/*
 * oma-dm-agent
 * Copyright (c) 2012 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*lib*/
#include <unistd.h>

/*sync-agent*/
#include <sync_agent.h>

/*dm-agent*/
#include "common/dm-status/oma_dm_status_db.h"
#include "common/dm_common.h"

#ifndef OMADM_AGENT_LOG
#undef LOG_TAG
#define LOG_TAG	"OMA_DM_DB"
#endif

/* static function define */
static int _busy_handler(void *pData, int count);
static DM_ERROR _query_exec(DB_HANDLER * db_handler, char *query, char *errMsg);
static db_stmt __query_prepare(DB_HANDLER * db_handler, char *query, int size);
static DM_ERROR __stmt_bind_text(DB_HANDLER * db_handler, db_stmt stmt, int index, const char *value);
static DM_ERROR _stmt_bind_int(DB_HANDLER * db_handler, db_stmt stmt, int index, const int value);
static DM_ERROR ___stmt_bind_null(DB_HANDLER * db_handler, db_stmt stmt, int index);
static DM_ERROR __stmt_step(DB_HANDLER * db_handler, db_stmt stmt);
/*static DM_ERROR				__stmt_reset(DB_HANDLER *db_handler, db_stmt stmt);*/
static DM_ERROR __stmt_finalize(DB_HANDLER * db_handler, db_stmt stmt);
static char *_stmt_column_text(db_stmt stmt, int index);
static int __stmt_column_int(db_stmt stmt, int index);
static DM_ERROR _get_table(DB_HANDLER * db_handler, char *query, char ***result, int *row_count, int *col_count);
static void __free_table(char **result);

static int _exist_table(DB_HANDLER * db_handler, const char *table_name);
static int _get_data_count(DB_HANDLER * db_handler, char *query, int size);
static char *_replace_engine_status_value(ENGINE_STATUS_VALUE engine_status_value);

static char *__create_table[] = {
	/* 1. create engine_status_tbl */
	"create table engine_status_tbl "
	    "( "
	    "engine_id integer, "
	    "engine_status integer default null, "
	    "server_id varchar(100) default null, "
	    "server_url varchar(100) default null, "
	    "correlator varchar(100) default null, "
	    "mo_path varchar(200) default null, "
	    "result_status integer not null, "
	    "ui_mode varchar(100) default null, " "task_id integer not null, " "ui_noti_type integer, " "download_click integer, " "update_time integer not null, " "constraint engine_status_tbl_pk primary key(engine_id)" ");",
};

static char *__delete_table[] = {
	/* 1. delete engine_status_tbl */
	"delete from engine_status_tbl",

};

DM_ERROR dm_open_agent(DB_HANDLER ** db_handler)
{
	_EXTERN_FUNC_ENTER;

	int ret = 0;
	char *errMsg = 0;

	if (*db_handler == NULL) {
		ret = sqlite3_open("/opt/usr/dbspace/.dm_service.db", db_handler);
		if (ret != SQLITE_OK)
			goto FINISH;

		/* register busy handler */
		ret = sqlite3_busy_handler(*db_handler, _busy_handler, 0);
		if (ret != SQLITE_OK) {
			_DEBUG_INFO("fail to register busy handler", errMsg);
			goto FINISH;
		}

		/* enable persist journal mode */
		ret = sqlite3_exec(*db_handler, "PRAGMA journal_mode = PERSIST", 0, 0, &errMsg);
		if (ret != SQLITE_OK) {
			_DEBUG_INFO("fail to change journal mode: %s", errMsg);
			goto FINISH;
		}
		_DEBUG_INFO("db_open success");
	} else {
		_DEBUG_INFO("db_already opened");
	}

	_EXTERN_FUNC_EXIT;
	return DM_OK;

 FINISH:

	_DEBUG_INFO("db_open failed(%d) : %s", ret, sqlite3_errmsg(*db_handler));

	if (errMsg != NULL)
		sqlite3_free(errMsg);

	sqlite3_close(*db_handler);
	*db_handler = 0;

	_EXTERN_FUNC_EXIT;
	return DM_ERR_UNKNOWN;
}

DM_ERROR dm_close_agent(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	int ret = 0;
	ret = sqlite3_close(db_handler);
	db_handler = 0;

	if (ret != SQLITE_OK) {
		_DEBUG_INFO("db_close failed(%d) : %s", ret, sqlite3_errmsg(db_handler));
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_DEBUG_INFO("db_close success");

	_EXTERN_FUNC_EXIT;

	return DM_OK;
}

DM_ERROR dm_create_table(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	if ((dm_begin_transaction(db_handler) != DM_OK)) {
		_DEBUG_INFO("mo_db_default_tbl_create failed");
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	/* 1. create engine_status_tbl */
	if (_exist_table(db_handler, "engine_status_tbl") == 0) {
		ret = _query_exec(db_handler, __create_table[0], "engine_status_tbl_create failed");
		if (ret != DM_OK)
			goto FINISH;
	}

	/* add normal engine */
	if (dm_isExist_engine_id(db_handler, 0) == 0) {
		engine_status status;
		memset(&status, 0x00, sizeof(engine_status));
		status.engine_id = 0;
		status.engine_status = 1;
		status.result_status = 0;
		status.task_id = 0;
		status.ui_noti_type = 0;
		status.download_click = 0;
		ret = dm_add_engine_status(db_handler, &status);
		_DEBUG_INFO("create dm service tbl : %d", ret);
	}

 FINISH:

	if (ret != DM_OK) {
		dm_end_transaction(db_handler, TRANSACTION_ROLLBACK_);
		ret = DM_ERR_UNKNOWN;
	} else {
		ret = dm_end_transaction(db_handler, TRANSACTION_COMMIT_);
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_delete_table(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	int i = 0;
	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	if ((dm_begin_transaction(db_handler) != DM_OK)) {
		_DEBUG_INFO("default_tbl_delete failed");
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	/* delete all default table */
	for (i = 0; i < DB_TABLE_NAME_MAX; i++) {
		ret = _query_exec(db_handler, __delete_table[i], "default_tbl_delete failed");
		if (ret != DM_OK)
			goto FINISH;
	}

 FINISH:

	if (ret != DM_OK) {
		dm_end_transaction(db_handler, TRANSACTION_ROLLBACK_);
		ret = DM_ERR_UNKNOWN;
	} else {
		ret = dm_end_transaction(db_handler, TRANSACTION_COMMIT_);
		_DEBUG_INFO("end transaction : %d", ret);
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_reset_table(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	if ((dm_begin_transaction(db_handler) != DM_OK)) {
		_DEBUG_INFO("default_tbl_delete failed");
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	/*ret = dm_delete_table(db_handler);
	   if (ret != DM_OK)
	   goto FINISH; */

	/* delete all default table */
	int i = 0;
	for (i = 0; i < DB_TABLE_NAME_MAX; i++) {
		ret = _query_exec(db_handler, __delete_table[i], "default_tbl_delete failed");
		if (ret != DM_OK)
			goto FINISH;
	}

	ret = dm_init_engine_status(db_handler);
	if (ret != DM_OK)
		goto FINISH;

 FINISH:

	if (ret != DM_OK) {
		dm_end_transaction(db_handler, TRANSACTION_ROLLBACK_);
		ret = DM_ERR_UNKNOWN;
	} else {
		ret = dm_end_transaction(db_handler, TRANSACTION_COMMIT_);
		_DEBUG_INFO("end transaction : %d", ret);
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_begin_transaction(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = _query_exec(db_handler, "begin immediate", "begin_transaction failed");
	_DEBUG_INFO("begin transaction : %d", ret);
/*	if (ret == SYNC_AGENT_DA_ERR_QUERY_FAILED)
		ret = SYNC_AGENT_DA_ERR_TRANSACTION_FAILED;*/

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_end_transaction(DB_HANDLER * db_handler, DB_TRANSACTION transaction)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	char *query = 0;
	char *errMsg = 0;

	if (transaction == TRANSACTION_COMMIT_) {
		query = "commit transaction";
		errMsg = "commit_transaction failed";
	} else if (transaction == TRANSACTION_ROLLBACK_) {
		query = "rollback transaction";
		errMsg = "rollback_transaction failed";
	}

	ret = _query_exec(db_handler, query, errMsg);
	if (ret == DM_ERR_UNKNOWN) {
		if (transaction == TRANSACTION_COMMIT_) {
			query = "rollback transaction";
			errMsg = "rollback_transaction failed";
			ret = _query_exec(db_handler, query, errMsg);
			if (ret == DM_OK)
				_DEBUG_INFO("rollback_transaction success");
		}
		ret = DM_ERR_UNKNOWN;
		_DEBUG_INFO("end transaction error : %d", ret);
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_init_engine_status(DB_HANDLER * db_handler)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_ERR_UNKNOWN;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	/* add normal engine */
	if (dm_isExist_engine_id(db_handler, 0) == 0) {
		engine_status status;
		memset(&status, 0x00, sizeof(engine_status));
		status.engine_id = 0;
		status.engine_status = 1;
		status.result_status = 0;
		status.task_id = 0;
		status.ui_noti_type = 0;
		status.download_click = 0;
		ret = dm_add_engine_status(db_handler, &status);
		_DEBUG_INFO("dm add engine status : %d", ret);
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_add_engine_status(DB_HANDLER * db_handler, engine_status * status)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	db_stmt stmt = 0;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");
	retvm_if(status == NULL, DM_ERR_UNKNOWN, "db status is NULL !!");

	char *query = "insert into engine_status_tbl (engine_id, engine_status, server_id, server_url, correlator, mo_path, result_status, ui_mode, task_id, ui_noti_type, download_click, update_time) " "values (?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?, ?)";

	stmt = __query_prepare(db_handler, query, strlen(query));
	if (stmt == NULL) {
		ret = DM_ERR_UNKNOWN;
		_DEBUG_INFO("ERROR DB");
		goto FINISH;
	}

	_stmt_bind_int(db_handler, stmt, 1, status->engine_id);
	_stmt_bind_int(db_handler, stmt, 2, status->engine_status);
	__stmt_bind_text(db_handler, stmt, 3, status->server_id);
	__stmt_bind_text(db_handler, stmt, 4, status->server_url);
	__stmt_bind_text(db_handler, stmt, 5, status->correlator);
	__stmt_bind_text(db_handler, stmt, 6, status->mo_path);
	_stmt_bind_int(db_handler, stmt, 7, status->result_status);
	__stmt_bind_text(db_handler, stmt, 8, status->ui_mode);
	_stmt_bind_int(db_handler, stmt, 9, status->task_id);
	_stmt_bind_int(db_handler, stmt, 10, status->ui_noti_type);
	_stmt_bind_int(db_handler, stmt, 11, status->download_click);
	_stmt_bind_int(db_handler, stmt, 12, (int)(time(&status->update_time)));

	ret = __stmt_step(db_handler, stmt);
	if (ret != DM_OK)
		ret = DM_ERR_UNKNOWN;

 FINISH:

	if (stmt != NULL)
		__stmt_finalize(db_handler, stmt);

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_delete_engine_status(DB_HANDLER * db_handler, int engine_id)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	db_stmt stmt = 0;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	char *query = "delete from engine_status_tbl where engine_id = ?";

	stmt = __query_prepare(db_handler, query, strlen(query));
	if (stmt == NULL) {
		ret = DM_ERR_UNKNOWN;
		goto FINISH;
	}

	_stmt_bind_int(db_handler, stmt, 1, engine_id);
	ret = __stmt_step(db_handler, stmt);

 FINISH:

	if (stmt != NULL)
		__stmt_finalize(db_handler, stmt);

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_update_engine_status(DB_HANDLER * db_handler, engine_status * status)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	db_stmt stmt = 0;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");
	retvm_if(status == NULL, DM_ERR_UNKNOWN, "db status is NULL !!");

	char *query = "update engine_status_tbl "
	    "set engine_status = ?, server_id = ?, server_url = ?, correlator = ?, mo_path = ?, result_status = ?, ui_mode = ?,  task_id = ?,  ui_noti_type = ?, download_click = ?, update_time = ? " "where engine_id = ?";

	stmt = __query_prepare(db_handler, query, strlen(query));
	if (stmt == NULL) {
		ret = DM_ERR_UNKNOWN;
		goto FINISH;
	}

	_stmt_bind_int(db_handler, stmt, 1, status->engine_status);
	__stmt_bind_text(db_handler, stmt, 2, status->server_id);
	__stmt_bind_text(db_handler, stmt, 3, status->server_url);
	__stmt_bind_text(db_handler, stmt, 4, status->correlator);
	__stmt_bind_text(db_handler, stmt, 5, status->mo_path);
	_stmt_bind_int(db_handler, stmt, 6, status->result_status);
	__stmt_bind_text(db_handler, stmt, 7, status->ui_mode);
	_stmt_bind_int(db_handler, stmt, 8, status->task_id);
	_stmt_bind_int(db_handler, stmt, 9, status->ui_noti_type);
	_stmt_bind_int(db_handler, stmt, 10, status->download_click);
	_stmt_bind_int(db_handler, stmt, 11, (int)(time(&status->update_time)));
	_stmt_bind_int(db_handler, stmt, 12, status->engine_id);

	ret = __stmt_step(db_handler, stmt);
	if (ret != DM_OK)
		ret = DM_ERR_UNKNOWN;

 FINISH:

	if (stmt != NULL)
		__stmt_finalize(db_handler, stmt);

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_update_engine_status_column(DB_HANDLER * db_handler, int engine_id, ENGINE_STATUS_VALUE engine_status_value, void *value)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	db_stmt stmt = 0;
	char query[MAX_QUERY_LENGTH] = { 0, };
	char *column_name = _replace_engine_status_value(engine_status_value);
	time_t update_time;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	snprintf(query, sizeof(query), "update engine_status_tbl set %s = ?, update_time= %d where engine_id = %d", column_name, (int)(time(&update_time)), engine_id);

	stmt = __query_prepare(db_handler, query, strlen(query));
	if (stmt == NULL) {
		ret = DM_ERR_UNKNOWN;
		goto FINISH;
	}

	if (engine_status_value == VALUE_ENGINE_STATUS || engine_status_value == VALUE_RESULT_STATUS || engine_status_value == VALUE_TASK_ID || engine_status_value == VALUE_UI_NOTI_TYPE || engine_status_value == VALUE_DOWNLOAD_CLICK) {
		int *engine_status = (int *)value;
		_stmt_bind_int(db_handler, stmt, 1, *engine_status);
	} else {
		char *str = (char *)value;
		__stmt_bind_text(db_handler, stmt, 1, str);
	}

	ret = __stmt_step(db_handler, stmt);
	if (ret != DM_OK)
		ret = DM_ERR_UNKNOWN;

 FINISH:

	if (stmt != NULL)
		__stmt_finalize(db_handler, stmt);

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_get_engine_status(DB_HANDLER * db_handler, int engine_id, engine_status ** status)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	db_stmt stmt = 0;
	char *query = "select * from engine_status_tbl where engine_id = ?";
	engine_status *temp_status = 0;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	stmt = __query_prepare(db_handler, query, strlen(query));
	if (stmt != NULL) {
		_stmt_bind_int(db_handler, stmt, 1, engine_id);

		if (__stmt_step(db_handler, stmt) == DM_ERR_MORE_DATA) {
			temp_status = (engine_status *) calloc(1, sizeof(engine_status));
			if (temp_status == NULL) {
				_DEBUG_INFO("alloc fail");
				_EXTERN_FUNC_EXIT;
				return COMMON_ERR_ALLOC;
			}
			temp_status->engine_id = __stmt_column_int(stmt, 0);
			temp_status->engine_status = __stmt_column_int(stmt, 1);
			temp_status->server_id = _stmt_column_text(stmt, 2);
			temp_status->server_url = _stmt_column_text(stmt, 3);
			temp_status->correlator = _stmt_column_text(stmt, 4);
			temp_status->mo_path = _stmt_column_text(stmt, 5);
			temp_status->result_status = __stmt_column_int(stmt, 6);
			temp_status->ui_mode = _stmt_column_text(stmt, 7);
			temp_status->task_id = __stmt_column_int(stmt, 8);
			temp_status->ui_noti_type = __stmt_column_int(stmt, 9);
			temp_status->download_click = __stmt_column_int(stmt, 10);
			temp_status->update_time = __stmt_column_int(stmt, 11);
		}
		ret = __stmt_finalize(db_handler, stmt);
	} else {
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	(*status) = temp_status;

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR dm_get_aII_engine_status(DB_HANDLER * db_handler, engine_status ** status_list, int *count)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	engine_status *temp_status = 0;
	char **result = 0;
	int row_count = 0;
	int col_count = 0;
	int index = 6;
	int i = 0;
	char query[MAX_QUERY_LENGTH] = { 0, };

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	snprintf(query, sizeof(query), "select * from engine_status_tbl");

	ret = _get_table(db_handler, query, &result, &row_count, &col_count);
	if ((ret == DM_OK) && (row_count != 0) && (result != NULL)) {
		*count = row_count;
		temp_status = (engine_status *) calloc(row_count, sizeof(engine_status));
		if (temp_status == NULL) {
			_DEBUG_INFO("memory_allocation failed");
			ret = COMMON_ERR_ALLOC;
			goto FINISH;
		}

		for (i = 0; i < row_count; i++) {
			temp_status[i].engine_id = ATOI(result[index]);
			index++;
			temp_status[i].engine_status = ATOI(result[index]);
			index++;
			temp_status[i].server_id = STRDUP(result[index]);
			index++;
			temp_status[i].server_url = STRDUP(result[index]);
			index++;
			temp_status[i].correlator = STRDUP(result[index]);
			index++;
			temp_status[i].mo_path = STRDUP(result[index]);
			index++;
			temp_status[i].result_status = ATOI(result[index]);
			index++;
			temp_status[i].ui_mode = STRDUP(result[index]);
			index++;
			temp_status[i].task_id = ATOI(result[index]);
			index++;
			temp_status[i].ui_noti_type = ATOI(result[index]);
			index++;
			temp_status[i].download_click = ATOI(result[index]);
			index++;
			temp_status[i].update_time = ATOI(result[index]);
			index++;
		}
	}

 FINISH:

	if (result != NULL)
		__free_table(result);

	*status_list = temp_status;

	_EXTERN_FUNC_EXIT;
	return ret;
}

int dm_isExist_engine_id(DB_HANDLER * db_handler, int engine_id)
{
	_EXTERN_FUNC_ENTER;

	int ret = 1;
	int dataCount = 0;
	char query[MAX_QUERY_LENGTH] = { 0, };

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	snprintf(query, sizeof(query), "select count(*) from engine_status_tbl where engine_id = %d", engine_id);

	dataCount = _get_data_count(db_handler, query, strlen(query));

	if (dataCount == 0)
		ret = 0;

	_EXTERN_FUNC_EXIT;
	return ret;
}

int _busy_handler(void *pData, int count)
{
	_INNER_FUNC_ENTER;

	_DEBUG_TRACE("__busy_handler %d called", count);

	/*  sleep time when SQLITE_LOCK */
	usleep(100000);

	_INNER_FUNC_EXIT;
	/* retry will be stopped if  busy handler return 0 */
	return RETRY_COUNT - count;
}

DM_ERROR _query_exec(DB_HANDLER * db_handler, char *query, char *errMsg)
{
	_INNER_FUNC_ENTER;

	char *queryMsg = 0;
	int ret = 0;

	retvm_if(db_handler == NULL, DM_ERR_UNKNOWN, "db handler is NULL !!");

	ret = sqlite3_exec(db_handler, query, 0, 0, &queryMsg);
	if (ret != SQLITE_OK) {
		_DEBUG_TRACE("%s(%d) : %s", errMsg, ret, queryMsg);

		if (queryMsg != NULL)
			sqlite3_free(queryMsg);

/*
		if (ret == SQLITE_BUSY)
			return DM_ERR_UNKNOWN;
*/
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	return DM_OK;
	_INNER_FUNC_EXIT;
}

db_stmt __query_prepare(DB_HANDLER * db_handler, char *query, int size)
{
	_INNER_FUNC_ENTER;

	int ret = 0;
	db_stmt stmt = 0;

	ret = sqlite3_prepare_v2(db_handler, query, size, &stmt, 0);
	if (ret != SQLITE_OK)
		_DEBUG_VERBOSE("sqlite3_query_prepare failed(%d) : %s ", ret, sqlite3_errmsg(db_handler));

	_INNER_FUNC_EXIT;

	return stmt;
}

DM_ERROR __stmt_bind_text(DB_HANDLER * db_handler, db_stmt stmt, int index, const char *value)
{
	_INNER_FUNC_ENTER;

	int ret = 0;

	if (value != NULL) {
		ret = sqlite3_bind_text(stmt, index, value, strlen(value), SQLITE_STATIC);
	} else {
		ret = ___stmt_bind_null(db_handler, stmt, index);
		if (ret == DM_OK) {
			ret = SQLITE_OK;
		}
	}

	if (ret != SQLITE_OK) {
		_DEBUG_VERBOSE("sqlite3_stmt_bind_text failed(%d) : %s ", ret, sqlite3_errmsg(db_handler));
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_INNER_FUNC_EXIT;
	return DM_OK;
}

DM_ERROR _stmt_bind_int(DB_HANDLER * db_handler, db_stmt stmt, int index, const int value)
{
	_INNER_FUNC_ENTER;

	int ret = 0;

	ret = sqlite3_bind_int(stmt, index, value);
	if (ret != SQLITE_OK) {
		_DEBUG_TRACE("sqlite3_stmt_bind_int failed(%d) : %s \n", ret, sqlite3_errmsg(db_handler));
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_INNER_FUNC_EXIT;

	return DM_OK;
}

DM_ERROR ___stmt_bind_null(DB_HANDLER * db_handler, db_stmt stmt, int index)
{
	_INNER_FUNC_ENTER;

	int ret = 0;

	ret = sqlite3_bind_null(stmt, index);
	if (ret != SQLITE_OK) {
		_DEBUG_VERBOSE("sqlite3_stmt_bind_null failed(%d) : %s", ret, sqlite3_errmsg(db_handler));
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_INNER_FUNC_EXIT;
	return DM_OK;
}

DM_ERROR __stmt_step(DB_HANDLER * db_handler, db_stmt stmt)
{
	_INNER_FUNC_ENTER;
	int ret = 0;

	ret = sqlite3_step(stmt);

	if (ret == SQLITE_ROW) {
		_INNER_FUNC_EXIT;
		return DM_ERR_MORE_DATA;
	}

	if (ret != SQLITE_DONE) {
		_DEBUG_VERBOSE("sqlite3_stmt_step failed(%d) : %s", ret, sqlite3_errmsg(db_handler));

/*		if (ret == SQLITE_BUSY)
			return DM_ERR_UNKNOWN;*/
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_INNER_FUNC_EXIT;
	return DM_OK;
}

/*DM_ERROR __stmt_reset(DB_HANDLER *db_handler, db_stmt stmt)
{
	int ret = 0;

	if (sqlite3_reset(stmt) != SQLITE_OK) {
		_DEBUG_INFO("sqlite3_stmt_reset failed(%d) : %s", ret, sqlite3_errmsg(db_handler));
		return DM_ERR_UNKNOWN;
	}

	return DM_OK;
}*/

DM_ERROR __stmt_finalize(DB_HANDLER * db_handler, db_stmt stmt)
{
	_INNER_FUNC_ENTER;
	int ret = 0;

	if (sqlite3_finalize(stmt) != SQLITE_OK) {
		_DEBUG_VERBOSE("sqlite3_stmt_finalize failed(%d) : %s", ret, sqlite3_errmsg(db_handler));
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_INNER_FUNC_EXIT;
	return DM_OK;
}

char *_stmt_column_text(db_stmt stmt, int index)
{
	_EXTERN_FUNC_ENTER;

	char *temp = 0;
	temp = (char *)sqlite3_column_text(stmt, index);

	_EXTERN_FUNC_EXIT;

	return STRDUP(temp);
}

int __stmt_column_int(db_stmt stmt, int index)
{
	_INNER_FUNC_ENTER;

	_INNER_FUNC_EXIT;
	return (int)sqlite3_column_int(stmt, index);
}

DM_ERROR _get_table(DB_HANDLER * db_handler, char *query, char ***result, int *row_count, int *col_count)
{
	_EXTERN_FUNC_ENTER;

	int ret = 0;
	char *errMsg;

	if (db_handler == NULL) {
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	ret = sqlite3_get_table(db_handler, query, result, row_count, col_count, &errMsg);

	if (ret != SQLITE_OK) {
		_DEBUG_TRACE("sqlite3_get_table failed(%d) : %s", ret, errMsg);

		__free_table(*result);

		if (errMsg != NULL)
			sqlite3_free(errMsg);

		_EXTERN_FUNC_EXIT;

		return DM_ERR_UNKNOWN;
	}

	_EXTERN_FUNC_EXIT;

	return DM_OK;
}

void __free_table(char **result)
{
	_INNER_FUNC_ENTER;

	if (result != NULL)
		sqlite3_free_table(result);

	_INNER_FUNC_EXIT;

}

int _exist_table(DB_HANDLER * db_handler, const char *table_name)
{
	_INNER_FUNC_ENTER;

	db_stmt stmt = 0;
	int table_count = 0;
	char *query = "select count(*) from sqlite_master where tbl_name= ?";

	stmt = __query_prepare(db_handler, query, strlen(query));
	if ((stmt != NULL) && (__stmt_bind_text(db_handler, stmt, 1, table_name) == DM_ERR_MORE_DATA)) {

		if (__stmt_step(db_handler, stmt) == DM_ERR_MORE_DATA)
			table_count = __stmt_column_int(stmt, 0);

		__stmt_finalize(db_handler, stmt);

		if (table_count > 0) {
			_INNER_FUNC_EXIT;
			return 1;
		}
	}

	_INNER_FUNC_EXIT;

	return 0;
}

int _get_data_count(DB_HANDLER * db_handler, char *query, int size)
{
	_INNER_FUNC_ENTER;

	db_stmt stmt = 0;
	int data_count = 0;

	stmt = __query_prepare(db_handler, query, size);
	if (stmt != NULL) {
		if (__stmt_step(db_handler, stmt) == DM_ERR_MORE_DATA)
			data_count = __stmt_column_int(stmt, 0);

		__stmt_finalize(db_handler, stmt);
	}

	_INNER_FUNC_EXIT;

	return data_count;
}

static char *_replace_engine_status_value(ENGINE_STATUS_VALUE engine_status_value)
{
	_INNER_FUNC_ENTER;

	char *temp = 0;

	switch (engine_status_value) {
	case VALUE_ENGINE_ID:
		temp = "engine_id";
		break;
	case VALUE_ENGINE_STATUS:
		temp = "engine_status";
		break;
	case VALUE_SERVER_ID:
		temp = "server_id";
		break;
	case VALUE_SERVER_URL:
		temp = "server_url";
		break;
	case VALUE_CORRELATOR:
		temp = "correlator";
		break;
	case VALUE_MO_PATH:
		temp = "mo_path";
		break;
	case VALUE_RESULT_STATUS:
		temp = "result_status";
		break;
	case VALUE_UI_MODE:
		temp = "ui_mode";
		break;
	case VALUE_TASK_ID:
		temp = "task_id";
		break;
	case VALUE_UI_NOTI_TYPE:
		temp = "ui_noti_type";
		break;
	case VALUE_DOWNLOAD_CLICK:
		temp = "download_click";
		break;
	case VALUE_UPDATE_TIME:
		temp = "update_time";
		break;
	default:
		temp = 0;
	}

	_INNER_FUNC_EXIT;

	return temp;
}
