/*
 * oma-dm-agent
 * Copyright (c) 2012 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*sync-agent*/
#include <sync_agent.h>

/*dm-agent*/
#include "common/dm-status/oma_dm_status_db_handler.h"
#include "common/dm-status/oma_dm_status_db.h"

#ifndef OMADM_AGENT_LOG
#undef LOG_TAG
#define LOG_TAG	"OMA_DM_DB"
#endif

static db_handler_mgr *pDBHandlerMgr = 0;

/* static function define */
static DM_ERROR _add_agent_handler(unsigned int key, DB_HANDLER * pHandler);
static DM_ERROR _remove_agent_handler(unsigned int key);
static DB_HANDLER *_get_agent_handler(unsigned int key);
static DM_ERROR __alloc_agent_dc_handlerTable();
static DM_ERROR __alloc_agent_dc_handlerTable_mutex();
static DM_ERROR _free_agent_dc_handlerTable();
static DM_ERROR _free_agent_dc_handlerTable_mutex();
static DM_ERROR _create_agentDBHandlerMgrInfo();
static void ___destroy_agentDBHashTableValue(void *pHandler);
static void ___destory_agentDBHashTableKey(void *key);

/* test function define*/
/*static void					__print_agentDBHashTableLog();*/

/* general function implementation */
/*
 * ==================================
 * external API (1. about db handler mgr)
 * ==================================
 */
DM_ERROR Alloc_DB_Handler_Mgr()
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	/* alloc the agentDBHandlerMgr */
	pDBHandlerMgr = (db_handler_mgr *) calloc(1, sizeof(db_handler_mgr));
	if (pDBHandlerMgr == NULL) {
		_DEBUG_INFO("alloc fail");
		_EXTERN_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	_DEBUG_INFO("[%s] pMoDBHandlerMgr alloc success !\n", __func__);

	/* alloc the agentDBHandlerMgr information */
	ret = _create_agentDBHandlerMgrInfo();
	if (ret != DM_OK)
		_DEBUG_INFO("[%s] __create_agentDBHandlerMgrInfo alloc fail !\n", __func__);
	else
		_DEBUG_INFO("[%s] __create_agentDBHandlerMgrInfo alloc success !\n", __func__);

	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR Free_DB_Handler_Mgr()
{
	_EXTERN_FUNC_ENTER;
	DM_ERROR ret = DM_OK;

	if (pDBHandlerMgr != NULL) {

		ret = _free_agent_dc_handlerTable_mutex();
		if (ret != DM_OK) {
			_EXTERN_FUNC_EXIT;
			return DM_ERR_UNKNOWN;
		}

		ret = _free_agent_dc_handlerTable();
		if (ret != DM_OK) {
			_EXTERN_FUNC_EXIT;
			return DM_ERR_UNKNOWN;
		}
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

void Free_Memory_Engine_Status(engine_status ** status, int count)
{
	_EXTERN_FUNC_ENTER;
	engine_status *temp_status = *status;

	int i;
	for (i = 0; i < count; i++) {
		MEMORY_FREE(temp_status[i].server_id);
		MEMORY_FREE(temp_status[i].server_url);
		MEMORY_FREE(temp_status[i].correlator);
		MEMORY_FREE(temp_status[i].ui_mode);
		MEMORY_FREE(temp_status[i].mo_path);
	}

	MEMORY_FREE(temp_status);
	*status = 0;
	_EXTERN_FUNC_EXIT;
}

/*
 * ==================================
 * external API (2. about agent db )
 * ==================================
 */
DM_ERROR DB_Open()
{
	_EXTERN_FUNC_ENTER;
	int ret = 0;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {

		_DEBUG_INFO("db_handler is 0");
		_DEBUG_INFO("should register db_handler to db_handler_hash_table");
		ret = dm_open_agent(&db_handler);
		if (ret != DM_OK) {
			_EXTERN_FUNC_EXIT;
			return ret;
		}

		if (db_handler != NULL) {
			/* add agentDB handler to agentDBHandleMgr */
			ret = _add_agent_handler(GET_THREAD_ID, db_handler);
			if (ret != DM_OK) {
				_EXTERN_FUNC_EXIT;
				return ret;
			}
		}
	} else {
		_DEBUG_INFO("db_handler is not null");
		_DEBUG_INFO("should not register db_handler to db_handler_hash_table");
	}

	_DEBUG_INFO("agent_db_agent_open_success");
	_EXTERN_FUNC_EXIT;
	return DM_OK;
}

void DB_Close()
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler != NULL) {

		/* remove agentDB handler to agentDBHandleMgr */
		ret = _remove_agent_handler(GET_THREAD_ID);
		if (ret != DM_OK) {
			_DEBUG_INFO("agent_db_agent_close_fail");
			_EXTERN_FUNC_EXIT;
			return;
		}
		_DEBUG_INFO("agent_db_agent_close_success");
	}
	_EXTERN_FUNC_EXIT;
}

DM_ERROR DB_Create_Table()
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_create_table(db_handler);
}

DM_ERROR DB_Delete_Table()
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_delete_table(db_handler);
}

DM_ERROR DB_Reset_Table()
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_reset_table(db_handler);
}

DM_ERROR DB_Begin_Transaction()
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_begin_transaction(db_handler);
}

DM_ERROR DB_End_Transaction(DB_TRANSACTION transaction)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_end_transaction(db_handler, transaction);
}

DM_ERROR Init_Engine_Status()
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_init_engine_status(db_handler);
}

DM_ERROR Add_Engine_Status(engine_status * status)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_add_engine_status(db_handler, status);
}

DM_ERROR Delete_Engine_Status(int engine_id)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_EXTERN_FUNC_EXIT;
	return dm_delete_engine_status(db_handler, engine_id);
}

DM_ERROR Update_Engine_Status(engine_status * status)
{
	_EXTERN_FUNC_ENTER;

	retvm_if(status == NULL, DM_ERR_UNKNOWN, "engine status is NULL !!");
	_DEBUG_INFO("status [%d]", status);

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	int ret = dm_isExist_engine_id(db_handler, status->engine_id);
	if (ret == 0) {
		_EXTERN_FUNC_EXIT;
		return dm_add_engine_status(db_handler, status);
	} else {
		_DEBUG_INFO("status->engine_id [%d]", status->engine_id);
		_DEBUG_INFO("status->download_click [%d]", status->download_click);
		_EXTERN_FUNC_EXIT;
		return dm_update_engine_status(db_handler, status);
	}
}

DM_ERROR Update_Engine_Status_Column(int engine_id, ENGINE_STATUS_VALUE engine_status_value, void *value)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_DEBUG_INFO("engine_status_value[%d]",engine_status_value);
	if( engine_status_value == VALUE_RESULT_STATUS) {
		_DEBUG_INFO("value [%d]", *(int*)value);
	}

	_EXTERN_FUNC_EXIT;
	return dm_update_engine_status_column(db_handler, engine_id, engine_status_value, value);
}

DM_ERROR Get_Engine_Status(int engine_id, engine_status ** status)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}
	_EXTERN_FUNC_EXIT;
	return dm_get_engine_status(db_handler, engine_id, status);
}

DM_ERROR Get_AII_Engine_Status(engine_status ** status_list, int *count)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_EXTERN_FUNC_EXIT;
	return dm_get_aII_engine_status(db_handler, status_list, count);
}

int IsExist_Engine_id(int engine_id)
{
	_EXTERN_FUNC_ENTER;

	DB_HANDLER *db_handler = _get_agent_handler(GET_THREAD_ID);
	if (db_handler == NULL) {
		_DEBUG_INFO("[%s] no handler for key\n", __func__);
		_EXTERN_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	_EXTERN_FUNC_EXIT;

	return dm_isExist_engine_id(db_handler, engine_id);
}

/*
 * ==================================
 *  DB module internal API
 * ==================================
 */
DM_ERROR _add_agent_handler(unsigned int key, DB_HANDLER * pHandler)
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr == NULL || pDBHandlerMgr->db_handlerTable_mutex == NULL) {
		return DM_ERR_UNKNOWN;
	}

	DM_ERROR ret = DM_OK;

	unsigned int *pKey_copy = (unsigned int *)calloc(1, sizeof(unsigned int));
	if (pKey_copy == NULL) {
		_INNER_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}

	*pKey_copy = key;

	pthread_mutex_lock(pDBHandlerMgr->db_handlerTable_mutex);
	if (pDBHandlerMgr->db_handlerTable == NULL) {
		ret = __alloc_agent_dc_handlerTable();
		if (ret != DM_OK)
			goto addagentHandler;
	}

	/* insert to handler */
	_DEBUG_TRACE("[%s] Key  =  %d, Handler  =  %p  !\n", __func__, key, pHandler);
	g_hash_table_insert(pDBHandlerMgr->db_handlerTable, pKey_copy, pHandler);

#ifdef DC_PRINT
	/* For test log */
	__print_agentDBHashTableLog();
#endif
	pthread_mutex_unlock(pDBHandlerMgr->db_handlerTable_mutex);
/* pKey_copy will be free from _remove_agent_handler function.
	if(pKey_copy != NULL) {
		free(pKey_copy);
	}
*/	
	_INNER_FUNC_EXIT;
	return ret;

 addagentHandler:
	pthread_mutex_unlock(pDBHandlerMgr->db_handlerTable_mutex);
	if(pKey_copy != NULL) {
		free(pKey_copy);
	}
	_INNER_FUNC_EXIT;
	return ret;
}

DM_ERROR _remove_agent_handler(unsigned int key)
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr == NULL || pDBHandlerMgr->db_handlerTable_mutex == NULL || pDBHandlerMgr->db_handlerTable == NULL) {
		_INNER_FUNC_EXIT;
		return DM_ERR_UNKNOWN;
	}

	/* remove the handler to hash table */
	pthread_mutex_lock(pDBHandlerMgr->db_handlerTable_mutex);
	int isSuccess = g_hash_table_remove(pDBHandlerMgr->db_handlerTable, &key);
	pthread_mutex_unlock(pDBHandlerMgr->db_handlerTable_mutex);

#ifdef DC_PRINT
	/* for test log */
	__print_agentDBHashTableLog();
#endif

	_DEBUG_TRACE("[%s] end !\n", __func__);
	if (isSuccess == 1) {
		_EXTERN_FUNC_EXIT;
		return DM_OK;
	}

	_INNER_FUNC_EXIT;

	return DM_ERR_UNKNOWN;
}

DB_HANDLER *_get_agent_handler(unsigned int key)
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr == NULL || pDBHandlerMgr->db_handlerTable_mutex == NULL || pDBHandlerMgr->db_handlerTable == NULL) {
		_DEBUG_TRACE("[%s] DM_ERR_UNKNOWN !\n", __func__);
		_INNER_FUNC_EXIT;

		return 0;
	}

	/* get the handler to hash table */
	pthread_mutex_lock(pDBHandlerMgr->db_handlerTable_mutex);
	DB_HANDLER *pHandler = g_hash_table_lookup(pDBHandlerMgr->db_handlerTable, &key);
	pthread_mutex_unlock(pDBHandlerMgr->db_handlerTable_mutex);

#ifdef DC_PRINT
	if (pHandler != NULL) {
		_DEBUG_TRACE("[%s] ######### > Value Search Success(key = %d, handler = %p)\n", __func__, key, pHandler);
	} else {
		_DEBUG_TRACE("[%s] ######### > Value Search Fail(key = %d, handler = %p)\n", __func__, key, pHandler);
	}
#endif
	_INNER_FUNC_EXIT;

	return pHandler;
}

/* static function implementation */
static DM_ERROR __alloc_agent_dc_handlerTable()
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr != NULL) {
		pDBHandlerMgr->db_handlerTable = g_hash_table_new_full(g_int_hash, g_int_equal, ___destory_agentDBHashTableKey, ___destroy_agentDBHashTableValue);
		if (pDBHandlerMgr->db_handlerTable == NULL) {
			_DEBUG_TRACE("[%s] Not Create agent dc handler hash table \n", __func__);
			_INNER_FUNC_EXIT;
			return DM_ERR_UNKNOWN;
		} else {
			_DEBUG_TRACE("[%s] create agent dc handler hash table \n", __func__);
		}
	}

	_INNER_FUNC_EXIT;

	return DM_OK;
}

static DM_ERROR __alloc_agent_dc_handlerTable_mutex()
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr != NULL) {

		pDBHandlerMgr->db_handlerTable_mutex = (pthread_mutex_t *) calloc(1, sizeof(pthread_mutex_t));
		if (pDBHandlerMgr->db_handlerTable_mutex == NULL) {
			_DEBUG_TRACE("[%s] Not Create agent dc handler hash table mutex \n", __func__);
			_INNER_FUNC_EXIT;
			return COMMON_ERR_ALLOC;
		} else {

			if (pthread_mutex_init(pDBHandlerMgr->db_handlerTable_mutex, 0) != 0) {
				_DEBUG_TRACE("[%s] Not init agent dc handler hash table mutex\n", __func__);
				_INNER_FUNC_EXIT;
				return DM_ERR_UNKNOWN;
			}
			_DEBUG_TRACE("[%s] Init agent dc handler hash table mutex\n", __func__);
		}
	}

	_INNER_FUNC_EXIT;

	return DM_OK;
}

static DM_ERROR _free_agent_dc_handlerTable()
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr != NULL && pDBHandlerMgr->db_handlerTable != NULL) {
		g_hash_table_destroy(pDBHandlerMgr->db_handlerTable);
		_DEBUG_TRACE("[%s] db_handlerTable free success !\n", __func__);
		pDBHandlerMgr->db_handlerTable = 0;;
	}
	_INNER_FUNC_EXIT;

	return DM_OK;
}

static DM_ERROR _free_agent_dc_handlerTable_mutex()
{
	_INNER_FUNC_ENTER;

	if (pDBHandlerMgr != NULL && pDBHandlerMgr->db_handlerTable_mutex != NULL) {

		int ret = pthread_mutex_destroy(pDBHandlerMgr->db_handlerTable_mutex);
		if (ret != 0) {
			_DEBUG_TRACE("[%s] db_handlerTable_mutex free error !\n", __func__);
			return DM_ERR_UNKNOWN;
		} else {
			free(pDBHandlerMgr->db_handlerTable_mutex);
			pDBHandlerMgr->db_handlerTable_mutex = 0;
			_DEBUG_TRACE("[%s] db_handlerTable_mutex free success !\n", __func__);
		}
	}
	_INNER_FUNC_EXIT;

	return DM_OK;
}

static DM_ERROR _create_agentDBHandlerMgrInfo()
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	if (pDBHandlerMgr != NULL) {

		/* alloc agent db hander hash table */
		ret = __alloc_agent_dc_handlerTable();
		if (ret != DM_OK)
			goto allocAgentDBHandlerMgrInfoErr;

		/* alloc agent db hander hash table mutex */
		ret = __alloc_agent_dc_handlerTable_mutex();
		if (ret != DM_OK)
			goto allocAgentDBHandlerMgrInfoErr;
	}
	_INNER_FUNC_EXIT;

	return DM_OK;

 allocAgentDBHandlerMgrInfoErr:
	Free_DB_Handler_Mgr();

	_INNER_FUNC_EXIT;
	return DM_ERR_UNKNOWN;
}

static void ___destroy_agentDBHashTableValue(void *pHandler)
{
	_INNER_FUNC_ENTER;

	_DEBUG_VERBOSE("[%s] start \n", __func__);
	if ((DB_HANDLER *) pHandler != 0) {
		dm_close_agent((DB_HANDLER *) pHandler);
	}

	_INNER_FUNC_EXIT;
}

static void ___destory_agentDBHashTableKey(void *key)
{
	_INNER_FUNC_ENTER;

	_DEBUG_VERBOSE("[%s] start \n", __func__);
	if ((unsigned int *)key != 0) {
		free((unsigned int *)key);
	}
	_INNER_FUNC_EXIT;

}

/*
static void __print_agentDBHashTableLog()
{
	_DEBUG_VERBOSE("[%s] ---------HashKeyList --------\n", __func__);
	int size = g_hash_table_size(pDBHandlerMgr->db_handlerTable);
	_DEBUG_VERBOSE("[%s] pDBHandlerMgr->db_handlerTable Size is %d\n", __func__, size);
	_DEBUG_VERBOSE("[%s] ----------------------------------------\n", __func__);
	GList *iter = 0;
	GList *keyList = g_hash_table_get_keys(pDBHandlerMgr->db_handlerTable);
	for (iter = keyList; iter != 0; iter = g_list_next(iter))
		_DEBUG_VERBOSE("[%s] Handler Key  %d\n", __func__, *((unsigned int *)(iter->data)));
	_DEBUG_VERBOSE("[%s] ----------------------------------------\n", __func__);
}
*/
