/*
 * oma-dm-agent
 * Copyright (c) 2012 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*lib*/
#include <glib.h>
#include <glib/gprintf.h>

/*sync-agent*/
#include <sync_agent.h>

/*dm-agent*/
#include "dm-engine/cp/dm_cp_struct.h"
#include "dm-engine/cp/dm_cp_parser.h"
#include "dm-engine/cp/dm_cp_processor.h"
#include "serviceadapter/protocolbinder/syncml_protocol_binder.h"

#ifndef OMADM_AGENT_LOG
#undef LOG_TAG
#define LOG_TAG	"OMA_DM_CP"
#endif

static String_Node *___get_Last_String_Node(String_Node * node);
static DM_CP_Bootstrap *__parse_Bootstrap(WBXMLTreeNode * wbxml_dom_node);
static DM_CP_Napdef *__parse_Napdef(WBXMLTreeNode * wbxml_dom_node);
static DM_CP_Application *__parse_Application(WBXMLTreeNode * wbxml_dom_node);
static DM_CP_PxLogical *__parse_PxLogical(WBXMLTreeNode * wbxml_dom_node);
static void ___set_Param(WBXMLTreeNode * wbxml_dom_node, char **set_value);
static sync_agent_pb_error_e _characteristic_start_reverse_converter_function(sync_agent_pb_protocol_binder_reverse_info_s * pReverse_binder, WBXMLTreeNode * wbxml_dom_node, Content_Ptr * pContent);

const sync_agent_pb_function_info_s cp_binder_function_info[] = {
	{TNDS_TAG_CHARACTERISTIC_START, "characteristic", NULL, _characteristic_start_reverse_converter_function, false},
};

DM_CP *parsing_cp_stream(CP_TYPE cp_type, const char *wbxml_stream, unsigned int msgSize)
{
	_EXTERN_FUNC_ENTER;

	retvm_if((wbxml_stream) == NULL, NULL, "wbxml_stream is NULL!!");

	sync_agent_pb_protocol_binder_function_set_s *pBinder_function_set = NULL;
	sync_agent_pb_error_e pb_error = sync_agent_init_binder_function_set(sizeof(cp_binder_function_info) / sizeof(sync_agent_pb_function_info_s),
									     (sync_agent_pb_function_info_s *) cp_binder_function_info, &pBinder_function_set);

	if (pb_error != SYNC_AGENT_PB_RETURN_OK) {
		_DEBUG_INFO("Failed to sync_agent_init_binder_function_set()");
		_EXTERN_FUNC_EXIT;
		return 0;
	}

	/*
	 * reverse protocol binder initialization
	 */
	sync_agent_pb_protocol_e protocol = SYNC_AGENT_PB_PROTOCOL_PROV10;
	sync_agent_pb_protocol_binder_reverse_info_s *pBinder = 0;
	DM_CP *dm_cp = (DM_CP *) calloc(1, sizeof(DM_CP));
	if (dm_cp == NULL) {
		_DEBUG_INFO("alloc fail");
		free_syncml_binder_function_set(pBinder_function_set);
		_EXTERN_FUNC_EXIT;
		return 0;
	}

	switch (cp_type) {
	case CP_TYPE_WBXML:
		pb_error = sync_agent_init_reverse_protocol_binder((char *)wbxml_stream, msgSize, SYNC_AGENT_PB_DECODING_WBXML, &protocol, pBinder_function_set, dm_cp, &pBinder);
		break;
	case CP_TYPE_XML:
		pb_error = sync_agent_init_reverse_protocol_binder((char *)wbxml_stream, msgSize, SYNC_AGENT_PB_DECODING_XML, &protocol, pBinder_function_set, dm_cp, &pBinder);
		break;
	case CP_TYPE_NONE:
		break;
	default:
		break;
	}

	if (pb_error != SYNC_AGENT_PB_RETURN_OK) {
		_DEBUG_INFO("Failed to call sync_agent_init_reverse_protocol_binder() - [%d]", pb_error);
	}

	Protocol_Element protocol_element;
	char *protocol_element_name = 0;
	Content_Ptr content = 0;

	while (1) {
		pb_error = sync_agent_next_element(pBinder, &protocol_element, &protocol_element_name, &content);

		if (pb_error == SYNC_AGENT_PB_RETURN_HAS_NO_MORE_ELEMENT) {
			break;
		} else if (pb_error != SYNC_AGENT_PB_RETURN_OK) {
			/*
			 * process error code
			 */
		}
	}
	_EXTERN_FUNC_EXIT;

	return dm_cp;
}

static sync_agent_pb_error_e _characteristic_start_reverse_converter_function(sync_agent_pb_protocol_binder_reverse_info_s * pReverse_binder, WBXMLTreeNode * wbxml_dom_node, Content_Ptr * pContent)
{
	_INNER_FUNC_ENTER;

	sync_agent_pb_error_e pb_error = SYNC_AGENT_PB_RETURN_OK;

	retvm_if((pReverse_binder) == NULL, SYNC_AGENT_PB_RETURN_OUT_OF_MEMORY, "pReverse_binder is NULL!!");
	retvm_if((wbxml_dom_node) == NULL, SYNC_AGENT_PB_RETURN_OUT_OF_MEMORY, "wbxml_dom_node is NULL!!");

	void *user_data = sync_agent_get_user_data_from_reverse_protocol_binder(pReverse_binder);
	DM_CP *dm_cp = (DM_CP *) user_data;

	WBXMLList *attrs = wbxml_dom_node->attrs;
	WB_ULONG index = 0;
	WB_ULONG nb_attrs = wbxml_list_len(attrs);

	while (index < nb_attrs) {
		WBXMLAttribute *attr = (WBXMLAttribute *) wbxml_list_get(attrs, index);

		if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(attr)) == 0) {
			if (WBXML_STRCMP("APPLICATION", wbxml_attribute_get_xml_value(attr)) == 0) {
				DM_CP_Application *application = __parse_Application(wbxml_dom_node);
				if (application == NULL) {
					index++;
					continue;
				}
				if (dm_cp->applications == NULL) {
					dm_cp->applications = application;
				} else {
					DM_CP_Application *cursor_ptr = dm_cp->applications;
					while (cursor_ptr != 0) {
						if (cursor_ptr->next == NULL) {
							cursor_ptr->next = application;
							break;
						}
						cursor_ptr = cursor_ptr->next;
					}
				}
			} else if (WBXML_STRCMP("BOOTSTRAP", wbxml_attribute_get_xml_value(attr)) == 0) {
				DM_CP_Bootstrap *bootstrap = __parse_Bootstrap(wbxml_dom_node);
				if (bootstrap == NULL) {
					index++;
					continue;
				}
				if (dm_cp->bootstraps == NULL) {
					dm_cp->bootstraps = bootstrap;
				} else {
					DM_CP_Bootstrap *cursor_ptr = dm_cp->bootstraps;
					while (cursor_ptr != NULL) {
						if (cursor_ptr->next == NULL) {
							cursor_ptr->next = bootstrap;
							break;
						}
						cursor_ptr = cursor_ptr->next;
					}
				}
			} else if (WBXML_STRCMP("NAPDEF", wbxml_attribute_get_xml_value(attr)) == 0) {
				DM_CP_Napdef *napdef = __parse_Napdef(wbxml_dom_node);
				if (napdef == NULL) {
					index++;
					continue;
				}
				if (dm_cp->napdefs == NULL) {
					dm_cp->napdefs = napdef;
				} else {
					DM_CP_Napdef *cursor_ptr = dm_cp->napdefs;
					while (cursor_ptr != NULL) {
						if (cursor_ptr->next == NULL) {
							cursor_ptr->next = napdef;
							break;
						}
						cursor_ptr = cursor_ptr->next;
					}
				}
			} else if (WBXML_STRCMP("PXLOGICAL", wbxml_attribute_get_xml_value(attr)) == 0) {
				DM_CP_PxLogical *pxLogical = __parse_PxLogical(wbxml_dom_node);
				if (pxLogical == NULL) {
					index++;
					continue;
				}
				if (dm_cp->pxLogical == NULL) {
					dm_cp->pxLogical = pxLogical;
				} else {
					DM_CP_PxLogical *cursor_ptr = dm_cp->pxLogical;
					while (cursor_ptr != NULL) {
						if (cursor_ptr->next == NULL) {
							cursor_ptr->next = pxLogical;
							break;
						}
						cursor_ptr = cursor_ptr->next;
					}
				}
			}
		}

		index++;
	}

	_INNER_FUNC_EXIT;
	return pb_error;
}

static String_Node *___get_Last_String_Node(String_Node * node)
{
	_INNER_FUNC_ENTER;

	String_Node *new_node = (String_Node *) calloc(1, sizeof(String_Node));
	if (new_node == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return 0;
	}
	new_node->next = NULL;

	String_Node *cursor_ptr = node;
	while (cursor_ptr != NULL) {
		if (cursor_ptr->next == NULL) {
			cursor_ptr->next = new_node;
			break;
		}

		cursor_ptr = cursor_ptr->next;
	}

	_INNER_FUNC_EXIT;

	return new_node;
}

static DM_CP_Bootstrap *__parse_Bootstrap(WBXMLTreeNode * wbxml_dom_node)
{
	_INNER_FUNC_ENTER;

	retvm_if((wbxml_dom_node) == NULL, NULL, "wbxml_dom_node is NULL!!");

	DM_CP_Bootstrap *bootstrap = (DM_CP_Bootstrap *) calloc(1, sizeof(DM_CP_Bootstrap));
	if (bootstrap == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return 0;
	}

	WBXMLTreeNode *pCursor = wbxml_dom_node->children;

	while (pCursor != NULL) {
		if (pCursor->type == WBXML_TREE_ELEMENT_NODE) {
			const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor->name);
			if (!strcmp((const char *)tag_name, "parm")) {
				WBXMLList *param_attrs = pCursor->attrs;
				WB_ULONG param_index = 0;
				WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
				while (param_index < param_nb_attrs) {
					WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
					if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
						if (WBXML_STRCMP("NAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(bootstrap->name));
							if(bootstrap->name != NULL) {
								_DEBUG_VERBOSE("NAME : %s", bootstrap->name);
							}
						} else if (WBXML_STRCMP("NETWORK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(bootstrap->network);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("bootstrap network get null");
								param_index++;
								continue;
							}
							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("NETWORK : %s", target_ptr->data);
							}
							if (bootstrap->network == NULL) {
								bootstrap->network = target_ptr;
							} else {
								_DEBUG_VERBOSE("free network");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("COUNTRY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(bootstrap->country));
							if(bootstrap->country != NULL) {
								_DEBUG_VERBOSE("COUNTRY : %s", bootstrap->country);
							}
						} else if (WBXML_STRCMP("PROXY-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(bootstrap->proxy_id);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("boot strap PROXY-ID get null");
								param_index++;
								continue;
							}
							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("PROXY-ID : %s", target_ptr->data);
							}
							if (bootstrap->proxy_id == NULL) {
								bootstrap->proxy_id = target_ptr;
							} else {
								_DEBUG_VERBOSE("free proxy id");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("PROVURL", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(bootstrap->provurl));
							if(bootstrap->provurl != NULL) {
								_DEBUG_VERBOSE("PROVURL : %s", bootstrap->provurl);
							}
						} else if (WBXML_STRCMP("CONTEXT-ALLOW", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(bootstrap->context_allow));
							if(bootstrap->context_allow != NULL) {
								_DEBUG_VERBOSE("CONTEXT-ALLOW : %s", bootstrap->context_allow);
							}
						}
					}
					param_index++;
				}
			}
		}
		pCursor = pCursor->next;
	}

	_INNER_FUNC_EXIT;
	return bootstrap;
}

static DM_CP_Napdef *__parse_Napdef(WBXMLTreeNode * wbxml_dom_node)
{
	_INNER_FUNC_ENTER;

	retvm_if((wbxml_dom_node) == NULL, NULL, "wbxml_dom_node is NULL!!");

	DM_CP_Napdef *napdef = (DM_CP_Napdef *) calloc(1, sizeof(DM_CP_Napdef));
	if (napdef == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return 0;
	}
	WBXMLTreeNode *pCursor = wbxml_dom_node->children;

	while (pCursor != 0) {
		if (pCursor->type == WBXML_TREE_ELEMENT_NODE) {
			const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor->name);
			if (!strcmp((const char *)tag_name, "parm")) {
				WBXMLList *param_attrs = pCursor->attrs;
				WB_ULONG param_index = 0;
				WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
				while (param_index < param_nb_attrs) {
					WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
					if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
						if (WBXML_STRCMP("NAPID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->nap_ID));
							if(napdef->nap_ID != NULL) {
								_DEBUG_VERBOSE("NAPID : %s", napdef->nap_ID);
							}
						} else if (WBXML_STRCMP("BEARER", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(napdef->bearer);
							if(target_ptr == NULL) {
								_DEBUG_VERBOSE("___get_Last_String_Node fail");
								free_cp_napdef(napdef);
								napdef = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("BEARER : %s", target_ptr->data);
							}
							if (napdef->bearer == NULL) {
								napdef->bearer = target_ptr;
							} else {
								_DEBUG_VERBOSE("free BEARER ");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("NAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->name));
							if(napdef->name != NULL) {
								_DEBUG_VERBOSE("NAME : %s", napdef->name);
							}
						} else if (WBXML_STRCMP("INTERNET", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->internet));
							if(napdef->internet != NULL) {
								_DEBUG_VERBOSE("INTERNET : %s", napdef->internet);
							}
						} else if (WBXML_STRCMP("NAP-ADDRESS", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->nap_address));
							if(napdef->nap_address != NULL) {
								_DEBUG_VERBOSE("NAP-ADDRESS : %s", napdef->nap_address);
							}
						} else if (WBXML_STRCMP("NAP-ADDRTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->nap_addr_type));
							if(napdef->nap_addr_type != NULL) {
								_DEBUG_VERBOSE("NAP-ADDRTYPE : %s", napdef->nap_addr_type);
							}
						} else if (WBXML_STRCMP("DNS-ADDR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(napdef->dns_addr);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("DNS-ADDR is NULL");
								param_index++;
								continue;
							}

							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("DNS-ADDR : %s", target_ptr->data);
							}
							if (napdef->dns_addr == NULL) {
								napdef->dns_addr = target_ptr;
							} else {
								_DEBUG_VERBOSE("free DNS-ADDR ");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("CALLTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->call_type));
							if(napdef->call_type != NULL) {
								_DEBUG_VERBOSE("CALLTYPE : %s", napdef->call_type);
							}
						} else if (WBXML_STRCMP("LOCAL-ADDR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->local_addr));
							if(napdef->local_addr != NULL) {
								_DEBUG_VERBOSE("LOCAL-ADDR : %s", napdef->local_addr);
							}
						} else if (WBXML_STRCMP("LOCAL-ADDRTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->local_addr_type));
							if(napdef->local_addr_type != NULL) {
								_DEBUG_VERBOSE("LOCAL-ADDRTYPE : %s", napdef->local_addr_type);
							}
						} else if (WBXML_STRCMP("LINKSPEED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->link_speed));
							if(napdef->link_speed != NULL) {
								_DEBUG_VERBOSE("LINKSPEED : %s", napdef->link_speed);
							}
						} else if (WBXML_STRCMP("DNLINKSPEED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->dn_link_speed));
							if(napdef->dn_link_speed != NULL) {
								_DEBUG_VERBOSE("DNLINKSPEED : %s", napdef->dn_link_speed);
							}
						} else if (WBXML_STRCMP("LINGER", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->linger));
							if(napdef->linger != NULL) {
								_DEBUG_VERBOSE("LINGER : %s", napdef->linger);
							}
						} else if (WBXML_STRCMP("DELIVERY-ERR-SDU", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->delivery_err_sdu));
							if(napdef->delivery_err_sdu != NULL) {
								_DEBUG_VERBOSE("DELIVERY-ERR-SDU : %s", napdef->delivery_err_sdu);
							}
						} else if (WBXML_STRCMP("DELIVERY-ORDER", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->delivery_order));
							if(napdef->delivery_order != NULL) {
								_DEBUG_VERBOSE("DELIVERY-ORDER : %s", napdef->delivery_order);
							}
						} else if (WBXML_STRCMP("TRAFFIC-CLASS", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->traffic_class));
							if(napdef->traffic_class != NULL) {
								_DEBUG_VERBOSE("TRAFFIC-CLASS : %s", napdef->traffic_class);
							}
						} else if (WBXML_STRCMP("MAX-SDU-SIZE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->max_sdu_size));
							if(napdef->max_sdu_size != NULL) {
								_DEBUG_VERBOSE("MAX-SDU-SIZE : %s", napdef->max_sdu_size);
							}
						} else if (WBXML_STRCMP("MAX-BITRATE-UPLINK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->max_bitrate_uplink));
							if(napdef->max_bitrate_uplink != NULL) {
								_DEBUG_VERBOSE("MAX-BITRATE-UPLINK : %s", napdef->max_bitrate_uplink);
							}
						} else if (WBXML_STRCMP("MAX-BITRATE-DNLINK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->max_bitrate_dnlink));
							if(napdef->max_bitrate_dnlink != NULL) {
								_DEBUG_VERBOSE("MAX-BITRATE-DNLINK : %s", napdef->max_bitrate_dnlink);
							}
						} else if (WBXML_STRCMP("RESIDUAL-BER", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->residual_ber));
							if(napdef->residual_ber != NULL) {
								_DEBUG_VERBOSE("RESIDUAL-BER : %s", napdef->residual_ber);
							}
						} else if (WBXML_STRCMP("SDU-ERROR-RATIO", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->sdu_error_ratio));
							if(napdef->sdu_error_ratio != NULL) {
								_DEBUG_VERBOSE("SDU-ERROR-RATIO : %s", napdef->sdu_error_ratio);
							}
						} else if (WBXML_STRCMP("TRAFFIC-HANDL-PRIO", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->traffic_handl_prio));
							if(napdef->traffic_handl_prio != NULL) {
								_DEBUG_VERBOSE("TRAFFIC-HANDL-PRIO : %s", napdef->traffic_handl_prio);
							}
						} else if (WBXML_STRCMP("TRANSFER-DELAY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->transfer_delay));
							if(napdef->transfer_delay != NULL) {
								_DEBUG_VERBOSE("TRANSFER-DELAY : %s", napdef->transfer_delay);
							}
						} else if (WBXML_STRCMP("GUARANTEED-BITRATE-UPLINK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->guaranteed_bitrate_uplink));
							if(napdef->guaranteed_bitrate_uplink != NULL) {
								_DEBUG_VERBOSE("GUARANTEED-BITRATE-UPLINK : %s", napdef->guaranteed_bitrate_uplink);
							}
						} else if (WBXML_STRCMP("GUARANTEED-BITRATE-DNLINK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->guaranteed_bitrate_dnlink));
							if(napdef->guaranteed_bitrate_dnlink != NULL) {
								_DEBUG_VERBOSE("GUARANTEED-BITRATE-DNLINK : %s", napdef->guaranteed_bitrate_dnlink);
							}
						} else if (WBXML_STRCMP("MAX-NUM-RETRY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->max_num_retry));
							if(napdef->max_num_retry != NULL) {
								_DEBUG_VERBOSE("MAX-NUM-RETRY : %s", napdef->max_num_retry);
							}
						} else if (WBXML_STRCMP("FIRST-RETRY-TIMEOUT", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->first_retry_timeout));
							if(napdef->first_retry_timeout != NULL) {
								_DEBUG_VERBOSE("FIRST-RETRY-TIMEOUT : %s", napdef->first_retry_timeout);
							}
						} else if (WBXML_STRCMP("REREG-THRESHOLD", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->rereg_threshold));
							if(napdef->rereg_threshold != NULL) {
								_DEBUG_VERBOSE("REREG-THRESHOLD : %s", napdef->rereg_threshold);
							}
						} else if (WBXML_STRCMP("T-BIT", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(napdef->t_bit));
							if(napdef->t_bit != NULL) {
								_DEBUG_VERBOSE("T-BIT : %s", napdef->t_bit);
							}
						}
					}
					param_index++;
				}
			} else if (!strcmp((const char *)tag_name, "characteristic")) {
				WBXMLList *attrs = pCursor->attrs;
				WB_ULONG index = 0;
				WB_ULONG nb_attrs = wbxml_list_len(attrs);

				while (index < nb_attrs) {
					WBXMLAttribute *attr = (WBXMLAttribute *) wbxml_list_get(attrs, index);
					if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(attr)) == 0) {
						if (WBXML_STRCMP("NAPAUTHINFO", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Napdef_AuthInfo *auth_info = (DM_CP_Napdef_AuthInfo *) calloc(1, sizeof(DM_CP_Napdef_AuthInfo));
							if (auth_info == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_napdef(napdef);
								napdef = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (napdef->auth_info == NULL) {
								napdef->auth_info = auth_info;
							} else {
								DM_CP_Napdef_AuthInfo *cursor_ptr = napdef->auth_info;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = auth_info;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_AuthInfo = pCursor->children;
							while (pCursor_AuthInfo != 0) {
								if (pCursor_AuthInfo->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_AuthInfo->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_AuthInfo->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("AUTHTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AuthInfo, &(auth_info->auth_type));
													if(auth_info->auth_type != NULL) {
														_DEBUG_VERBOSE("AUTHTYPE : %s", auth_info->auth_type);
													}
												} else if (WBXML_STRCMP("AUTHNAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AuthInfo, &(auth_info->auth_name));
													if(auth_info->auth_name != NULL) {
														_DEBUG_VERBOSE("AUTHNAME : %s", auth_info->auth_name);
													}
												} else if (WBXML_STRCMP("AUTHSECRET", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AuthInfo, &(auth_info->auth_secret));
													if(auth_info->auth_secret != NULL) {
														_DEBUG_VERBOSE("AUTHSECRET : %s", auth_info->auth_secret);
													}
												} else if (WBXML_STRCMP("AUTH-ENTITY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													String_Node *target_ptr = ___get_Last_String_Node(auth_info->auth_entity);
													if (target_ptr == NULL) {
														_DEBUG_VERBOSE("auth entity is NULL");
														param_index++;
														continue;
													}
													___set_Param(pCursor_AuthInfo, &(target_ptr->data));
													if(target_ptr->data != NULL) {
														_DEBUG_VERBOSE("AUTH-ENTITY : %s", target_ptr->data);
													}
													if (auth_info->auth_entity == NULL) {
														auth_info->auth_entity = target_ptr;
													} else {
														_DEBUG_VERBOSE("free auth entity");
														free_string_node(target_ptr);
													}
												} else if (WBXML_STRCMP("SPI", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AuthInfo, &(auth_info->spi));
													if(auth_info->spi != NULL ) {
														_DEBUG_VERBOSE("SPI : %s", auth_info->spi);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_AuthInfo = pCursor_AuthInfo->next;
							}
						} else if (WBXML_STRCMP("VALIDITY", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Napdef_validity *validity = (DM_CP_Napdef_validity *) calloc(1, sizeof(DM_CP_Napdef_validity));
							if (validity == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_napdef(napdef);
								napdef = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (napdef->validity == NULL) {
								napdef->validity = validity;
							} else {
								DM_CP_Napdef_validity *cursor_ptr = napdef->validity;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = validity;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_Validity = pCursor->children;
							while (pCursor_Validity != 0) {
								if (pCursor_Validity->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_Validity->name);
									if (!strcmp((const char *)tag_name, "parm")) {
										WBXMLList *param_attrs = pCursor_Validity->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("COUNTRY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Validity, &(validity->country));
													if(validity->country != NULL) {
														_DEBUG_VERBOSE("COUNTRY : %s", validity->country);
													}
												} else if (WBXML_STRCMP("NETWORK", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Validity, &(validity->network));
													if(validity->network != NULL) {
														_DEBUG_VERBOSE("NETWORK : %s", validity->network);
													}
												} else if (WBXML_STRCMP("SID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Validity, &(validity->sid));
													if(validity->sid != NULL) {
														_DEBUG_VERBOSE("SID : %s", validity->sid);
													}
												} else if (WBXML_STRCMP("SOC", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Validity, &(validity->soc));
													if(validity->soc != NULL) {
														_DEBUG_VERBOSE("SOC : %s", validity->soc);
													}
												} else if (WBXML_STRCMP("VALIDUNTIL", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Validity, &(validity->validuntil));
													if(validity->validuntil != NULL) {
														_DEBUG_VERBOSE("VALIDUNTIL : %s", validity->validuntil);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_Validity = pCursor_Validity->next;
							}
						}
					}
					index++;
				}
			}
		}
		pCursor = pCursor->next;
	}

	_INNER_FUNC_EXIT;

	return napdef;
}

static DM_CP_Application *__parse_Application(WBXMLTreeNode * wbxml_dom_node)
{
	_INNER_FUNC_ENTER;

	retvm_if((wbxml_dom_node) == NULL, NULL, "wbxml_dom_node is NULL!!");

	DM_CP_Application *application = (DM_CP_Application *) calloc(1, sizeof(DM_CP_Application));
	if (application == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return 0;
	}
	WBXMLTreeNode *pCursor = wbxml_dom_node->children;

	while (pCursor != 0) {
		if (pCursor->type == WBXML_TREE_ELEMENT_NODE) {
			const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor->name);
			if (strcmp((const char *)tag_name, "parm") == 0) {
				WBXMLList *param_attrs = pCursor->attrs;
				WB_ULONG param_index = 0;
				WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
				while (param_index < param_nb_attrs) {
					WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
					if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
						if (WBXML_STRCMP("APPID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(application->app_ID));
							if(application->app_ID != NULL) {
								_DEBUG_VERBOSE("APPID : %s", application->app_ID);
							}
						} else if (WBXML_STRCMP("PROVIDER-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(application->provider_ID));
							if(application->provider_ID != NULL) {
								_DEBUG_VERBOSE("PROVIDER-ID : %s", application->provider_ID);
							}
						} else if (WBXML_STRCMP("NAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(application->name));
							if(application->name != NULL) {
								_DEBUG_VERBOSE("NAME : %s", application->name);
							}
						} else if (WBXML_STRCMP("AACCEPT", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(application->accept));
							if(application->accept != NULL) {
								_DEBUG_VERBOSE("AACCEPT : %s", application->accept);
							}
						} else if (WBXML_STRCMP("APROTOCOL", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(application->protocol));
							if(application->protocol != NULL) {
								_DEBUG_VERBOSE("APROTOCOL : %s", application->protocol);
							}
						} else if (WBXML_STRCMP("TO-PROXY", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(application->to_proxy);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("TO-PROXY is NULL");
								param_index++;
								continue;
							}

							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("PROXY : %s", target_ptr->data);
							}

							if (application->to_proxy == NULL) {
								application->to_proxy = target_ptr;
							} else {
								_DEBUG_VERBOSE("free PROXY ");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("TO-NAPID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(application->to_napid);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("TO-NAPID is NULL");
								param_index++;
								continue;
							}

							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("NAPID : %s", target_ptr->data);
							}
							if (application->to_napid == NULL) {
								application->to_napid = target_ptr;
							} else {
								_DEBUG_VERBOSE("free NAPID ");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("ADDR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(application->addr);
							if (target_ptr == NULL) {
								_DEBUG_VERBOSE("ADDR is NULL");
								param_index++;
								continue;
							}

							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("ADDR : %s", target_ptr->data);
							}
							if (application->addr == NULL) {
								application->addr = target_ptr;
							} else {
								_DEBUG_VERBOSE("free ADDR ");
								free_string_node(target_ptr);
							}
						}
						break;
					}
					param_index++;
				}
			} else if (!strcmp((const char *)tag_name, "characteristic")) {
				WBXMLList *attrs = pCursor->attrs;
				WB_ULONG index = 0;
				WB_ULONG nb_attrs = wbxml_list_len(attrs);

				while (index < nb_attrs) {
					WBXMLAttribute *attr = (WBXMLAttribute *) wbxml_list_get(attrs, index);

					if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(attr)) == 0) {

						if (WBXML_STRCMP("APPADDR", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Application_AppAddr *app_addr = (DM_CP_Application_AppAddr *) calloc(1, sizeof(DM_CP_Application_AppAddr));
							if (app_addr == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_application(application);
								application = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (application->appAddr == NULL) {
								application->appAddr = app_addr;
							} else {
								DM_CP_Application_AppAddr *cursor_ptr = application->appAddr;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = app_addr;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_AppAddr = pCursor->children;
							while (pCursor_AppAddr != 0) {
								if (pCursor_AppAddr->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_AppAddr->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_AppAddr->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("ADDR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAddr, &(app_addr->addr));
													if(app_addr->addr != NULL) {
														_DEBUG_VERBOSE("ADDR : %s", app_addr->addr);
													}
												} else if (WBXML_STRCMP("ADDRTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAddr, &(app_addr->addr_type));
													if(app_addr->addr_type != NULL) {
														_DEBUG_VERBOSE("ADDRTYPE : %s", app_addr->addr_type);
													}
												}
												break;
											}
											param_index++;
										}
									} else if (strcmp((const char *)tag_name, "characteristic") == 0) {
										WBXMLList *addr_attrs = pCursor_AppAddr->attrs;
										WB_ULONG addr_index = 0;
										WB_ULONG addr_nb_attrs = wbxml_list_len(addr_attrs);

										while (addr_index < addr_nb_attrs) {
											WBXMLAttribute *addr_attr = (WBXMLAttribute *) wbxml_list_get(addr_attrs, addr_index);
											if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(addr_attr)) == 0) {
												if (WBXML_STRCMP("PORT", wbxml_attribute_get_xml_value(addr_attr)) == 0) {
													DM_CP_Port *port = (DM_CP_Port *) calloc(1, sizeof(DM_CP_Port));
													if (port == NULL) {
														_DEBUG_VERBOSE("alloc fail");
														free_cp_application(application);
														application = NULL;
														_INNER_FUNC_EXIT;
														return 0;
													}
													if (app_addr->port == NULL) {
														app_addr->port = port;
													} else {
														DM_CP_Port *cursor_ptr = app_addr->port;
														while (cursor_ptr != NULL) {
															if (cursor_ptr->next == NULL) {
																cursor_ptr->next = port;
																break;
															}
															cursor_ptr = cursor_ptr->next;
														}
													}
													WBXMLTreeNode *pCursor_Port = pCursor_AppAddr->children;
													while (pCursor_Port != 0) {
														if (pCursor_Port->type == WBXML_TREE_ELEMENT_NODE) {
															const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_Port->name);
															if (strcmp((const char *)tag_name, "parm") == 0) {
																WBXMLList *_param_attrs = pCursor_Port->attrs;
																WB_ULONG _param_index = 0;
																WB_ULONG _param_nb_attrs = wbxml_list_len(_param_attrs);
																while (_param_index < _param_nb_attrs) {
																	WBXMLAttribute *_param_attr = (WBXMLAttribute *) wbxml_list_get(_param_attrs, _param_index);
																	if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(_param_attr)) == 0) {
																		if (WBXML_STRCMP("PORTNBR", wbxml_attribute_get_xml_value(_param_attr)) == 0) {
																			___set_Param(pCursor_Port, &(port->portnbr));
																			if(port->portnbr != NULL) {
																				_DEBUG_VERBOSE("[%d] PORTNBR : %s", _param_index, port->portnbr);
																			}
																		} else if (WBXML_STRCMP("SERVICE", wbxml_attribute_get_xml_value(_param_attr)) == 0) {
																			String_Node *target_ptr = ___get_Last_String_Node(port->service);
																			if (target_ptr == NULL) {
																				_DEBUG_VERBOSE("SERVICE is NULL");
																				_param_index++;
																				continue;
																			}

																			___set_Param(pCursor_Port, &(target_ptr->data));
																			if(target_ptr->data != NULL) {
																				_DEBUG_VERBOSE("SERVICE : %s", target_ptr->data);
																			}
																			if (port->service == NULL) {
																				port->service = target_ptr;
																			} else {
																				_DEBUG_VERBOSE("free port service");
																				free_string_node(target_ptr);
																			}
																		}
																		break;
																	}
																	_param_index++;
																	_DEBUG_VERBOSE("param_index : %d", _param_index);
																}
															}
														}
														pCursor_Port = pCursor_Port->next;
													}
												}
											}
											addr_index++;
										}
									}
								}

								pCursor_AppAddr = pCursor_AppAddr->next;
							}
						} else if (WBXML_STRCMP("APPAUTH", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Application_AppAuth *app_auth = (DM_CP_Application_AppAuth *) calloc(1, sizeof(DM_CP_Application_AppAuth));
							if (app_auth == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_application(application);
								application = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (application->appAuth == NULL) {
								application->appAuth = app_auth;
							} else {
								DM_CP_Application_AppAuth *cursor_ptr = application->appAuth;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = app_auth;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_AppAuth = pCursor->children;
							while (pCursor_AppAuth != 0) {
								if (pCursor_AppAuth->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_AppAuth->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_AppAuth->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("AAUTHLEVEL", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAuth, &(app_auth->auth_level));
													if(app_auth->auth_level != NULL) {
														_DEBUG_VERBOSE("AAUTHLEVEL : %s", app_auth->auth_level);
													}
												} else if (WBXML_STRCMP("AAUTHTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAuth, &(app_auth->auth_type));
													if(app_auth->auth_type != NULL) {
														_DEBUG_VERBOSE("AAUTHTYPE : %s", app_auth->auth_type);
													}
												} else if (WBXML_STRCMP("AAUTHNAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAuth, &(app_auth->auth_name));
													if(app_auth->auth_name != NULL) {
														_DEBUG_VERBOSE("AAUTHNAME : %s", app_auth->auth_name);
													}
												} else if (WBXML_STRCMP("AAUTHSECRET", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAuth, &(app_auth->auth_secret));
													if(app_auth->auth_secret != NULL) {
														_DEBUG_VERBOSE("AAUTHSECRET : %s", app_auth->auth_secret);
													}
												} else if (WBXML_STRCMP("AAUTHDATA", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_AppAuth, &(app_auth->auth_data));
													if(app_auth->auth_data != NULL) {
														_DEBUG_VERBOSE("AAUTHDATA : %s", app_auth->auth_data);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_AppAuth = pCursor_AppAuth->next;
							}
						} else if (WBXML_STRCMP("RESOURCE", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Application_Resource *app_resource = (DM_CP_Application_Resource *) calloc(1, sizeof(DM_CP_Application_Resource));
							if (app_resource == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_application(application);
								application = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (application->resource == NULL) {
								application->resource = app_resource;
							} else {
								DM_CP_Application_Resource *cursor_ptr = application->resource;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = app_resource;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_Resource = pCursor->children;
							while (pCursor_Resource != 0) {
								if (pCursor_Resource->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_Resource->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_Resource->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("URI", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->uri));
													if(app_resource->uri != NULL) {
														_DEBUG_VERBOSE("URI : %s", app_resource->uri);
													}
												} else if (WBXML_STRCMP("NAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->name));
													if(app_resource->name != NULL) {
														_DEBUG_VERBOSE("NAME : %s", app_resource->name);
													}
												} else if (WBXML_STRCMP("AACCEPT", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->accept));
													if(app_resource->accept != NULL) {
														_DEBUG_VERBOSE("AACCEPT : %s", app_resource->accept);
													}
												} else if (WBXML_STRCMP("AAUTHTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->auth_type));
													if(app_resource->auth_type != NULL) {
														_DEBUG_VERBOSE("AAUTHTYPE : %s", app_resource->auth_type);
													}
												} else if (WBXML_STRCMP("AAUTHNAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->auth_name));
													if(app_resource->auth_name != NULL) {
														_DEBUG_VERBOSE("AAUTHNAME : %s", app_resource->auth_name);
													}
												} else if (WBXML_STRCMP("AAUTHSECRET", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->auth_secret));
													if(app_resource->auth_secret != NULL) {
														_DEBUG_VERBOSE("AAUTHSECRET : %s", app_resource->auth_secret);
													}
												} else if (WBXML_STRCMP("AAUTHDATA", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->auth_data));
													if(app_resource->auth_data != NULL) {
														_DEBUG_VERBOSE("AAUTHDATA : %s", app_resource->auth_data);
													}
												} else if (WBXML_STRCMP("STARTPAGE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Resource, &(app_resource->start_page));
													if(app_resource->start_page != NULL) {
														_DEBUG_VERBOSE("STARTPAGE : %s", app_resource->start_page);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_Resource = pCursor_Resource->next;
							}
						}
					}

					index++;
				}
			}
		}

		pCursor = pCursor->next;
	}

	_INNER_FUNC_EXIT;

	return application;
}

static DM_CP_PxLogical *__parse_PxLogical(WBXMLTreeNode * wbxml_dom_node)
{
	_INNER_FUNC_ENTER;

	retvm_if((wbxml_dom_node) == NULL, NULL, "wbxml_dom_node is NULL!!");

	DM_CP_PxLogical *pxLogical = (DM_CP_PxLogical *) calloc(1, sizeof(DM_CP_PxLogical));
	if (pxLogical == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return 0;
	}
	WBXMLTreeNode *pCursor = wbxml_dom_node->children;

	while (pCursor != 0) {
		if (pCursor->type == WBXML_TREE_ELEMENT_NODE) {
			const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor->name);
			if (strcmp((const char *)tag_name, "parm") == 0) {
				WBXMLList *param_attrs = pCursor->attrs;
				WB_ULONG param_index = 0;
				WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
				while (param_index < param_nb_attrs) {
					WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
					if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
						if (WBXML_STRCMP("PROXY-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->proxy_id));
							if(pxLogical->proxy_id != NULL) {
								_DEBUG_VERBOSE("PROXY-ID : %s", pxLogical->proxy_id);
							}
						} else if (WBXML_STRCMP("PROXY-PW", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->proxy_pw));
							if(pxLogical->proxy_pw != NULL) {
								_DEBUG_VERBOSE("PROXY-PW : %s", pxLogical->proxy_pw);
							}
						} else if (WBXML_STRCMP("PPGAUTH-TYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->ppg_auth_type));
							if(pxLogical->ppg_auth_type != NULL) {
								_DEBUG_VERBOSE("PPGAUTH-TYPE : %s", pxLogical->ppg_auth_type);
							}
						} else if (WBXML_STRCMP("PROXY-PROVIDER-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->proxy_provider_id));
							if(pxLogical->proxy_provider_id != NULL) {
								_DEBUG_VERBOSE("PROXY-PROVIDER-ID : %s", pxLogical->proxy_provider_id);
							}
						} else if (WBXML_STRCMP("NAME", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->name));
							if(pxLogical->name != NULL) {
								_DEBUG_VERBOSE("NAME : %s", pxLogical->name);
							}
						} else if (WBXML_STRCMP("DOMAIN", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							String_Node *target_ptr = ___get_Last_String_Node(pxLogical->domain);
							if(target_ptr == NULL) {
								_DEBUG_VERBOSE("___get_Last_String_Node fail");
								free_cp_pxlogical(pxLogical);
								pxLogical = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							___set_Param(pCursor, &(target_ptr->data));
							if(target_ptr->data != NULL) {
								_DEBUG_VERBOSE("DOMAIN : %s", target_ptr->data);
							}
							if (pxLogical->domain == 0) {
								pxLogical->domain = target_ptr;
							} else {
								_DEBUG_VERBOSE("free DOMAIN ");
								free_string_node(target_ptr);
							}
						} else if (WBXML_STRCMP("TRUST", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->trust));
							if(pxLogical->trust != NULL) {
								_DEBUG_VERBOSE("TRUST : %s", pxLogical->trust);
							}
						} else if (WBXML_STRCMP("MASTER", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->master));
							if(pxLogical->master != NULL) {
								_DEBUG_VERBOSE("MASTER : %s", pxLogical->master);
							}
						} else if (WBXML_STRCMP("STARTPAGE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->start_page));
							if(pxLogical->start_page != NULL) {
								_DEBUG_VERBOSE("STARTPAGE : %s", pxLogical->start_page);
							}
						} else if (WBXML_STRCMP("BASAUTH-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->bas_auth_id));
							if(pxLogical->bas_auth_id != NULL) {
								_DEBUG_VERBOSE("BASAUTH-ID : %s", pxLogical->bas_auth_id);
							}
						} else if (WBXML_STRCMP("BASAUTH-PW", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->bas_auth_pw));
							if(pxLogical->bas_auth_pw != NULL) {
								_DEBUG_VERBOSE("BASAUTH-PW : %s", pxLogical->bas_auth_pw);
							}
						} else if (WBXML_STRCMP("WSP-VERSION", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->wsp_version));
							if(pxLogical->wsp_version != NULL) {
								_DEBUG_VERBOSE("WSP-VERSION : %s", pxLogical->wsp_version);
							}
						} else if (WBXML_STRCMP("PUSHENABLED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->push_enabled));
							if(pxLogical->push_enabled != NULL) {
								_DEBUG_VERBOSE("PUSHENABLED : %s", pxLogical->push_enabled);
							}
						} else if (WBXML_STRCMP("PULLENABLED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
							___set_Param(pCursor, &(pxLogical->pull_enabled));
							if(pxLogical->pull_enabled != NULL) {
								_DEBUG_VERBOSE("PULLENABLED : %s", pxLogical->pull_enabled);
							}
						}
						break;
					}
					param_index++;
				}
			} else if (!strcmp((const char *)tag_name, "characteristic")) {
				WBXMLList *attrs = pCursor->attrs;
				WB_ULONG index = 0;
				WB_ULONG nb_attrs = wbxml_list_len(attrs);

				while (index < nb_attrs) {
					WBXMLAttribute *attr = (WBXMLAttribute *) wbxml_list_get(attrs, index);

					if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(attr)) == 0) {

						if (WBXML_STRCMP("PXAUTHINFO", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_PxLogical_PxAuthInfo *pxAuthInfo = (DM_CP_PxLogical_PxAuthInfo *) calloc(1, sizeof(DM_CP_PxLogical_PxAuthInfo));
							if (pxAuthInfo == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_pxlogical(pxLogical);
								pxLogical = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (pxLogical->pxAuthInfo == NULL) {
								pxLogical->pxAuthInfo = pxAuthInfo;
							} else {
								DM_CP_PxLogical_PxAuthInfo *cursor_ptr = pxLogical->pxAuthInfo;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = pxAuthInfo;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_PxAuthInfo = pCursor->children;
							while (pCursor_PxAuthInfo != 0) {
								if (pCursor_PxAuthInfo->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_PxAuthInfo->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_PxAuthInfo->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("PXAUTH-TYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxAuthInfo, &(pxAuthInfo->pxAuth_type));
													if(pxAuthInfo->pxAuth_type != NULL) {
														_DEBUG_VERBOSE("PXAUTH-TYPE : %s", pxAuthInfo->pxAuth_type);
													}
												} else if (WBXML_STRCMP("PXAUTH-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxAuthInfo, &(pxAuthInfo->pxAuth_id));
													if(pxAuthInfo->pxAuth_id != NULL) {
														_DEBUG_VERBOSE("PXAUTH-ID : %s", pxAuthInfo->pxAuth_id);
													}
												} else if (WBXML_STRCMP("PXAUTH-PW", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxAuthInfo, &(pxAuthInfo->pxAuth_pw));
													if(pxAuthInfo->pxAuth_pw != NULL) {
														_DEBUG_VERBOSE("PXAUTH-PW : %s", pxAuthInfo->pxAuth_pw);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_PxAuthInfo = pCursor_PxAuthInfo->next;
							}
						} else if (WBXML_STRCMP("PORT", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_Port *port = (DM_CP_Port *) calloc(1, sizeof(DM_CP_Port));
							if (port == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_pxlogical(pxLogical);
								pxLogical = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (pxLogical->port == NULL) {
								pxLogical->port = port;
							} else {
								DM_CP_Port *cursor_ptr = pxLogical->port;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = port;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_Port = pCursor->children;
							while (pCursor_Port != 0) {
								if (pCursor_Port->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_Port->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_Port->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("PORTNBR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_Port, &(port->portnbr));
													if(port->portnbr != NULL) {
														_DEBUG_VERBOSE("AAUTHLEVEL : %s", port->portnbr);
													}
												} else if (WBXML_STRCMP("SERVICE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													String_Node *target_ptr = ___get_Last_String_Node(port->service);

													if(target_ptr == NULL) {
														_DEBUG_ERROR("___get_Last_String_Node fail - target_ptr NULL ERROR");
														free_cp_pxlogical(pxLogical);
														pxLogical = NULL;
														_INNER_FUNC_EXIT;
														return 0;
													}

													___set_Param(pCursor_Port, &(target_ptr->data));
													if(target_ptr->data != NULL) {
														_DEBUG_VERBOSE("SERVICE : %s", target_ptr->data);
													}
													if (port->service == NULL) {
														port->service = target_ptr;
													} else {
														_DEBUG_VERBOSE("free SERVICE ");
														free_string_node(target_ptr);
													}
												}
												break;
											}
											param_index++;
										}
									}
								}
								pCursor_Port = pCursor_Port->next;
							}
						} else if (WBXML_STRCMP("PXPHYSICAL", wbxml_attribute_get_xml_value(attr)) == 0) {
							DM_CP_PxLogical_PxPhysical *pxPhysical = (DM_CP_PxLogical_PxPhysical *) calloc(1, sizeof(DM_CP_PxLogical_PxPhysical));
							if (pxPhysical == NULL) {
								_DEBUG_VERBOSE("alloc fail");
								free_cp_pxlogical(pxLogical);
								pxLogical = NULL;
								_INNER_FUNC_EXIT;
								return 0;
							}
							if (pxLogical->pxPhysical == NULL) {
								pxLogical->pxPhysical = pxPhysical;
							} else {
								DM_CP_PxLogical_PxPhysical *cursor_ptr = pxLogical->pxPhysical;
								while (cursor_ptr != NULL) {
									if (cursor_ptr->next == NULL) {
										cursor_ptr->next = pxPhysical;
										break;
									}
									cursor_ptr = cursor_ptr->next;
								}
							}
							WBXMLTreeNode *pCursor_PxPhysical = pCursor->children;
							while (pCursor_PxPhysical != 0) {
								if (pCursor_PxPhysical->type == WBXML_TREE_ELEMENT_NODE) {
									const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_PxPhysical->name);
									if (strcmp((const char *)tag_name, "parm") == 0) {
										WBXMLList *param_attrs = pCursor_PxPhysical->attrs;
										WB_ULONG param_index = 0;
										WB_ULONG param_nb_attrs = wbxml_list_len(param_attrs);
										while (param_index < param_nb_attrs) {
											WBXMLAttribute *param_attr = (WBXMLAttribute *) wbxml_list_get(param_attrs, param_index);
											if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(param_attr)) == 0) {
												if (WBXML_STRCMP("PHYSICAL-PROXY-ID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->physical_proxy_id));
													if(pxPhysical->physical_proxy_id != NULL) {
														_DEBUG_VERBOSE("PHYSICAL-PROXY-ID : %s", pxPhysical->physical_proxy_id);
													}
												} else if (WBXML_STRCMP("DOMAIN", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													String_Node *target_ptr = ___get_Last_String_Node(pxPhysical->domain);
													if(target_ptr == NULL) {
														_DEBUG_VERBOSE("___get_Last_String_Node fail");
														free_cp_pxlogical(pxLogical);
														pxLogical = NULL;
														_INNER_FUNC_EXIT;
														return 0;
													}
													___set_Param(pCursor_PxPhysical, &(target_ptr->data));
													if(target_ptr->data != NULL) {
														_DEBUG_VERBOSE("DOMAIN : %s", target_ptr->data);
													}
													if (pxPhysical->domain == NULL) {
														pxPhysical->domain = target_ptr;
													} else {
														_DEBUG_VERBOSE("free DOMAIN ");
														free_string_node(target_ptr);
													}
												} else if (WBXML_STRCMP("PXADDR", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->pxAddr));
													if(pxPhysical->pxAddr != NULL) {
														_DEBUG_VERBOSE("PXADDR : %s", pxPhysical->pxAddr);
													}
												} else if (WBXML_STRCMP("PXADDRTYPE", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->pxAddrType));
													if(pxPhysical->pxAddrType != NULL) {
														_DEBUG_VERBOSE("PXADDRTYPE : %s", pxPhysical->pxAddrType);
													}
												} else if (WBXML_STRCMP("PXADDR-FQDN", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->pxAddr_FQDN));
													if(pxPhysical->pxAddr_FQDN != NULL) {
														_DEBUG_VERBOSE("PXADDR-FQDN : %s", pxPhysical->pxAddr_FQDN);
													}
												} else if (WBXML_STRCMP("WSP-VERSION", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->wsp_version));
													if(pxPhysical->wsp_version != NULL) {
														_DEBUG_VERBOSE("WSP-VERSION : %s", pxPhysical->wsp_version);
													}
												} else if (WBXML_STRCMP("PUSHENABLED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->push_enabled));
													if(pxPhysical->push_enabled != NULL) {
														_DEBUG_VERBOSE("PUSHENABLED : %s", pxPhysical->push_enabled);
													}
												} else if (WBXML_STRCMP("PULLENABLED", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													___set_Param(pCursor_PxPhysical, &(pxPhysical->pull_enabled));
													if(pxPhysical->pull_enabled != NULL) {
														_DEBUG_VERBOSE("PULLENABLED : %s", pxPhysical->pull_enabled);
													}
												} else if (WBXML_STRCMP("TO-NAPID", wbxml_attribute_get_xml_value(param_attr)) == 0) {
													String_Node *target_ptr = ___get_Last_String_Node(pxPhysical->to_napid);
													if(target_ptr == NULL) {
														_DEBUG_VERBOSE("___get_Last_String_Node fail");
														free_cp_pxlogical(pxLogical);
														pxLogical = NULL;
														_INNER_FUNC_EXIT;
														return 0;
													}
													___set_Param(pCursor_PxPhysical, &(target_ptr->data));
													if(target_ptr->data != NULL) {
														_DEBUG_VERBOSE("TO-NAPID : %s", target_ptr->data);
													}
													if (pxPhysical->to_napid == NULL) {
														pxPhysical->to_napid = target_ptr;
													} else {
														_DEBUG_VERBOSE("free TO-NAPID ");
														free_string_node(target_ptr);
													}
												}
												break;
											}
											param_index++;
										}
									} else if (strcmp((const char *)tag_name, "characteristic") == 0) {
										WBXMLList *addr_attrs = pCursor_PxPhysical->attrs;
										WB_ULONG addr_index = 0;
										WB_ULONG addr_nb_attrs = wbxml_list_len(addr_attrs);

										while (addr_index < addr_nb_attrs) {
											WBXMLAttribute *addr_attr = (WBXMLAttribute *) wbxml_list_get(addr_attrs, addr_index);
											if (WBXML_STRCMP("type", wbxml_attribute_get_xml_name(addr_attr)) == 0) {
												if (WBXML_STRCMP("PORT", wbxml_attribute_get_xml_value(addr_attr)) == 0) {
													DM_CP_Port *port = (DM_CP_Port *) calloc(1, sizeof(DM_CP_Port));
													if (port == NULL) {
														_DEBUG_VERBOSE("alloc fail");
														free_cp_pxlogical(pxLogical);
														pxLogical = NULL;
														_INNER_FUNC_EXIT;
														return 0;
													}
													if (pxPhysical->port == NULL) {
														pxPhysical->port = port;
													} else {
														DM_CP_Port *cursor_ptr = pxPhysical->port;
														while (cursor_ptr != NULL) {
															if (cursor_ptr->next == NULL) {
																cursor_ptr->next = port;
																break;
															}
															cursor_ptr = cursor_ptr->next;
														}
													}
													WBXMLTreeNode *pCursor_Port = pCursor_PxPhysical->children;
													while (pCursor_Port != 0) {
														if (pCursor_Port->type == WBXML_TREE_ELEMENT_NODE) {
															const WB_UTINY *tag_name = wbxml_tag_get_xml_name(pCursor_Port->name);
															if (strcmp((const char *)tag_name, "parm") == 0) {
																WBXMLList *_param_attrs = pCursor_Port->attrs;
																WB_ULONG _param_index = 0;
																WB_ULONG _param_nb_attrs = wbxml_list_len(_param_attrs);
																while (_param_index < _param_nb_attrs) {
																	WBXMLAttribute *_param_attr = (WBXMLAttribute *) wbxml_list_get(_param_attrs, _param_index);
																	if (WBXML_STRCMP("name", wbxml_attribute_get_xml_name(_param_attr)) == 0) {
																		if (WBXML_STRCMP("PORTNBR", wbxml_attribute_get_xml_value(_param_attr)) == 0) {
																			___set_Param(pCursor_Port, &(port->portnbr));
																			if(port->portnbr != NULL) {
																				_DEBUG_VERBOSE("[%d] PORTNBR : %s", _param_index, port->portnbr);
																			}
																		} else if (WBXML_STRCMP("SERVICE", wbxml_attribute_get_xml_value(_param_attr)) == 0) {
																			String_Node *target_ptr = ___get_Last_String_Node(port->service);
																			if(target_ptr==NULL) {
																				_DEBUG_VERBOSE("___get_Last_String_Node fail");
																				free_cp_pxlogical(pxLogical);
																				pxLogical = NULL;
																				_INNER_FUNC_EXIT;
																				return 0;
																			}
																			___set_Param(pCursor_Port, &(target_ptr->data));
																			if(target_ptr->data != NULL) {
																				_DEBUG_VERBOSE("SERVICE : %s", target_ptr->data);
																			}
																			if (port->service == NULL) {
																				port->service = target_ptr;
																			} else {
																				_DEBUG_VERBOSE("free port service");
																				free_string_node(target_ptr);
																			}
																		}
																		break;
																	}
																	_param_index++;
																	_DEBUG_VERBOSE("param_index : %d", _param_index);
																}
															}
														}
														pCursor_Port = pCursor_Port->next;
													}
												}
											}
											addr_index++;
										}
									}
								}
								pCursor_PxPhysical = pCursor_PxPhysical->next;
							}
						}
					}
					index++;
				}
			}
		}

		pCursor = pCursor->next;
	}

	_INNER_FUNC_EXIT;

	return pxLogical;
}

static void ___set_Param(WBXMLTreeNode * wbxml_dom_node, char **set_value)
{
	_INNER_FUNC_ENTER;

	retm_if((wbxml_dom_node) == NULL, "wbxml_dom_node is NULL!!");

	WBXMLList *attr_list = wbxml_dom_node->attrs;
	WB_ULONG index = 0;
	WB_ULONG nb_attrs = wbxml_list_len(attr_list);

	while (index < nb_attrs) {
		WBXMLAttribute *attr = (WBXMLAttribute *) wbxml_list_get(attr_list, index);
		if (WBXML_STRCMP("value", wbxml_attribute_get_xml_name(attr)) == 0) {
			const char *value_str = 0;
			value_str = (const char *)wbxml_attribute_get_xml_value(attr);
			int value_len = 0;
			if (value_str != NULL) {
				value_len = strlen(value_str);
			}
			int str_len = 0;
			*set_value = (char *)calloc(value_len + 1, sizeof(char));
			if (*set_value == NULL) {
				break;
			}

			str_len = g_strlcpy(*set_value, value_str, value_len + 1);
			if (str_len >= sizeof(*set_value)) {
				_DEBUG_VERBOSE("buffer overflow");
				break;
			}
			break;
		}
		index++;
	}
	_INNER_FUNC_EXIT;
}
