/*
 * oma-dm-agent
 * Copyright (c) 2012 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*lib*/
#include <stdio.h>
#include <aul.h>

#include <vconf.h>
#include <vconf-keys.h>

/*sync-agent*/
#include <sync_agent.h>

/*dm-agent*/
#include "common/dm_common.h"
#include "common/util/util.h"
#include "framework/platform-event-handler/dm_platform_event_handler_internal.h"
#include "framework/task/oma_dm_task_request.h"
#include "dm-engine/fumo/fumo_account.h"
#include "dm-engine/fumo/fumo_engine_internal.h"
#include "dm-engine/cp/dm_cp_processor.h"
#include "mo-handler/dm_mo_common.h"
#include "mo-handler/dm_mo_common_internal.h"
#include "ipc_common.h"
#include "ipc_agent.h"

#ifndef OMADM_AGENT_LOG
#undef LOG_TAG
#define LOG_TAG	"OMA_DM_IPC"
#endif

static gint64 timeout;
static GMutex mutex;
static GCond cond;

#define TIMEOUT_VAL_UIC 60	//uic time out time
#define REST_TIMEOUT_VAL 10

static gint64 uic_timeout_val = 0;

static uic_res_data *res_uic_data = NULL;

static int (*_get_event_callback(int event_id)) (sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_software_update(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_cancel(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_download(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_install(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_set_interval(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_account_registration(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_get_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_set_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_reset_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_uic_alert(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_get_fumo_config(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_set_fumo_config_wifi_only(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_callback_check_pin_code(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_callback_default(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int __event_add_dmacc(sync_agent_event_data_s * request, sync_agent_event_data_s * response);
static int __event_remove_dmacc(sync_agent_event_data_s * request, sync_agent_event_data_s * response);

static int _IPC_noti(int noti_type, int noti_id, int input_param, char *input_string, void *user_data);
static sync_agent_event_ipc_cancel_e __dm_task_process_check_cancel_flag(void *user_data);

static int __check_ui_mode(char **ui_mode, char *input_string);
static int __launch_ui(char *ui_mode, int noti_type);

static int __uic_response(int noti_id, void *user_data);

static int __check_ui_mode(char **ui_mode, char *input_string)
{
	_INNER_FUNC_ENTER;

	if (*ui_mode == NULL) {
		_DEBUG_VERBOSE("ui mode is NULL");
		if (input_string != NULL) {
			_DEBUG_VERBOSE("input_string : %s", input_string);
			if (strcmp(input_string, OMADM_ALERT_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_NOTI_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_BACKGROUND_UI_TYPE) == 0) {

				*ui_mode = strdup(input_string);
			} else {
				//do nothing
			}
		} else {
			_DEBUG_VERBOSE("ERROR UI MODE NULL");
			_INNER_FUNC_EXIT;
			return RESPONSE_TYPE_CANCEL;
		}
	} else {
		_DEBUG_VERBOSE("ui mode : %s", *ui_mode);
		_DEBUG_VERBOSE("input_string : %s", input_string);
		if (input_string != NULL) {
			if (strcmp(input_string, OMADM_ALERT_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_NOTI_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_UI_TYPE) == 0 || strcmp(input_string, OMADM_FUMO_BACKGROUND_UI_TYPE) == 0) {

				str_free(ui_mode);
				*ui_mode = strdup(input_string);
			} else {
				//do nothing
			}
		} else {
			//do nothing
		}
	}

	_INNER_FUNC_EXIT;
	return RESPONSE_TYPE_OK;
}

static int __launch_ui(char *ui_mode, int noti_type)
{
	_INNER_FUNC_ENTER;

	int ui_ret = 1;

	if (ui_mode != NULL) {
		_DEBUG_INFO("ui_mode[%s], noti_type[%d]", ui_mode, noti_type);

		if (strcmp(ui_mode, OMADM_FUMO_UI_TYPE) == 0) {
			ui_ret = launch_oma_dm_fumo_ui();
			_DEBUG_INFO("fumo ui : %d", ui_ret);
		} else if (strcmp(ui_mode, OMADM_FUMO_NOTI_UI_TYPE) == 0) {
			DM_ERROR ret_poll = DM_OK;
			char *server_id = NULL;
			ret_poll = get_Serverid(SAMSUNG_FUMO_TYPE, &server_id);
			if (ret_poll!= DM_OK || server_id == NULL) {
				ui_ret = launch_om_dm_fumo_noti_ui(NOTI_TYPE_USERINTERACTION, NULL, NULL, POLL_EVENT);
			} else {
				ui_ret = launch_om_dm_fumo_noti_ui(NOTI_TYPE_USERINTERACTION, NULL, server_id, POLL_EVENT);
				free(server_id);
			}
			_DEBUG_INFO("noti ui : %d", ui_ret);
		} else if (strcmp(ui_mode, OMADM_FUMO_BACKGROUND_UI_TYPE) == 0) {
			if (noti_type == NOTI_INSTALL) {
				ui_ret = launch_om_dm_fumo_noti_ui(NOTI_TYPE_BACKGRUOUND, NULL, NULL, 0);
				_DEBUG_INFO("noti ui : %d", ui_ret);
			}
			_DEBUG_INFO("background ui");
			_INNER_FUNC_EXIT;
			return 0;
		} else if (strcmp(ui_mode, OMADM_ALERT_UI_TYPE) == 0) {
			ui_ret = launch_om_dm_fumo_alert_ui();
			_DEBUG_INFO("alert ui : %d", ui_ret);
		} else {
			_INNER_FUNC_EXIT;
			return -1;
		}

		if (ui_ret == 0) {
			_DEBUG_INFO("UI launch fail");
			_INNER_FUNC_EXIT;
			return -1;
		}

	} else {
		_DEBUG_INFO("ERROR UI MODE NULL");
		_INNER_FUNC_EXIT;
		return -1;
	}
	_INNER_FUNC_EXIT;
	return 1;
}

static int (*_get_event_callback(int event_id)) (sync_agent_event_data_s * request, sync_agent_event_data_s * response) {
	_DEBUG_TRACE("%s: start (event_id:%d)\n", __func__, event_id);

	switch (event_id) {
	case EVENT_SOFTWARE_UPDATE:	// Type 0, Replies Exist

		return __event_callback_software_update;
		break;

	case EVENT_CANCEL:	// Type 0, No Reply

		return __event_callback_cancel;
		break;

	case EVENT_DOWNLOAD:	// Type 0, No Reply

		return __event_callback_download;
		break;

	case EVENT_INSTALL:	// Type 0, No Reply
		return __event_callback_install;
		break;

	case EVENT_SET_INTERVAL:
		return __event_callback_set_interval;
		break;

	case EVENT_ACCOUNT_REGISTRATION:
		return __event_callback_account_registration;
		break;

	case EVENT_GET_FUMO_CONFIG:
		return __event_callback_get_fumo_config;
		break;

	case EVENT_SET_FUMO_CONFIG:
		return __event_callback_set_fumo_config_wifi_only;
		break;

	case EVENT_GET_PROFILE:
		return __event_callback_get_profile;
		break;

	case EVENT_SET_PROFILE:
		return __event_callback_set_profile;
		break;

	case EVENT_SET_RESET_PROFILE:
		return __event_callback_reset_profile;
		break;

	case EVENT_UIC_ALERT:
		return __event_callback_uic_alert;
		break;

	case EVENT_CHECK_PIN_CODE:
		return __event_callback_check_pin_code;
		break;

	case EVENT_ADD_DMACC:
		return __event_add_dmacc;
		break;

	case EVENT_REMOVE_DMACC:
		return __event_remove_dmacc;
		break;

	default:
		return __event_callback_default;
		break;
	}

}

// Type 0, No Reply
static int __event_callback_software_update(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

#if 1
	//      IPC_EVENT_TYPE_0;
	/*sleep(10); // for test */

	DM_ERROR ret;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	UI_TYPE ui_type;
	char *session_id = NULL;
	char *fota_ui_type = NULL;
	char *server_id = NULL;
	Event_Type triger_type = POLL_EVENT;
	int task_id = 0;
	SERVICE_SERVER_TYPE server_type = SAMSUNG_FUMO_TYPE;
	OPER_MODE operationMode = OPER_UNKNOWN;
	NotI_Type noti_type = NOTI_TYPE_NOT_SPECIFIED;

	sync_agent_get_event_data_param_int(request, (int*) &ui_type);
	_DEBUG_INFO("UI TYPE : %d\n", ui_type);

	switch (ui_type) {
	case FOTA_COMMON_UI:
		_DEBUG_INFO("common ui");
		fota_ui_type = OMADM_FUMO_UI_TYPE;
		sync_agent_get_event_data_param_int(request, &task_id);
		_DEBUG_INFO("get task id : %d", task_id);
		sync_agent_get_event_data_param_int(request, (int*) &server_type);
		triger_type = UI_EVENT;
		ret = get_Serverid(server_type, &server_id);
		//ret =get_Serverid(SAMSUNG_FUMO_TYPE , &server_id);
		//ret =get_Serverid(GCF_TYPE , &server_id);
		_DEBUG_INFO("get server id : %s : %d", server_id, ret);
		operationMode = OPER_USER_INITIATION;
		break;
	case FOTA_NOTI_UI:
		_DEBUG_INFO("noti ui");
		fota_ui_type = OMADM_FUMO_NOTI_UI_TYPE;
		sync_agent_get_event_data_param_str(request, &session_id);
		sync_agent_get_event_data_param_str(request, &server_id);
		sync_agent_get_event_data_param_int(request,(int*) &triger_type);
		_DEBUG_INFO("get session id : %s", session_id);
		_DEBUG_INFO("get server id : %s", server_id);
		_DEBUG_INFO("get triger_type : %d", triger_type);

		operationMode = OPER_SERVER_INITIATION_FROM_IP;
		break;
	case FOTA_BACKGROUND_UI:
		_DEBUG_INFO("background ui");
		break;
	case ADMIN_UI:
		_DEBUG_INFO("admin ui");
		fota_ui_type = OMADM_ADMIN_UI_TYPE;
		sync_agent_get_event_data_param_int(request, (int*) &server_type);
		ret = get_Serverid(server_type, &server_id);
		_DEBUG_INFO("get server id : %s : %d", server_id, ret);
		operationMode = OPER_USER_INITIATION;
		break;
	case CP_UI:
		_DEBUG_INFO("cp ui");
		break;
	default:
		_DEBUG_INFO("not use ui");
		break;
	}

	if (server_id != NULL) {
		_DEBUG_INFO("fumo server id = %s", server_id);
		Event_Contents *ev_data = (Event_Contents *) calloc(1, sizeof(Event_Contents));
		if (ev_data == NULL) {
			_DEBUG_INFO("alloc fail");
			goto error;
		}
		//ev_data->type = UI_EVENT;
		//ev_data->type = IP_PUSH_EVENT;
		ev_data->type = triger_type;
		_DEBUG_INFO("get triger_type : %d", triger_type);
		if (session_id != NULL)
			ev_data->session_id = strdup(session_id);
		if (server_id != NULL) {
			ev_data->server_id = strdup(server_id);
			str_free(&server_id);
		}
		ev_data->noti_type = noti_type;
		ev_data->ui_mode = strdup(fota_ui_type);

		ret = dm_fumo_common_task_request(ev_data);
		if (ret != DM_OK) {
			event_data_free((void *)ev_data);
			goto error;
		}
	} else {
		goto error;
	}

	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

#else

	// IPC_EVENT_TYPE_1;
	int response_int = REPLY_FAIL_TO_CONNECT;	//initial value

	//Todo ------------------------------------------------------------------------------------------------------------------------    (from here)
	sleep(10);		// for test

	response_int = REPLY_SUCCESS_TO_CONNECT;	// resulted value
	// ----------------------------------------------------------------------------------------------------------------------------------    (end)
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &response_int);
#endif

	str_free(&(session_id));
	str_free(&(server_id));
	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	str_free(&(session_id));
	str_free(&(server_id));
	return 0;

}

// Type 0, No Reply
static int __event_callback_cancel(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	_DEBUG_VERBOSE(">>> >>> >>> Connect to server	 CANCELED !!!\n");

	int task_id;
	sync_agent_get_event_data_param_int(request, &task_id);
	_DEBUG_VERBOSE("task id : %d\n", task_id);

	/*sync_agent_cancel_task(task_id); */
	cancel_task(task_id);

	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
}

// Type 0, No Reply
static int __event_callback_download(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	DM_ERROR ret = DM_OK;

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	int response_type = 0;
	RESUME_STATUS download_clicked = 0;

	sync_agent_get_event_data_param_int(request, &response_type);
	_DEBUG_INFO("DOWNLOAD BUTTON TYPE : %d\n", response_type);

	ret = DB_Open();
	_DEBUG_INFO("db open : %d", ret);

	if (response_type == RESPONSE_TYPE_OK) {
		/*download button clicked */
		download_clicked = RESUME_STATUS_DOWNLOAD;
		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_DOWNLOAD_CLICK, &download_clicked);
		_DEBUG_INFO("update engine status clolumn download click : %d", ret);

	} else if (RESPONSE_TYPE_CANCEL) {
		/*later button clicked */
		download_clicked = RESUME_STATUS_DOWNLOAD_LATER;
		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_DOWNLOAD_CLICK, &download_clicked);
		_DEBUG_INFO("update engine status clolumn download click : %d", ret);
	}
	DB_Close();

	if (response_type == RESPONSE_TYPE_OK) {
		ret = fumo_serviceEngine_task_request();
		if (ret != DM_OK) {
			goto error;
		}
	}
	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

// Type 0, No Reply
static int __event_callback_install(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	DM_ERROR ret = DM_OK;

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	int response_type = 0;
	RESUME_STATUS download_clicked = 0;

	sync_agent_get_event_data_param_int(request, &response_type);
	_DEBUG_VERBOSE("INSTALL BUTTON TYPE : %d\n", response_type);

	ret = DB_Open();
	_DEBUG_VERBOSE("db open : %d", ret);

	if (response_type == RESPONSE_TYPE_OK) {
		/*download button clicked */
		download_clicked = RESUME_STATUS_INSTALL;
		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_DOWNLOAD_CLICK, &download_clicked);
		_DEBUG_VERBOSE("update engine status clolumn install click : %d", ret);

	} else if (RESPONSE_TYPE_CANCEL) {
		/*later button clicked */
		download_clicked = RESUME_STATUS_INSTALL_LATER;
		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_DOWNLOAD_CLICK, &download_clicked);
		_DEBUG_VERBOSE("update engine status clolumn install click : %d", ret);
	}
	DB_Close();

	ret = fumo_serviceEngine_task_request();
	if (ret != DM_OK) {
		goto error;
	}
	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

// Type 0, No Reply
static int __event_callback_set_interval(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	DM_ERROR ret = DM_OK;
	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	int request_param = INTERVAL_CANCEL;

	sync_agent_get_event_data_param_int(request, &request_param);
	_DEBUG_INFO("interval val from UI : %d\n", request_param);

	switch (request_param) {
	case RESPONSE_TYPE_REMINDER_INTERVAL_1_HOUR:
		request_param = INTERVAL_1HOUR;
		break;
	case RESPONSE_TYPE_REMINDER_INTERVAL_3_HOURS:
		request_param = INTERVAL_3HOURS;
		break;
	case RESPONSE_TYPE_REMINDER_INTERVAL_6_HOURS:
		request_param = INTERVAL_6HOURS;
		break;
	case RESPONSE_TYPE_REMINDER_INTERVAL_12_HOURS:
		request_param = INTERVAL_12HOURS;
		break;
	case RESPONSE_TYPE_REMINDER_INTERVAL_1_DAY:
		request_param = INTERVAL_1DAY;
		break;
	case RESPONSE_TYPE_CANCEL:
		request_param = INTERVAL_CANCEL;
		break;
	default:
		break;
	}

	_DEBUG_INFO(" interval val in fumo engine : %d\n", request_param);
	if (request_param != INTERVAL_CANCEL) {
		ret = fumo_schedule_task_request(request_param);
		if (ret != DM_OK) {
			goto error;
		}
		_DEBUG_INFO("fumo schedule task result  : %d\n", ret);
	} else {
/*
		ret = fumo_schedule_task_request(request_param);
		if (ret != DM_OK) {
			goto error;
		}
*/
		ret = DB_Open();
		_DEBUG_INFO("db open : %d", ret);

		int engine_return = FUMO_USER_CANCELLED;// 401;
		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_RESULT_STATUS, &engine_return);
		_DEBUG_INFO("ret[%d] service engine status generic alert status : %d", ret, engine_return);
		DB_Close();

		ret = fumo_serviceEngine_task_request();
		if (ret != DM_OK) {
			goto error;
		}
		_DEBUG_INFO("fumo schedule cancel : %d \n", ret);
	}

	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;

 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 1;
}

// Type 0, No Reply
static int __event_callback_account_registration(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

   /*
	DM_ERROR push_ret = DM_OK;
	push_ret = register_fumo_ip_push();
	if (push_ret == -1) {
		_DEBUG_INFO("fail register fumo ip push : %d ", push_ret);
	} else {
		_DEBUG_INFO("success register fumo ip push : %d ", push_ret);
	}
   */

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

	int ret = -1;
	//temp code
	int db_ret = 0;
	db_ret = sync_agent_open_agent();
	_DEBUG_INFO("OPEN DACI : %d", db_ret);

	ret = register_fota_account();

	if (ret == 0) {
		_DEBUG_INFO("fail register fota account : %d ", ret);
		goto error;
	} else {
		_DEBUG_INFO("success register fota account : %d ", ret);

		char * str_version_url = NULL;
		str_version_url = vconf_get_str(VERSION_URL_KEY);
		if(str_version_url == NULL) {
			_DEBUG_ERROR("vconf_get_str[%s] failed", VERSION_URL_KEY);
			ret = EVENT_RESPONSE_FAIL;
			_INNER_FUNC_EXIT;
			goto error;
		}
		char * str_version_filename = NULL;
		str_version_filename = vconf_get_str(VERSION_FILENAME_KEY);
		if(str_version_filename == NULL) {
			_DEBUG_ERROR("vconf_get_str[%s] failed", VERSION_FILENAME_KEY);
			ret = EVENT_RESPONSE_FAIL;
			goto error;
		}

		int exist_version = -1;
		//ret = register_fota_polling(str_version_url, str_version_filename, &exist_version);
		_DEBUG_INFO("exist_version [%d]", exist_version);

	}

	db_ret = sync_agent_close_agent();
	_DEBUG_INFO("CLOSE DACI : %d", db_ret);

	//2012.11.20
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;

 error:
	db_ret = sync_agent_close_agent();
	_DEBUG_INFO("CLOSE DACI : %d", db_ret);

	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

// Type 1, Fumo setting data Reply
static int __event_callback_get_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

	char *profile_name = NULL;
	SERVICE_SERVER_TYPE sv_type = NO_ENGINE_TYPE;

	char *client_id = NULL;
	char *client_pwd = NULL;
	char *client_sourceUrl = NULL;
	char *client_targetUrl = NULL;
	int client_isBase64 = 0;
	char *client_nextNonce = NULL;
	char *client_authType = NULL;

	char *server_url = NULL;
	char *server_pwd = NULL;
	char *server_sourceUrl = NULL;
	char *server_targetUrl = NULL;
	int server_isBase64 = 0;
	char *server_nextNonce = NULL;
	char *server_authType = NULL;

	sync_agent_dm_mo_error_e db_ret = SYNC_AGENT_DM_MO_SUCCESS;
	char *server_id = NULL;

	sync_agent_get_event_data_param_int(request, (int*) &sv_type);
	_DEBUG_VERBOSE("server type : %d\n", sv_type);

	db_ret = sync_agent_open_mo();
	_DEBUG_VERBOSE("mo db open : %d", db_ret);

	switch (sv_type) {
	case SAMSUNG_FUMO_TYPE:
		{
			profile_name = "Fumo";
		}
		break;
	case SAMSUNG_FMM_TYPE:
		{
			profile_name = "Lawmo";
		}
		break;
	case GCF_TYPE:
		{
			profile_name = "gcf";
		}
		break;
	default:
		{
			_DEBUG_VERBOSE("not exist server id error");
			goto error;
		}
	}

	get_Serverid(sv_type, &server_id);
	_DEBUG_VERBOSE("server id : %s", server_id);

	if (server_id != NULL) {
		DM_ERROR err = get_client_dmacc(server_id, &client_id, &client_pwd, &client_sourceUrl, &client_targetUrl, &client_isBase64, &client_nextNonce, &client_authType);
		if (err != DM_OK) {
			_DEBUG_VERBOSE("get_client_dmacc() fail !!");
			goto error;
		} else {
			_DEBUG_VERBOSE("get_client_dmacc() success !!");
		}

		err = get_server_dmacc(server_id, &server_url, &server_pwd, &server_sourceUrl, &server_targetUrl, &server_isBase64, &server_nextNonce, &server_authType);
		if (err != DM_OK) {
			_DEBUG_VERBOSE("get_server_dmacc() fail !!");
			goto error;
		} else {
			_DEBUG_VERBOSE("get_server_dmacc() success !!");
		}

		db_ret = sync_agent_close_mo();
		_DEBUG_VERBOSE("mo db close : %d", db_ret);

		_DEBUG_VERBOSE("profile_name : %s", profile_name);
		_DEBUG_VERBOSE("client_targetUrl : %s", client_targetUrl);
		_DEBUG_VERBOSE("server_id : %s", server_url);
		_DEBUG_VERBOSE("server_pwd : %s", server_pwd);
		_DEBUG_VERBOSE("client_id : %s", client_id);
		_DEBUG_VERBOSE("client_pwd : %s", client_pwd);
		_DEBUG_VERBOSE("client_authType : %s", client_authType);
		_DEBUG_VERBOSE("server_authType : %s", server_authType);

		//2012.11.20
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, profile_name);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, client_targetUrl);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, server_url);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, server_pwd);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, client_id);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, client_pwd);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, client_authType);
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, server_authType);

		str_free(&server_nextNonce);
		str_free(&client_nextNonce);
		str_free(&client_sourceUrl);
		str_free(&server_sourceUrl);
		str_free(&server_targetUrl);
		str_free(&client_targetUrl);
		str_free(&server_url);
		str_free(&server_pwd);
		str_free(&client_id);
		str_free(&client_pwd);
		str_free(&client_authType);
		str_free(&server_authType);
	} else {
		_DEBUG_VERBOSE("not exist server id error");
		goto error;
	}

	_INNER_FUNC_EXIT;
	return 0;
 error:

	db_ret = sync_agent_close_mo();
	_DEBUG_VERBOSE("mo db close : %d", db_ret);
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_set_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

	int ret = -1;
	SERVICE_SERVER_TYPE sv_type = NO_ENGINE_TYPE;
	char *server_id = NULL;

	dm_acc *dmaccount = (dm_acc *) calloc(1, sizeof(dm_acc));
	if (dmaccount == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		goto error;
	}
	//update 2012.11.20
	sync_agent_get_event_data_param_int(request, (int*) &sv_type);
	_DEBUG_VERBOSE("server profile type : %d", sv_type);
	sync_agent_get_event_data_param_str(request, &(dmaccount->server_url));
	_DEBUG_VERBOSE("server url : %s", (dmaccount->server_url));
	sync_agent_get_event_data_param_str(request, &(dmaccount->server_id));
	_DEBUG_VERBOSE("server id : %s", dmaccount->server_id);
	sync_agent_get_event_data_param_str(request, &(dmaccount->server_pw));
	_DEBUG_VERBOSE("server_pw : %s", dmaccount->server_pw);
	sync_agent_get_event_data_param_str(request, &(dmaccount->user_id));
	_DEBUG_VERBOSE("user_id : %s", dmaccount->user_id);
	sync_agent_get_event_data_param_str(request, &(dmaccount->user_pw));
	_DEBUG_VERBOSE("user_pw : %s", dmaccount->user_pw);
	sync_agent_get_event_data_param_str(request, &(dmaccount->user_auth_type));
	_DEBUG_VERBOSE("user_auth_type : %s", dmaccount->user_auth_type);
	sync_agent_get_event_data_param_str(request, &(dmaccount->server_auth_type));
	_DEBUG_VERBOSE("server_auth_type : %s", dmaccount->server_auth_type);

	get_Serverid(sv_type, &server_id);
	_DEBUG_VERBOSE("server id : %s", server_id);

	if (server_id != NULL) {
		ret = update_dmacc(server_id, dmaccount);
		if (ret != 1) {
			_DEBUG_VERBOSE("update dm acc fail", ret);
			goto error;

		} else {
			_DEBUG_VERBOSE("update dm acc success", ret);
		}

	} else {
		goto error;
	}

	_DEBUG_VERBOSE("update dm acc : %d", ret);
	free_dm_acc(dmaccount);

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	free_dm_acc(dmaccount);
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_reset_profile(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	SERVICE_SERVER_TYPE sv_type = NO_ENGINE_TYPE;

	sync_agent_get_event_data_param_int(request, (int*) &sv_type);
	_DEBUG_VERBOSE("PROFILE TYPE : %d\n", sv_type);

	switch (sv_type) {
	case SAMSUNG_FUMO_TYPE:
		{
			reset_dmacc_msctestserver();
		}
		break;
	case SAMSUNG_FMM_TYPE:
		{
			reset_dmacc_mscserver();
		}
		break;
	case GCF_TYPE:
		{
			reset_dmacc_gcf();
		}
		break;
	default:
		{
			goto error;
		}
	}

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_uic_alert(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	uic_type type = UIC_NONE_TYPE;
	RESPONSE_TYPE response_ret = EVENT_RESPONSE_FAIL;
	char *input_text = NULL;
	int count = 0;
	int i = 0;

	/*
	if (type == UIC_DISPLAY_TYPE) {
		return 0;
	}
	*/
	free_res_data(res_uic_data);

	sync_agent_get_event_data_param_int(request, (int*) &response_ret);
	_DEBUG_TRACE("response is %d", response_ret);

	sync_agent_get_event_data_param_int(request, (int*) &type);
	_DEBUG_TRACE("uic type %d", type);

	/*alloc reponse data */
	res_uic_data = calloc(1, sizeof(uic_res_data) + 1);
	res_uic_data->status = response_ret;

	if (response_ret == RESPONSE_TYPE_OK) {

		_DEBUG_TRACE("response ok");

		switch (type) {
		case UIC_CONFIRMATION_TYPE:
			break;
		case UIC_INPUT_TEXT_TYPE:{
				sync_agent_get_event_data_param_str(request, &input_text);
				_DEBUG_TRACE("input text %s", input_text);

				res_uic_data->input_text = strdup(input_text);
			}
			break;
		case UIC_SINGLE_CHOICE_TYPE:
		case UIC_MULTIPLE_CHOICE_TYPE:{
				sync_agent_get_event_data_param_int(request, &count);
				_DEBUG_TRACE("count  is %d\n", count);

				if (count != 0) {
					/*single, multiple choice */
					for (i = 0; i < count; ++i) {
						char *input = NULL;
						char *value = NULL;
						sync_agent_get_event_data_param_str(request, &input);
						_DEBUG_TRACE(" choice : %s", input);
						if (input == NULL) {
							input = strdup("NULL");
							res_uic_data->res_multi_data = g_list_append(res_uic_data->res_multi_data, input);
						} else {
							_DEBUG_TRACE("value data : %s", input);
							value = strdup(input);
							res_uic_data->res_multi_data = g_list_append(res_uic_data->res_multi_data, value);
						}
					}
				} else {
					_DEBUG_TRACE("not choice");
				}
			}
			break;
		case UIC_DISPLAY_TYPE:
		default:
			break;
		}
	} else {
		/*
		 * time out
		 * error
		 * ...
		 */
		_DEBUG_TRACE("response fail");
	}

	//signal off
	g_mutex_lock(&(mutex));
	g_cond_signal(&(cond));
	g_mutex_unlock(&(mutex));

	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_get_fumo_config(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

	int wifi_only = -1;
	sync_agent_open_agent();

	wifi_only = get_wifi_only_config();
	sync_agent_close_agent();

	if (wifi_only == -1) {
		_DEBUG_VERBOSE("get wifi only config fail\n");
		goto error;
	} else {
		_DEBUG_VERBOSE("get wifi only config success\n");
	}

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_DEBUG_VERBOSE("wifi only : %d", wifi_only);
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &wifi_only);

	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_set_fumo_config_wifi_only(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	int ret = 1;
	int wifi_only_param;

	sync_agent_get_event_data_param_int(request, &wifi_only_param);

	sync_agent_open_agent();
	ret = set_wifi_only_config(wifi_only_param);
	sync_agent_close_agent();
	if (ret != 1) {
		goto error;
	}

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
 error:
	//2012.11.20
	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_check_pin_code(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	//2012.11.20
	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;

	CP_ERROR err = CP_ERROR_SUCCESS;

	int ext_id = 0;
	sync_agent_get_event_data_param_int(request, &ext_id);
	_DEBUG_VERBOSE("ext id : %d", ext_id);

	char *user_pin = NULL;

	int response_type = RESPONSE_TYPE_NONE;
	sync_agent_get_event_data_param_int(request, &response_type);
	_DEBUG_VERBOSE("response type :  %d", response_type);

	int response_value = 0;

	if (response_type == RESPONSE_TYPE_OK) {
		_DEBUG_VERBOSE("check pin code !!");
		sync_agent_get_event_data_param_str(request, &user_pin);
		_DEBUG_VERBOSE("user pin : %s", user_pin);

		int is_correct = 0;
		err = process_Check_Pincode(0, user_pin, &is_correct, ext_id);
		if (err != CP_ERROR_SUCCESS) {
			_DEBUG_VERBOSE("process_Check_Pincode() fail !!");
			sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

			response_value = 0;
			/* todo : exception handling */
			sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &response_value);
		} else {
			_DEBUG_VERBOSE("process_Check_Pincode() success !!");

			sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

			if (is_correct != 0) {
				response_value = 1;
			} else {
				response_value = 0;
			}
			sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &response_value);

			/* process_Setting_CP() */
			if (is_correct == 1) {
				_DEBUG_VERBOSE("Valid PIN !!");

				err = process_Setting_CP(ext_id);
				if (err != CP_ERROR_SUCCESS) {
					_DEBUG_VERBOSE("process_Setting_CP() fail !!");
				} else {
					_DEBUG_VERBOSE("process_Setting_CP() success !!");
				}
			} else if (is_correct == 0) {
				_DEBUG_VERBOSE("Invalid PIN !!");
			}
		}
	} else {		/* response_type == RESPONSE_TYPE_CANCEL */
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

		response_value = 1;	/* meaningless value */
		sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &response_value);

		if (ext_id != -1) {
			_DEBUG_VERBOSE("fail logic !!");
			err = process_Invalid_Pincode(ext_id);
			if (err != CP_ERROR_SUCCESS) {
				_DEBUG_VERBOSE("process_Invalid_Pincode() fail !!");
				/* todo : exception handling */
			} else {
				_DEBUG_VERBOSE("process_Invalid_Pincode() success !!");
			}
		} else {
			_DEBUG_VERBOSE("cancel logic !!");
			/* do nothing !! */
		}
	}

	str_free(&(user_pin));
	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_add_dmacc(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	sync_agent_dm_mo_error_e db_ret = SYNC_AGENT_DM_MO_SUCCESS;
	DM_ERROR ret = DM_OK;

	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	SERVICE_SERVER_TYPE sv_type = NO_ENGINE_TYPE;

	sync_agent_get_event_data_param_int(request, (int*) &sv_type);
	_DEBUG_VERBOSE("server type : %d\n", sv_type);

	db_ret = sync_agent_open_mo();
	_DEBUG_VERBOSE("mo db open : %d", db_ret);

	ret = DB_Open();
	_DEBUG_VERBOSE("dm status open : %d", db_ret);

	ret = DB_Begin_Transaction();
	_DEBUG_VERBOSE("begin transaction : %d", ret);

	switch (sv_type) {
		//todo will change ipc refactoring
	case SAMSUNG_FUMO_TYPE:
		{
			_DEBUG_VERBOSE("add mscservertest dmacc");
			ret = add_dm_acc(sv_type);
		}
		break;
	case SAMSUNG_FMM_TYPE:
		{
			_DEBUG_VERBOSE("add mscserver dmacc");
			ret = add_dm_acc(sv_type);
		}
		break;
	case GCF_TYPE:
		{
			_DEBUG_VERBOSE("add gcf dmacc");
			ret = add_dm_acc(sv_type);
		}
		break;
	default:
		{
			_DEBUG_VERBOSE("not exist server id error");
		}
	}

	_DEBUG_VERBOSE("add dm acc error :%d", ret);

	if (ret != DM_OK) {
		ev_response = EVENT_RESPONSE_FAIL;
		goto error;
	}

	ret = DB_End_Transaction(TRANSACTION_COMMIT_);
	_DEBUG_VERBOSE("end transaction : %d", ret);

	db_ret = sync_agent_close_mo();
	_DEBUG_VERBOSE("mo db close : %d", db_ret);

	DB_Close();

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;

 error:

	ret = DB_End_Transaction(TRANSACTION_ROLLBACK_);
	_DEBUG_VERBOSE("end transaction : %d", ret);

	db_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_ROLLBACK);
	_DEBUG_VERBOSE("end transaction : %d", db_ret);

	db_ret = sync_agent_close_mo();
	_DEBUG_VERBOSE("mo db close : %d", db_ret);

	DB_Close();

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_remove_dmacc(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;

	sync_agent_dm_mo_error_e db_ret = SYNC_AGENT_DM_MO_SUCCESS;
	DM_ERROR ret = DM_OK;
	engine_status *status = NULL;
	ENGINE_STATUS engine_status = DM_IDLE;

	EVENT_RESPONSE ev_response = EVENT_RESPONSE_OK;
	SERVICE_SERVER_TYPE sv_type = NO_ENGINE_TYPE;

	sync_agent_get_event_data_param_int(request, (int*) &sv_type);
	_DEBUG_VERBOSE("server type : %d\n", sv_type);

	//mo db
	db_ret = sync_agent_open_mo();
	_DEBUG_VERBOSE("mo db open : %d", db_ret);

	//config, alram db
	ret = DB_Open();
	_DEBUG_VERBOSE("config alarm open : %d", db_ret);

	int is_exist = IsExist_Engine_id(IDLE_ENGINE);
	if (is_exist == 0)
		goto error;

	ret = Get_Engine_Status(IDLE_ENGINE, &status);
	if (ret != DM_OK) {
		_DEBUG_ERROR("Get_Engine_Status is not DM_OK");
		goto error;
	}

	if(status == NULL) {
		_DEBUG_ERROR("status is null");
		goto error;
	}

	_DEBUG_VERBOSE("STATUS VALUE : %d ", status->engine_status);
	engine_status = status->engine_status;

	if (engine_status == DM_BEFORE_BOOTSTRAP) {
		_DEBUG_VERBOSE("not bootstrap");
		goto error;
	}

	db_ret = sync_agent_begin_transaction_mo();
	_DEBUG_VERBOSE("end transaction : %d", db_ret);

	ret = DB_Begin_Transaction();
	_DEBUG_VERBOSE("end transaction : %d", ret);

	switch (sv_type) {

	case UI_SAMSUNG_FUMO_TYPE:
		{
			_DEBUG_VERBOSE("remove mscservertest dmacc");
			ret = remove_dm_acc(DM_DMACC_MSCTESTSERVER);
			if (ret != DM_OK) {
				_DEBUG_ERROR("remove dm acc error :%d", ret);
				goto error;
			} else {
				//config db
				int is_exist = IsExist_Engine_id(FUMO_SERVICE_ENGINE);
				if (is_exist == 0) {
					_DEBUG_VERBOSE("not exist fumo engine");
				}

				ret = Delete_Engine_Status(FUMO_SERVICE_ENGINE);
				if (ret != DM_OK) {
					_DEBUG_ERROR("delete engine status :%d", ret);
					goto error;
				}
				//alarm db
				delete_alarm_item(FUMO_INTERVAL, FUMO_INTERVAL_TYPE);
				_DEBUG_TRACE(" delete fumo reminder in config db ");

				//delete delta file
				delete_fumo_contents(MEMORY_INTERNAL);
				delete_fumo_contents(MEMORY_SD_CARD);

				//change fumo state 10
				char *fumo_state = NULL;
				fumo_state = g_strdup_printf("%s%s%s%s", MO_ROOT, MO_SEPERATOR, FUMO_SERVICE, DM_FUMO_STATE_OP);
				set_current_fumo_state(fumo_state, IDLE_START);
				str_free(&fumo_state);
			}
		}
		break;
	case UI_SAMSUNG_FMM_TYPE:
		{
			_DEBUG_VERBOSE("remove mscserver dmacc");
			ret = remove_dm_acc(DM_DMACC_MSCSERVER);
			if (ret != DM_OK) {
				_DEBUG_ERROR("remove dm acc error :%d", ret);
				goto error;
			} else {

				//config db
				int is_exist = IsExist_Engine_id(LAWMO_SERVICE_ENGINE);
				if (is_exist == 0) {
					_DEBUG_VERBOSE("not exist fumo engine");
				}

				ret = Delete_Engine_Status(LAWMO_SERVICE_ENGINE);
				if (ret != DM_OK) {
					_DEBUG_ERROR("delete engine status :%d", ret);
					goto error;
				}

			}
		}
		break;
	case UI_GCF_TYPE:
		{
			_DEBUG_VERBOSE("remove gcf dmacc");
			ret = remove_dm_acc(DM_DMACC_GCF);
			if (ret != DM_OK) {
				_DEBUG_ERROR("remove dm acc error :%d", ret);
				goto error;
			} else {
				_DEBUG_VERBOSE("remove dm acc");
			}
		}
		break;
	default:
		{
			_DEBUG_VERBOSE("not exist server id error");
		}
	}

	ret = DB_End_Transaction(TRANSACTION_COMMIT_);
	_DEBUG_VERBOSE("end transaction : %d", ret);

	db_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_COMMIT);
	_DEBUG_VERBOSE("end transaction : %d", db_ret);

	_DEBUG_VERBOSE("remove dm acc error :%d", ret);

	db_ret = sync_agent_close_mo();
	_DEBUG_VERBOSE("mo db close : %d", db_ret);

	DB_Close();

	if (status != NULL) {
		Free_Memory_Engine_Status(&status, 1);
	}

	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
 error:

	ret = DB_End_Transaction(TRANSACTION_ROLLBACK_);
	_DEBUG_VERBOSE("end transaction : %d", ret);

	db_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_ROLLBACK);
	_DEBUG_VERBOSE("end transaction : %d", db_ret);

	db_ret = sync_agent_close_mo();
	_DEBUG_VERBOSE("mo db close : %d", db_ret);

	DB_Close();

	if (status != NULL) {
		Free_Memory_Engine_Status(&status, 1);
	}

	ev_response = EVENT_RESPONSE_FAIL;
	sync_agent_append_event_data_param(response, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &ev_response);

	_INNER_FUNC_EXIT;
	return 0;
}

static int __event_callback_default(sync_agent_event_data_s * request, sync_agent_event_data_s * response)
{
	_INNER_FUNC_ENTER;
	retvm_if((request) == NULL, 0, "request is NULL!!");

	_DEBUG_VERBOSE("Undefined Event !!! \n");

	_INNER_FUNC_EXIT;
	return 0;
}

int set_IPC()
{
	_EXTERN_FUNC_ENTER;

	sync_agent_event_error_e error = SYNC_AGENT_EVENT_IPC_ERR;
	int event_id = 0;
	for (event_id = 1; event_id < EVENT_NUM; event_id++) {
		error = sync_agent_set_event_callback(event_id, _get_event_callback(event_id));
		if (error != SYNC_AGENT_EVENT_SUCCESS) {
			_DEBUG_INFO("FAILURE: sync_agent_set_event_callback (event id: %d)\n", event_id);
			break;
		} else {
			_DEBUG_INFO("SUCCESS: sync_agent_set_event_callback (event id: %d)\n", event_id);

			g_mutex_init(&(mutex));
			g_cond_init(&(cond));
		}
	}

	_DEBUG_INFO("%s: end (error:%d, event_id:%d) \n", __func__, error, event_id);
	_EXTERN_FUNC_EXIT;
	return error;
}

int noti_start(char *ui_mode)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti start-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ENGINE_START, 0, ui_mode, NULL);
}

int noti_download(int file_size)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_connect_to_server-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_DOWNLOAD, file_size, NULL, NULL);
}

int noti_install()
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_install-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_INSTALL, 0, NULL, NULL);
}

int noti_send_download_info(int filesize, char *path)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_send_download_info-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_DOWNLOAD_INFO, filesize, path, NULL);
}

int noti_low_battery(char *battery_level)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_low_battery-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_LOW_BATTERY, 0, battery_level, NULL);
}

int noti_update_result(int result)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_update_result-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_UPDATE_RESULT, result, NULL, NULL);
}

int alert_uic_display(uic_data * uic_value)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_alert_ui_display-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ALERT_DISPLAY, 0, OMADM_ALERT_UI_TYPE, uic_value);
}

int alert_uic_confirmation(uic_data * uic_value)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_alert_ui_confirmation-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ALERT_CONFIRMATION, 0, OMADM_ALERT_UI_TYPE, uic_value);
}

int alert_uic_inputText(uic_data * uic_value)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_alert_ui_inputText-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ALERT_INPUTTEXT, 0, OMADM_ALERT_UI_TYPE, uic_value);
}

int alert_uic_single_choice(uic_data * uic_value)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_alert_ui_single_choice-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ALERT_SINGLE_CHOICE, 0, OMADM_ALERT_UI_TYPE, uic_value);
}

int alert_uic_multiple_choice(uic_data * uic_value)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_alert_ui_multi_choice-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ALERT_MULTIPLE_CHOICE, 0, OMADM_ALERT_UI_TYPE, uic_value);
}

int noti_engine_fail(int prev_noti_id, int fail_type)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_engine_fail-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_ENGINE_FAIL, prev_noti_id, 0, &fail_type);
}

int noti_wifi_only_download_fail()
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_wifi_only_download_fail-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_WIFI_ONLY_DOWNLOAD_FAIL, 0, 0, 0);
}

int noti_memory_full(int file_size)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_memory_full-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_MEMORY_FULL, file_size, 0, 0);
}

int noti_over_max_file_size(int file_size)
{
	_EXTERN_FUNC_ENTER;
	_DEBUG_INFO("-----------------------------------------------------------------------noti_over_max_file_size-------------------------------------------------------------------------------------");
	_EXTERN_FUNC_EXIT;
//	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_OVER_BIG_SIZE, 0, 0, 0);
	return _IPC_noti(IPC_NOTI_TYPE_0, NOTI_OVER_BIG_SIZE, file_size, 0, 0);
}

static int _IPC_noti(int noti_type, int noti_id, int input_param, char *input_string, void *user_data)
{
	_INNER_FUNC_ENTER;

	int ret = RESPONSE_TYPE_NONE;
	int task_id = -1;
	char *ui_mode = NULL;
	int count = 0;
	GList *iter = NULL;
	uic_data *uic_value = NULL;
	sync_agent_event_data_s *noti = NULL;
	sync_agent_event_data_s *reply_msg = NULL;
	sync_agent_event_error_e err = SYNC_AGENT_EVENT_FAIL;

	task_id = sync_agent_get_self_request_id();
	_DEBUG_TRACE("task : id %d ", task_id);
	if (task_id == -1) {
		ret = RESPONSE_TYPE_CANCEL;
		_INNER_FUNC_EXIT;
		goto return_part;
	}

	ret = get_fumo_ui_mode(&ui_mode);
	_DEBUG_TRACE("get ui mode : %d", ret);
	if (ret != DM_OK) {
		ret = RESPONSE_TYPE_CANCEL;
		_INNER_FUNC_EXIT;
		return ret;
	}

	_DEBUG_TRACE("ui mode : %s", ui_mode);
	ret = __check_ui_mode(&ui_mode, input_string);
	if (ret != RESPONSE_TYPE_OK) {
		str_free(&(ui_mode));
		_INNER_FUNC_EXIT;
		return RESPONSE_TYPE_CANCEL;
	}
	_DEBUG_TRACE("ui mode : %s", ui_mode);

	ret = __launch_ui(ui_mode, noti_type);
	if (ret == -1) {
		str_free(&(ui_mode));
		_INNER_FUNC_EXIT;
		return RESPONSE_TYPE_CANCEL;
	} else if (ret == 0) {
		str_free(&(ui_mode));
		_INNER_FUNC_EXIT;
		return RESPONSE_TYPE_OK;
	}

	_DEBUG_TRACE("ui mode : %s", ui_mode);
	_DEBUG_TRACE("create noti  noti_id : %d", noti_id);

	noti = sync_agent_create_noti(noti_id);
	if (noti == NULL) {
		_DEBUG_TRACE("failed to sync_agent_create_noti");
		str_free(&(ui_mode));
		_INNER_FUNC_EXIT;
		return ret;
	}

	_DEBUG_TRACE("append_eventdata");
	switch (noti_id) {
	case NOTI_UNKNOWN:
		break;
	case NOTI_DOWNLOAD:	//Type 0
		{
			_DEBUG_TRACE("download file size : %d", input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);
		}
		break;
	case NOTI_INSTALL:	//Type 0
		break;
	case NOTI_DOWNLOAD_INFO:	//Type 0
		{
			if (&input_param == NULL || input_string == NULL) {
				_DEBUG_TRACE("Error: input param: NULL\n");
				ret = RESPONSE_TYPE_CANCEL;
				_INNER_FUNC_EXIT;
				goto return_part;
			}
			_DEBUG_TRACE("agent noti: filesize:%d, path:%s\n", input_param, input_string);

			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, input_string);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &task_id);
		}
		break;
	case NOTI_ENGINE_START:	//Type 0
		{
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &task_id);
		}
		break;
	case NOTI_ENGINE_FAIL:	//Type 0
		{
			int *fail_type = (int *)(user_data);

			_DEBUG_TRACE("noti id : %d\n", input_param);
			_DEBUG_TRACE("fail  : %d\n", *fail_type);

			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, fail_type);
		}
		break;
	case NOTI_WIFI_ONLY_DOWNLOAD_FAIL:	//Type 0
		{
			_DEBUG_TRACE("wifi only downlaod fail ipc");
		}
		break;
	case NOTI_MEMORY_FULL:	//Type 0
		{
			_DEBUG_TRACE("memory full ipc : %d", input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);

		}
		break;
	case NOTI_OVER_BIG_SIZE:	//Type 0
		{
			_DEBUG_TRACE("over max size : %d", input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);
		}
		break;
	case NOTI_LOW_BATTERY:	//Type 0
		{
			if (input_string == NULL) {

				_DEBUG_TRACE("Error : input string : NULL\n");
				ret = RESPONSE_TYPE_CANCEL;
				_INNER_FUNC_EXIT;
				goto return_part;
			}

			_DEBUG_TRACE("agent noti - battery level : %s\n", input_string);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, input_string);
		}
		break;
	case NOTI_UPDATE_RESULT:
		{
			_DEBUG_TRACE("fota update result : %d", input_param);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &input_param);
		}
		break;
	case NOTI_ALERT_DISPLAY:	//Type 2, response event
	case NOTI_ALERT_CONFIRMATION:	//Type 2, response event
	case NOTI_ALERT_INPUTTEXT:	//Type 2, response event
		{
			_DEBUG_TRACE("uic flow noti type : %d", noti_id);
			uic_value = (uic_data *) user_data;

			uic_timeout_val = atoi(uic_value->pUic_option->max_t);

			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->min_t);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->max_t);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->default_data);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->maxlen);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->input_type);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->echo_type);

			_DEBUG_TRACE("display : %s\n", uic_value->display);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->display);
		}
		break;
	case NOTI_ALERT_SINGLE_CHOICE:	//Type 2, response event
	case NOTI_ALERT_MULTIPLE_CHOICE:	//Type 2, response event
		{
			uic_value = (uic_data *) user_data;

			uic_timeout_val = atoi(uic_value->pUic_option->max_t);

			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->min_t);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->max_t);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->default_data);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->maxlen);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->input_type);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->pUic_option->echo_type);

			_DEBUG_TRACE("display : %s\n", uic_value->display);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, uic_value->display);

			count = g_list_length(uic_value->multi_data);
			_DEBUG_TRACE("choice data count : %d\n", count);
			sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_INTEGER, &count);

			for (iter = uic_value->multi_data; iter != NULL; iter = g_list_next(iter)) {
				_DEBUG_TRACE("choice data : %s", (char *)(iter->data));
				sync_agent_append_event_data_param(noti, SYNC_AGENT_EVENT_PARAM_TYPE_STRING, (char *)(iter->data));
			}
		}
		break;

	default:
		break;
	}

	if (noti_id != NOTI_ALERT_DISPLAY && noti_id != NOTI_ALERT_CONFIRMATION && noti_id != NOTI_ALERT_INPUTTEXT && noti_id != NOTI_ALERT_SINGLE_CHOICE && noti_id != NOTI_ALERT_MULTIPLE_CHOICE) {
		DB_Begin_Transaction();

		ret = Update_Engine_Status_Column(FUMO_SERVICE_ENGINE, VALUE_UI_NOTI_TYPE, &noti_id);
		if (ret != DM_OK) {
			_DEBUG_TRACE("update engine status column : %d, %d\n", ret, noti_id);
			ret = RESPONSE_TYPE_CANCEL;
			goto return_part;
		}
	}

	_DEBUG_TRACE("sendNoti");
	reply_msg = sync_agent_send_noti(ui_mode, noti, __dm_task_process_check_cancel_flag, ui_mode, &err);
	if(reply_msg == NULL) {
		_DEBUG_TRACE("sendNoti return : %d", err);
	} else {
		_DEBUG_TRACE("sendNoti return : %d , event_num[%d] , reply_msg data[%s]", err, reply_msg->event_num, reply_msg->data);
	}
	if (err != SYNC_AGENT_EVENT_SUCCESS) {
		if (err == SYNC_AGENT_EVENT_IPC_ERR) {
			//todo : ui off
			if (noti_id != NOTI_ALERT_DISPLAY && noti_id != NOTI_ALERT_CONFIRMATION && noti_id != NOTI_ALERT_INPUTTEXT && noti_id != NOTI_ALERT_SINGLE_CHOICE && noti_id != NOTI_ALERT_MULTIPLE_CHOICE) {
				DB_End_Transaction(TRANSACTION_ROLLBACK_);
			}
		}

		_DEBUG_TRACE("sendNoti (noti id:%d), FAIL\n", noti_id);
		ret = RESPONSE_TYPE_CANCEL;
		goto return_part;

	} else {

		_DEBUG_TRACE("sendNoti (noti id:%d), success\n", noti_id);
		_DEBUG_TRACE("IPC RESPONSE\n");

		if (noti_id != NOTI_ALERT_DISPLAY && noti_id != NOTI_ALERT_CONFIRMATION && noti_id != NOTI_ALERT_INPUTTEXT && noti_id != NOTI_ALERT_SINGLE_CHOICE && noti_id != NOTI_ALERT_MULTIPLE_CHOICE) {
			DB_End_Transaction(TRANSACTION_COMMIT_);
		} else {
			ret = __uic_response(noti_id, user_data);
			_DEBUG_TRACE("uic_response ret : %d", ret);
		}
	}

	if (ret == DM_OK) {
		ret = RESPONSE_TYPE_OK;
	}

	str_free(&(ui_mode));
	sync_agent_free_noti(noti);
	_INNER_FUNC_EXIT;
	return ret;

 return_part:

	str_free(&(ui_mode));
	sync_agent_free_noti(noti);
	_DEBUG_TRACE("%s : end : %d\n", __func__, ret);
	_INNER_FUNC_EXIT;
	return ret;
}

sync_agent_event_ipc_cancel_e __dm_task_process_check_cancel_flag(void *user_data)
{
	_INNER_FUNC_ENTER;

	retvm_if((user_data) == NULL, SYNC_AGENT_EVENT_IPC_CANCEL_NEED_CANCEL, "user_data is NULL!!");

	sync_agent_ec_boolean cancel_flag = sync_agent_check_cancel_flag();
	sync_agent_event_ipc_cancel_e result_cancel;
	int result_ui;
	char *ui_mode = (char *)user_data;

	if (cancel_flag == 0) {
		result_cancel = SYNC_AGENT_EVENT_IPC_CANCEL_UN_NEED_CANCEL;
		_DEBUG_VERBOSE("%s : cancel_flag = %d\n", __func__, result_cancel);
	} else {
		result_cancel = SYNC_AGENT_EVENT_IPC_CANCEL_NEED_CANCEL;
		_DEBUG_VERBOSE("%s : cancel_flag = %d\n", __func__, result_cancel);
	}

	if (strcmp(ui_mode, OMADM_FUMO_UI_TYPE) == 0) {
		result_ui = aul_app_is_running(OMA_DM_FUMO_UI_PKG);
		if (result_ui == 0) {
			_DEBUG_VERBOSE("oma dm fumo ui process off \n");
			result_cancel = SYNC_AGENT_EVENT_IPC_CANCEL_NEED_CANCEL;
		}
	} else if (strcmp(ui_mode, OMADM_FUMO_NOTI_UI_TYPE) == 0) {
		result_ui = aul_app_is_running(OMA_DM_FUMO_NOTI_UI_PKG);
		if (result_ui == 0) {
			_DEBUG_VERBOSE("oma dm noti ui process off \n");
			result_cancel = SYNC_AGENT_EVENT_IPC_CANCEL_NEED_CANCEL;
		}
	}

	_INNER_FUNC_EXIT;

	return result_cancel;
}

static int __uic_response(int noti_id, void *user_data)
{
	_INNER_FUNC_ENTER;
	int ret = RESPONSE_TYPE_NONE;

	switch (noti_id) {
	case NOTI_ALERT_DISPLAY:	//Type 0
		{
			_DEBUG_TRACE("display alert response ");
			ret = RESPONSE_TYPE_OK;
		}
		break;
	case NOTI_ALERT_CONFIRMATION:	//Type 0
		{
			_DEBUG_TRACE("confirmation alert response ");

			//signal on
			if (uic_timeout_val == TIMEOUT_VAL_UIC) {
				uic_timeout_val = TIMEOUT_VAL_UIC + REST_TIMEOUT_VAL;
			} else {
				uic_timeout_val += REST_TIMEOUT_VAL;
			}

			timeout = g_get_monotonic_time() + uic_timeout_val * G_TIME_SPAN_SECOND;

			_DEBUG_TRACE("timeout : %d ", timeout);

			g_mutex_lock(&(mutex));
			g_cond_wait_until(&(cond), &(mutex), timeout);
			g_mutex_unlock(&(mutex));

			if (res_uic_data == NULL) {
				/*signal by timeout */
				ret = RESPONSE_TYPE_TIMEOVER;
			} else {
				ret = res_uic_data->status;
				_DEBUG_TRACE("response is %d\n", ret);

				free_res_data(res_uic_data);
				res_uic_data = NULL;
			}
		}
		break;
	case NOTI_ALERT_INPUTTEXT:	//Type 0
		{
			//signal on
			if (uic_timeout_val == TIMEOUT_VAL_UIC) {
				uic_timeout_val = TIMEOUT_VAL_UIC + REST_TIMEOUT_VAL;
			} else {
				uic_timeout_val += REST_TIMEOUT_VAL;
			}

			timeout = g_get_monotonic_time() + uic_timeout_val * G_TIME_SPAN_SECOND;

			_DEBUG_TRACE("timeout : %d ", timeout);

			g_mutex_lock(&(mutex));
			g_cond_wait_until(&(cond), &(mutex), timeout);
			g_mutex_unlock(&(mutex));

			uic_data *uic_value = (uic_data *) user_data;

			if (res_uic_data == NULL) {
				/*signal by timeout */
				uic_value->res_data->input_text = strdup("NULL");
				_DEBUG_TRACE("INPUT TEXT CANCEL / TIMEOVER");
				ret = RESPONSE_TYPE_TIMEOVER;
			} else {
				ret = res_uic_data->status;
				_DEBUG_TRACE("response is %d\n", ret);
				_DEBUG_TRACE("input text alert response ");

				if (ret == RESPONSE_TYPE_OK) {
					if (res_uic_data->input_text == NULL) {
						res_uic_data->input_text = strdup("NULL");
					}

					uic_value->res_data->input_text = strdup(res_uic_data->input_text);

				} else {
					uic_value->res_data->input_text = strdup("NULL");
					_DEBUG_TRACE("INPUT TEXT CANCEL / TIMEOVER");
				}

				free_res_data(res_uic_data);
				res_uic_data = NULL;
			}
		}
		break;
	case NOTI_ALERT_SINGLE_CHOICE:	//Type 0
	case NOTI_ALERT_MULTIPLE_CHOICE:	//Type 0
		{

			//signal on
			if (uic_timeout_val == TIMEOUT_VAL_UIC) {
				uic_timeout_val = TIMEOUT_VAL_UIC + REST_TIMEOUT_VAL;
			} else {
				uic_timeout_val += REST_TIMEOUT_VAL;
			}

			timeout = g_get_monotonic_time() + uic_timeout_val * G_TIME_SPAN_SECOND;
			_DEBUG_TRACE("timeout : %d ", timeout);

			g_mutex_lock(&(mutex));
			g_cond_wait_until(&(cond), &(mutex), timeout);
			g_mutex_unlock(&(mutex));

			_DEBUG_TRACE("choice alert response ");

			uic_data *uic_value = (uic_data *) user_data;
			int count = 0;
			char *input = NULL;

			if (res_uic_data == NULL) {
				/*signal by timeout */
				ret = RESPONSE_TYPE_TIMEOVER;
				input = strdup("NULL");
				uic_value->res_data->res_multi_data = g_list_append(uic_value->res_data->res_multi_data, input);
				_DEBUG_TRACE("CHOICE CANCEL / TIMEOVER");
			} else {
				ret = res_uic_data->status;
				_DEBUG_TRACE("response is %d\n", ret);

				if (ret == RESPONSE_TYPE_OK) {

					if (res_uic_data->res_multi_data != NULL) {
						count = g_list_length(res_uic_data->res_multi_data);

						if (count == 0) {
							input = strdup("NULL");
							uic_value->res_data->res_multi_data = g_list_append(uic_value->res_data->res_multi_data, input);
						} else {
							GList *iter = NULL;
							for (iter = res_uic_data->res_multi_data; iter != NULL; iter = g_list_next(iter)) {
								input = NULL;
								input = strdup(iter->data);
								uic_value->res_data->res_multi_data = g_list_append(uic_value->res_data->res_multi_data, input);
							}
						}
					} else {
						input = strdup("NULL");
						uic_value->res_data->res_multi_data = g_list_append(uic_value->res_data->res_multi_data, input);
						_DEBUG_TRACE("CHOICE CANCEL / TIMEOVER");
					}
				}

				free_res_data(res_uic_data);
				res_uic_data = NULL;
			}
		}
		break;
	case NOTI_DOWNLOAD:
	case NOTI_INSTALL:
	case NOTI_DOWNLOAD_INFO:
	case NOTI_ENGINE_START:
	case NOTI_LOW_BATTERY:
	case NOTI_OVER_BIG_SIZE:
		break;

	default:
		break;
	}

	_INNER_FUNC_EXIT;
	return ret;
}
