/*
 * oma-dm-agent
 * Copyright (c) 2012 Samsung Electronics Co., Ltd.
 *
 * Licensed under the Apache License, Version 2.0 (the License);
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

/*lib*/
#include <string.h>

/*sync-agent*/
#include <sync_agent.h>

/*dm-agent*/
#include "common/dm_common.h"
#include "common/dm_common_def.h"
#include "common/util/util.h"
#include "framework/ui-event-handler/user-interaction/user_interaction.h"
#include "serviceadapter/sa_session.h"
#include "serviceadapter/sa_session_internal.h"
#include "serviceadapter/sa_elements.h"
#include "serviceadapter/sa_elements_internal.h"
#include "serviceadapter/sa_command.h"
#include "serviceadapter/sa_command_internal.h"
#include "serviceadapter/dm-phase-handler/dm_phase_handler.h"
#include "mo-handler/dm_mo_common.h"
#include "mo-handler/dm_mo_handler.h"
#include "dm-engine/dm_common_engine.h"

#ifndef OMADM_AGENT_LOG
#undef LOG_TAG
#define LOG_TAG	"OMA_DM_SA"
#endif

//#define DEFAULT_SERVER_MAXMSGSIZE 61440
#define DEFAULT_SERVER_MAXMSGSIZE 5120

static DM_ERROR _receive_alert_status(Session * session, Status * status);
static DM_ERROR _receive_results_status(Session * session, Status * status);

static DM_ERROR _receive_alert_cmd(Session * session, Command * command);
static DM_ERROR _receive_get_cmd(Session * session, Command * command);
static DM_ERROR _receive_results_cmd(Session * session, Command * command);
static DM_ERROR _receive_changes_cmd(Session * session, Command * command);
static DM_ERROR _receive_exec_cmd(Session * session, Command * command);
static DM_ERROR _receive_atomic_cmd(Session * session, Command * command);
static DM_ERROR _receice_sequence_cmd(Session * session, Command * command);

static DM_ERROR __receive_get_property_cmd(char *server_id, char *full_path, char *real_full_path, propoerty_type pr_type, Item ** pItem, int *resultStatus);
static DM_ERROR __rollback_atomic(Session * session, Command * pCommand);
static DM_ERROR __userInteration_operation(Command * command, GList ** response, int *result_status);

static void _free_pending_statuses(GList * pendingStatus);
static void __free_pending_status(PendingStatus * pendingStatus);

static void _set_hmac_value(char **key, char *value, int size);

static DM_ERROR _receive_get_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");
	retvm_if((command->private.access.items) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command->private.access.items is NULL!!");

	int resultStatus = DM_OK;

	Status *temp = NULL;
	Command *pCommand = NULL;

	char *full_path = NULL;
	int error_acl = 0;

	ret = create_results_cmd(session, &pCommand);
	if (ret != DM_OK)
		goto error;

	set_resultscommand_msgref(pCommand, command->msgID);
	set_resultscommand_cmdref(pCommand, command->cmdID);

	GList *iter = NULL;
	Item *item = NULL;
	char *format = NULL;
	char *contenttype = NULL;
	char *data = NULL;
	unsigned int size = 0;
	int moreData = 0;
	Item *pItem;
	int is_mo = 0;
	sync_agent_dm_mo_error_e err_code = SYNC_AGENT_DM_MO_SUCCESS;
	/* URL length over size : 414 error */
	for (iter = command->private.access.items; iter != NULL; iter = g_list_next(iter)) {
		item = (iter->data);

		if (item == NULL) {
			_DEBUG_TRACE("item null");
			resultStatus = DM_ERR_NOT_FOUND;
			goto error;
		}

		char *target = get_location_locuri(item->target);
		if (target != NULL) {
			full_path = g_strdup(target);
		}

		GList *pItems = NULL;
		GetType gettype = GET_TYPE_NOMAL;
		char *real_full_path = NULL;

		ret = get_type_mo_tree(full_path, &gettype, &real_full_path);
		if (ret != DM_OK)
			goto error;

		switch (gettype) {
		case GET_TYPE_NOMAL:
			{
				_DEBUG_TRACE(" get_type_nomal \n");

				err_code = sync_agent_is_exist_mo(full_path, &is_mo);
				_DEBUG_TRACE("is existed mo : %d", err_code);
				_DEBUG_TRACE("is mo : %d", is_mo);
				if (is_mo == 0) {
					_DEBUG_TRACE("NOT EXIST MO");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("EXIST MO");

				//error_acl = 1;
				err_code = sync_agent_check_acl_value(full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_GET, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype permission ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					ret = get_mo_node_format_contenttype(full_path, &format, &contenttype);
					if (ret != DM_OK)
						goto error;

					ret = get_mo_data(full_path, &data);
					if (ret != DM_OK)
						goto error;

					if (data != NULL) {
						size = strlen(data);
						if (session->targetMaxMsgSize - DEFAULT_MORE_REST_SIZE < size) {
							_DEBUG_TRACE("set more data ");
							moreData = 1;
						}
					}

					ret = construct_Item(full_path, format, contenttype, data, size, moreData, &pItem);
					if (ret != DM_OK)
						goto error;

					pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
				}
			}
			break;
		case GET_TYPE_STRUCT:
			{
				_DEBUG_TRACE(" get_type_struct  \n");
				err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
				_DEBUG_TRACE("is existed mo : %d", err_code);
				_DEBUG_TRACE("is mo : %d", is_mo);
				if (is_mo == 0) {
					str_free(&real_full_path);
					_DEBUG_TRACE("NOT EXIST MO");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("EXIST MO");

				//error_acl = 1;
				err_code = sync_agent_check_acl_value(real_full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_GET, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					str_free(&real_full_path);
					_DEBUG_TRACE("acl get fail");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					//free_command(pCommand);
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype PERMISSION ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					ret = get_struct(real_full_path, &pItems);
					if (ret != DM_OK) {
						str_free(&real_full_path);
						goto error;
					}

					str_free(&real_full_path);
					pCommand->private.results.items = pItems;
				}
			}
			break;
		case GET_TYPE_STRUCTDATA:
			{
				err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
				_DEBUG_TRACE("is existed mo : %d", err_code);
				_DEBUG_TRACE("is mo : %d", is_mo);
				if (is_mo == 0) {
					str_free(&real_full_path);
					_DEBUG_TRACE("NOT EXIST MO");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("EXIST MO");

				_DEBUG_TRACE(" get_type_structdata \n");
				//error_acl = 1;
				err_code = sync_agent_check_acl_value(real_full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_GET, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					str_free(&real_full_path);
					_DEBUG_TRACE("acl get fail");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					//free_command(pCommand);
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype PERMISSION ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					ret = get_structdata(real_full_path, &pItems);
					if (ret != DM_OK) {
						str_free(&real_full_path);
						goto error;
					}

					str_free(&real_full_path);
					pCommand->private.results.items = pItems;
				}
			}
			break;
		case GET_TYPE_TNDS:
			{
				_DEBUG_TRACE(" get_type_tnds \n");
				/*xml, wbxml */
				/*currnt XML */
				err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
				_DEBUG_TRACE("is existed mo : %d", err_code);
				if (is_mo == 0) {
					str_free(&real_full_path);
					_DEBUG_TRACE("NOT EXIST MO");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("EXIST MO");

				//error_acl = 1;
				err_code = sync_agent_check_acl_value(real_full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_GET, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					str_free(&real_full_path);
					_DEBUG_TRACE("acl get fail");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					//free_command(pCommand);
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype PERMISSION ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					const char *format = ELEMENT_XML;
					const char *contenttype = ELEMENT_DMTNDS_XML;
					char *output_stream = NULL;
					unsigned int size = 0;
					int moreData = 0;
					Item *pItem;

					ret = get_tnds(ELEMENT_DMTNDS_XML, full_path, &output_stream, &size);
					if (ret != DM_OK)
						goto error;

					_DEBUG_TRACE("max size : %d", session->targetMaxMsgSize);
					_DEBUG_TRACE("data size : %d", size);
					if (output_stream != NULL && session->targetMaxMsgSize - DEFAULT_MORE_REST_SIZE < size) {
						_DEBUG_TRACE("set more data ");
						moreData = 1;
					}

					ret = construct_Item(real_full_path, format, contenttype, output_stream, size, moreData, &pItem);
					if (ret != DM_OK) {
						str_free(&real_full_path);
						goto error;
					}

					str_free(&real_full_path);
					_DEBUG_TRACE("add cmd item");
					pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
				}
			}
			break;
		case GET_TYPE_ACL_PROP:
			{
				ret = __receive_get_property_cmd(session->pServer_id, full_path, real_full_path, PROP_TYPE_ACL, &pItem, &resultStatus);
				if (ret != DM_OK || resultStatus != DM_OK) {
					str_free(&real_full_path);
					goto error;
				}
				str_free(&real_full_path);
				_DEBUG_TRACE("add cmd item");
				pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
			}
			break;
		case GET_TYPE_SIZE_PROP:
			{
				ret = __receive_get_property_cmd(session->pServer_id, full_path, real_full_path, PROP_TYPE_SIZE, &pItem, &resultStatus);
				if (ret != DM_OK || resultStatus != DM_OK) {
					str_free(&real_full_path);
					goto error;
				}
				str_free(&real_full_path);
				_DEBUG_TRACE("add cmd item");
				pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
			}
			break;
		case GET_TYPE_TYPE_PROP:
			{
				ret = __receive_get_property_cmd(session->pServer_id, full_path, real_full_path, PROP_TYPE_TYPE, &pItem, &resultStatus);
				if (ret != DM_OK || resultStatus != DM_OK) {
					str_free(&real_full_path);
					goto error;
				}
				str_free(&real_full_path);
				_DEBUG_TRACE("add cmd item");
				pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
			}
			break;
		case GET_TYPE_FORMAT_PROP:
			{
				ret = __receive_get_property_cmd(session->pServer_id, full_path, real_full_path, PROP_TYPE_FORMAT, &pItem, &resultStatus);
				if (ret != DM_OK || resultStatus != DM_OK) {
					str_free(&real_full_path);
					goto error;
				}
				str_free(&real_full_path);
				_DEBUG_TRACE("add cmd item");
				pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
			}
			break;
		case GET_TYPE_NAME_PROP:
			{
				ret = __receive_get_property_cmd(session->pServer_id, full_path, real_full_path, PROP_TYPE_NAME, &pItem, &resultStatus);
				if (ret != DM_OK || resultStatus != DM_OK) {
					str_free(&real_full_path);
					goto error;
				}
				str_free(&real_full_path);
				_DEBUG_TRACE("add cmd item");
				pCommand->private.results.items = g_list_append(pCommand->private.results.items, pItem);
			}
			break;

		default:
			_DEBUG_TRACE(" get_type_default \n");
			break;
		}

		session->resultsCommand = g_list_append(session->resultsCommand, pCommand);
		_DEBUG_TRACE("add session result command  result command list length : %d", g_list_length(session->resultsCommand));

		ret = create_newstatus(session, resultStatus, command, item->source, item->target, COMMAND_TYPE_GET, &temp);
		if (ret != DM_OK)
			goto error;

		session->status = g_list_append(session->status, temp);

		str_free(&full_path);

	}

	_EXTERN_FUNC_EXIT;
	return ret;

 error:

	str_free(&full_path);
	free_command(pCommand);
	_DEBUG_TRACE(" error : %d\n", ret);

	if (item != NULL) {
		ret = create_newstatus(session, resultStatus, command, item->source, item->target, COMMAND_TYPE_GET, &temp);
		if (ret != DM_OK)
			goto ex_error;
	} else {
		ret = create_newstatus(session, resultStatus, command, command->source, command->target, COMMAND_TYPE_GET, &temp);
		if (ret != DM_OK)
			goto ex_error;
	}

	session->status = g_list_append(session->status, temp);
	_EXTERN_FUNC_EXIT;
	return ret;

 ex_error:

	free_status(temp);
	_EXTERN_FUNC_EXIT;
	return ret;

}

static DM_ERROR __receive_get_property_cmd(char *server_id, char *full_path, char *real_full_path, propoerty_type pr_type, Item ** pItem, int *resultStatus)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((server_id) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "server_id is NULL!!");
	retvm_if((full_path) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "full_path is NULL!!");
	retvm_if((real_full_path) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "real_full_path is NULL!!");

	sync_agent_dm_mo_error_e err_code = SYNC_AGENT_DM_MO_SUCCESS;
	int is_mo = 0;
	int error_acl = 0;
	char *format = NULL;
	char *contenttype = NULL;
	int moreData = 0;
	unsigned int size = 0;

	_DEBUG_TRACE(" get property ");

	err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
	_DEBUG_TRACE("is existed mo : %d", err_code);
	if (is_mo == 0) {
		_DEBUG_TRACE("NOT EXIST MO");
		*resultStatus = DM_ERR_NOT_FOUND;
		goto error;
	}
	_DEBUG_TRACE("EXIST MO");

	//error_acl = 1;
	err_code = sync_agent_check_acl_value(real_full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_GET, server_id, &error_acl);
	_DEBUG_TRACE("sync agent acl result : %d", err_code);
	if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
		_DEBUG_TRACE("acl get fail");
		*resultStatus = DM_ERR_NOT_FOUND;
		goto error;
	}
	_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
	if (error_acl == 0) {
		_DEBUG_TRACE("ACL PERMISSION ERROR");
		*resultStatus = DM_ERR_PERMISSION_DENIED;
		//free_command(pCommand);
		goto error;
	} else if (error_acl == 1) {
		_DEBUG_TRACE("accesstype PERMISSION ERROR");
		*resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
		goto error;
	} else {
		_DEBUG_TRACE("ACL PERMISSION");

		ret = get_mo_node_format_contenttype(real_full_path, &format, &contenttype);
		if (ret != DM_OK)
			goto error;

		char *property = NULL;
		ret = get_mo_node_property(real_full_path, pr_type, &property);
		if (ret != DM_OK) {
			*resultStatus = DM_ERR_PERMISSION_DENIED;
			//str_free(&real_full_path);
			goto error;
		}
		_DEBUG_TRACE("property value : %s", property);
		ret = construct_Item(full_path, format, contenttype, property, size, moreData, pItem);
		if (ret != DM_OK)
			goto error;
	}

	_DEBUG_TRACE(" end %d", ret);
	_INNER_FUNC_EXIT;
	return ret;
 error:

	_DEBUG_TRACE(" end eror : %d", ret);
	_INNER_FUNC_EXIT;
	return ret;
}

static DM_ERROR _receive_alert_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	Status *temp = NULL;
	int result_status = DM_OK;
	GList *response = NULL;

	int alert_data = command->private.alert.type;	//atoi(item->private.data);

	if (DM_ALERT_DISPLAY <= alert_data && DM_ALERT_MULTIPLE_CHOICE >= alert_data) {
		_DEBUG_TRACE("alert ui");
		ret = __userInteration_operation(command, &response, &result_status);
		_DEBUG_TRACE("status : %d", result_status);

	} else if (DM_ALERT_NEXT_MESSAGE <= alert_data && DM_ALERT_MULTIPLE_CHOICE > alert_data) {
		_DEBUG_TRACE("result status : %d", result_status);
		goto error;
	}

	if (ret != DM_OK) {
		ret = DM_OK;
		result_status = DM_ERR_NOT_EXECUTED;
		goto error;
	}

	ret = create_newstatus(session, result_status, command, command->source, command->target, COMMAND_TYPE_ALERT, &temp);
	if (ret != DM_OK) {
		goto ex_error;
	}

	switch (alert_data) {
	case DM_ALERT_DISPLAY:
	case DM_ALERT_CONFIRM_OR_REJECT:
		break;
	case DM_ALERT_TEXT_INPUT:
	case DM_ALERT_SINGLE_CHOICE:
	case DM_ALERT_MULTIPLE_CHOICE:
		{
			Item *pItem = NULL;
			GList *iter = NULL;

			if (response != NULL) {
				_DEBUG_TRACE("--------------------response data existed--------------------");
				for (iter = response; iter != NULL; iter = g_list_next(iter)) {
					if (iter->data != NULL) {
						char *value = NULL;
						value = strdup((char *)iter->data);
						ret = create_item_data(value, strlen(value), &pItem);
						if (ret != DM_OK) {
							if (temp->data != NULL) {
								str_free(&(temp->data));
							}
							temp->data = g_strdup_printf("%d", DM_ERR_NOT_EXECUTED);
							goto return_part;
						}

						temp->items = g_list_append(temp->items, pItem);
					}
				}
			} else {
				_DEBUG_TRACE("--------------------response data not existed--------------------");
				//do nothing
			}
		}
		break;

	}

 return_part:

	session->status = g_list_append(session->status, temp);
	_INNER_FUNC_EXIT;
	return ret;

 error:

	ret = create_newstatus(session, result_status, command, command->source, command->target, COMMAND_TYPE_ALERT, &temp);
	if (ret != DM_OK) {
		goto ex_error;
	}

	session->status = g_list_append(session->status, temp);
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_EXIT;
	return ret;

 ex_error:
	free_status(temp);
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_EXIT;
	return ret;
}

static DM_ERROR _receive_results_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	_INNER_FUNC_EXIT;
	return ret;

}

static DM_ERROR _receive_changes_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	ChangeType type = command->private.change.type;
	_DEBUG_TRACE(" start command type :  %d\n", type);

	int resultStatus = DM_OK;
	int error_acl = 0;
	int is_mo = 0;
	GList *item_iter = NULL;
	Item *item = NULL;
	char *targetUrl = NULL;
	char *sourceUrl = NULL;
	Status *temp = NULL;
	sync_agent_dm_mo_error_e err_code = SYNC_AGENT_DM_MO_SUCCESS;

	char *real_full_path = NULL;
	GetType gettype = GET_TYPE_NOMAL;

	for (item_iter = command->private.change.items; item_iter != NULL; item_iter = g_list_next(item_iter)) {
		item = item_iter->data;
		targetUrl = NULL;
		targetUrl = get_location_locuri(item->target);

		switch (type) {
		case CHANGE_ADD:
			_DEBUG_TRACE(" add command \n");
			err_code = sync_agent_is_exist_mo(targetUrl, &is_mo);
			_DEBUG_TRACE("is exist result : %d", err_code);
			_DEBUG_TRACE("is mo : %d", is_mo);
			if (is_mo == 1) {
				resultStatus = DM_ERR_ALEADY_EXISTS;
				goto error;
			} else {
				//error_acl = 1;
				err_code = sync_agent_check_acl_value(targetUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_ADD, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					if (err_code == SYNC_AGENT_DM_MO_NOT_EXIST_NODE) {
						error_acl = 2;
					} else {
						resultStatus = DM_ERR_PERMISSION_DENIED;
						goto error;
					}
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype PERMISSION ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					ret = add_itemtomo(item, session->pServer_id);
					if (ret != DM_OK)
						goto error;
				}
			}

			break;
		case CHANGE_REPLACE:
			_DEBUG_TRACE(" replace command \n");

			ret = get_type_mo_tree(targetUrl, &gettype, &real_full_path);
			if (ret != DM_OK)
				goto error;

			switch (gettype) {
			case GET_TYPE_NOMAL:
				{
					_DEBUG_TRACE("nomal type : %d", gettype);
					err_code = sync_agent_is_exist_mo(targetUrl, &is_mo);
					_DEBUG_TRACE("is existed mo: %d, result: %d", is_mo, err_code);
					if (is_mo == 0) {
						_DEBUG_TRACE("NOT EXIST MO");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					//error_acl = 1;
					err_code = sync_agent_check_acl_value(targetUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_REPLACE, session->pServer_id, &error_acl);
					_DEBUG_TRACE("sync agent acl result : %d", err_code);
					if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
						_DEBUG_TRACE("acl get fail");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
					if (error_acl == 0) {
						_DEBUG_TRACE("ACL PERMISSION ERROR");
						resultStatus = DM_ERR_PERMISSION_DENIED;
						goto error;
					} else if (error_acl == 1) {
						_DEBUG_TRACE("accesstype PERMISSION ERROR");
						resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
						goto error;
					} else {
						_DEBUG_TRACE("ACL PERMISSION");
						ret = check_mo_scope(targetUrl);
						if (ret != DM_OK) {
							//resultStatus = ret;
							//goto error;
							ret = replace_itemtomo(item); // Permanent node. But Accesstype is allow.
							if (ret != DM_OK) {
								goto error;
							} else { // Permanent node. But device need execution about replace cmd.
								char *server_id = NULL;
								targetUrl = get_location_locuri(item->target);
								server_id = session->pServer_id;
								_DEBUG_TRACE("server_id [%s]", server_id);
								//ret = engine_replace_start(server_id, targetUrl, &(session->service_engine_id), &resultStatus, item);
								//resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
							}
						} else {
							ret = replace_itemtomo(item);
							if (ret != DM_OK)
								goto error;
						}
					}
				}
				break;
			case GET_TYPE_ACL_PROP:
				{
					_DEBUG_TRACE(" property type : %d", gettype);
					err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
					_DEBUG_TRACE("is existed mo : %d", err_code);
					if (is_mo == 0) {
						_DEBUG_TRACE("NOT EXIST MO");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					/*do not check acl for replace property */
					/*err_code = sync_agent_check_acl_value(real_full_path, SYNC_AGENT_DM_MO_ACCESSTYPE_REPLACE, session->pServer_id, &error_acl);
					   _DEBUG_TRACE("sync agent acl result : %d", err_code);
					   if(err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					   resultStatus = DM_ERR_NOT_FOUND;
					   goto error;
					   }
					 */
					/*       error_acl = 2;
					   _DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
					   if(error_acl == 0) {
					   _DEBUG_TRACE("ACL PERMISSION ERROR");
					   resultStatus = DM_ERR_PERMISSION_DENIED;
					   goto error;
					   } else if(error_acl == 1) {
					   _DEBUG_TRACE("accesstype PERMISSION ERROR");
					   resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					   goto error;
					   } else { */
					_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", 2);
					ret = check_mo_scope(real_full_path);
					if (ret != DM_OK) {
						resultStatus = ret;
						goto error;
					} else {
						//check_scope == 1 (dynamic)
						ret = replace_itemtomo(item);
						if (ret != DM_OK)
							goto error;
					}
					//}
				}
				break;
			case GET_TYPE_NAME_PROP:
			case GET_TYPE_TYPE_PROP:
			case GET_TYPE_FORMAT_PROP:
			case GET_TYPE_SIZE_PROP:
				{
					_DEBUG_TRACE(" property type : %d", gettype);
					err_code = sync_agent_is_exist_mo(real_full_path, &is_mo);
					_DEBUG_TRACE("is existed mo : %d", err_code);
					if (is_mo == 0) {
						_DEBUG_TRACE("NOT EXIST MO");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
				}
				break;
			default:
				_DEBUG_TRACE("replace type error");
				break;
			}
			break;
		case CHANGE_DELETE:
			_DEBUG_TRACE(" delete command \n");
			err_code = sync_agent_is_exist_mo(targetUrl, &is_mo);
			_DEBUG_TRACE("is existed mo : %d", err_code);
			if (is_mo == 0) {
				_DEBUG_TRACE("NOT EXIST MO");
				resultStatus = DM_ERR_NOT_FOUND;
				goto error;
			} else {

				//error_acl = 1;
				err_code = sync_agent_check_acl_value(targetUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_DELETE, session->pServer_id, &error_acl);
				_DEBUG_TRACE("sync agent acl result : %d", err_code);
				if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
					_DEBUG_TRACE("acl get fail");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				}
				_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
				if (error_acl == 0) {
					_DEBUG_TRACE("ACL PERMISSION ERROR");
					resultStatus = DM_ERR_PERMISSION_DENIED;
					goto error;
				} else if (error_acl == 1) {
					_DEBUG_TRACE("accesstype PERMISSION ERROR");
					resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
					goto error;
				} else {
					_DEBUG_TRACE("ACL PERMISSION");
					ret = check_mo_scope(targetUrl);
					if (ret != DM_OK) {
						resultStatus = ret;
					} else {
						ret = delete_itemtomo(item);
						if (ret != DM_OK)
							goto error;
					}
				}
			}
			break;
		case CHANGE_COPY:
			_DEBUG_TRACE(" copy command \n");
			sourceUrl = NULL;
			sourceUrl = get_location_locuri(item->source);
			if (sourceUrl != NULL) {
				err_code = sync_agent_is_exist_mo(sourceUrl, &is_mo);
				_DEBUG_TRACE("is existed mo : %d", err_code);
				if (is_mo == 0) {
					_DEBUG_TRACE("NOT EXIT MO");
					resultStatus = DM_ERR_NOT_FOUND;
					goto error;
				} else {
					//error_acl = 1;
					err_code = sync_agent_check_acl_value(sourceUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_COPY, session->pServer_id, &error_acl);
					_DEBUG_TRACE("sync agent acl result : %d", err_code);
					if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
						_DEBUG_TRACE("acl get fail");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
					if (error_acl == 0) {
						_DEBUG_TRACE("ACL PERMISSION ERROR");
						resultStatus = DM_ERR_PERMISSION_DENIED;
						goto error;
					} else if (error_acl == 1) {
						_DEBUG_TRACE("accesstype PERMISSION ERROR");
						resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
						goto error;
					}
					err_code = sync_agent_check_acl_value(targetUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_COPY, session->pServer_id, &error_acl);
					_DEBUG_TRACE("sync agent acl result : %d", err_code);
					if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
						_DEBUG_TRACE("acl get fail");
						resultStatus = DM_ERR_NOT_FOUND;
						goto error;
					}
					_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
					if (error_acl == 0) {
						_DEBUG_TRACE("ACL PERMISSION ERROR");
						resultStatus = DM_ERR_PERMISSION_DENIED;
						goto error;
					} else if (error_acl == 1) {
						_DEBUG_TRACE("accesstype PERMISSION ERROR");
						resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
						goto error;
					} else {
						_DEBUG_TRACE("ACL PERMISSION");
						ret = copy_itemtomo(item);
						if (ret != DM_OK)
							goto error;
					}
				}
			} else {
				//do nothing
			}
			break;
		default:
			break;
		}

		ret = create_newstatus(session, resultStatus, command, item->source, item->target, command->type, &temp);
		if (ret != DM_OK)
			goto error;
		session->status = g_list_append(session->status, temp);
	}

	str_free(&(real_full_path));
	_INNER_FUNC_ENTER;
	return ret;

 error:
	ret = create_newstatus(session, resultStatus, command, command->source, command->target, command->type, &temp);
	if (ret != DM_OK)
		goto ex_error;
	session->status = g_list_append(session->status, temp);

	str_free(&(real_full_path));
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;
 ex_error:

	free_status(temp);
	str_free(&(real_full_path));
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;
}

static DM_ERROR _receive_exec_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	int resultStatus = DM_ERR_ACCEPTED_FOR_PROCESSING;
	int error_acl = 0;
	char *targetUrl = NULL;
	Item *item = NULL;
	item = command->private.exec.item;
	Status *temp = NULL;
	char *pCorrelator = NULL;
	char *server_id = NULL;
	int is_mo = 0;
	sync_agent_dm_mo_error_e err_code = SYNC_AGENT_DM_MO_SUCCESS;

	_DEBUG_TRACE("----------------------------------------------------------------------------exec cmd --------------------------------------------------------------------------- ");
	targetUrl = get_location_locuri(command->private.exec.item->target);
	/*todo: server_id */
	//error_acl = 1;
	err_code = sync_agent_is_exist_mo(targetUrl, &is_mo);
	_DEBUG_TRACE("is existed mo : %d", err_code);
	if (is_mo == 0) {
		_DEBUG_TRACE("NOT EXIT MO");
		resultStatus = DM_ERR_NOT_FOUND;
		goto error;
	} else {

		//error_acl = 1;
		err_code = sync_agent_check_acl_value(targetUrl, SYNC_AGENT_DM_MO_ACCESSTYPE_EXEC, session->pServer_id, &error_acl);
		_DEBUG_TRACE("sync agent acl result : %d", err_code);
		if (err_code != SYNC_AGENT_DM_MO_SUCCESS) {
			_DEBUG_TRACE("acl get fail");
			resultStatus = DM_ERR_NOT_FOUND;
			goto error;
		}
		_DEBUG_TRACE("----------------------------------------------------------------------------------------ACL RETURN--------------------------------------------------- : %d", error_acl);
		if (error_acl == 0) {
			_DEBUG_TRACE("ACL PERMISSION ERROR");
			resultStatus = DM_ERR_PERMISSION_DENIED;
			goto error;
		} else if (error_acl == 1) {
			_DEBUG_TRACE("accesstype PERMISSION ERROR");
			resultStatus = DM_ERR_COMMAND_NOT_ALLOWED;
			goto error;
		} else {
			_DEBUG_TRACE("ACL PERMISSION");
			targetUrl = get_location_locuri(item->target);
			pCorrelator = command->private.exec.correlator;
			server_id = session->pServer_id;

			ret = engine_start(server_id, targetUrl, pCorrelator, &(session->service_engine_id), &resultStatus);
			_DEBUG_TRACE("--------------------------------------------------------engine_start off %d-----------------------------------------------------------------\n", ret);
			if (ret != DM_OK) {
				//ret = DM_ERR_NOT_FOUND;
				goto error;
			}
		}
	}

	ret = create_newstatus(session, resultStatus, command, item->source, item->target, COMMAND_TYPE_EXEC, &temp);
	if (ret != DM_OK)
		goto ex_error;
	session->status = g_list_append(session->status, temp);

	_INNER_FUNC_ENTER;
	return ret;

 error:
	ret = create_newstatus(session, resultStatus, command, item->source, item->target, COMMAND_TYPE_EXEC, &temp);
	if (ret != DM_OK)
		goto ex_error;
	session->status = g_list_append(session->status, temp);

	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;
 ex_error:
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;

}

static DM_ERROR _receice_sequence_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	int resultStatus = DM_OK;
	//sync_agent_dm_mo_error_e mo_ret  = sync_agent_begin_transaction_mo();

	Status *temp = NULL;
	ret = create_newstatus(session, resultStatus, command, command->source, command->target, COMMAND_TYPE_SEQUENCE, &temp);
	if (ret != DM_OK)
		goto error;

	session->status = g_list_append(session->status, temp);

	ret = receive_cmd(session, command->private.sequence_atomic.commands);
	if (ret != DM_OK) {
		resultStatus = DM_ERR_COMMAND_FAILED;
	}
	//mo_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_COMMIT);

	_INNER_FUNC_ENTER;
	return ret;

 error:
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;

}

static DM_ERROR _receive_atomic_cmd(Session * session, Command * command)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	int resultStatus = DM_ERR_ACCEPTED_FOR_PROCESSING;

	GList *atomic_iter = NULL;
	Command *pCommand = NULL;
	Status *temp = NULL;
	sync_agent_dm_mo_error_e mo_ret = sync_agent_begin_transaction_mo();

	for (atomic_iter = command->private.sequence_atomic.commands; atomic_iter != NULL;) {
		pCommand = atomic_iter->data;

		if (pCommand->type == COMMAND_TYPE_GET) {
			ret = create_newstatus(session, DM_ERR_COMMAND_FAILED, pCommand, pCommand->source, pCommand->target, COMMAND_TYPE_GET, &temp);
			if (ret != DM_OK)
				goto error;

			session->status = g_list_append(session->status, temp);

			atomic_iter = g_list_next(atomic_iter);
			command->private.sequence_atomic.commands = g_list_remove(command->private.sequence_atomic.commands, pCommand);
		}
	}

	ret = receive_cmd(session, command->private.sequence_atomic.commands);
	if (ret != DM_OK) {
		resultStatus = DM_ERR_ATOMIC_FAILED;
	}

	if (resultStatus == DM_ERR_ACCEPTED_FOR_PROCESSING) {
		mo_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_COMMIT);
	} else {
		mo_ret = sync_agent_end_transaction_mo(SYNC_AGENT_DM_MO_TRANSACTION_ROLLBACK);
		__rollback_atomic(session, command);
	}

	Status *temp_status = NULL;
	ret = create_newstatus(session, resultStatus, command, pCommand->source, pCommand->target, COMMAND_TYPE_ATOMIC, &temp_status);
	if (ret != DM_OK)
		goto error;

	session->status = g_list_append(session->status, temp_status);

	_INNER_FUNC_ENTER;
	return ret;

 error:
	_DEBUG_TRACE(" error : %d\n", ret);
	_INNER_FUNC_ENTER;
	return ret;

}

static DM_ERROR _receive_alert_status(Session * session, Status * status)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((status) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "status is NULL!!");

	unsigned int code = get_statuscode(status);
	if (code >= 400 && code != DM_ERR_REQUIRE_REFRESH) {
		//401   unauthorized
		//407   Authentication required
		//405   command not allowed
		//406   optional feature not supported
		//500   command failed
		//412   Incomplete command
		//415 unsupported media type or format
		//404 not found ->it have to be considered

		//TODO error handling it's error~!!!!!
		ret = code;
	}
	_DEBUG_TRACE(" end, Error Type : %d\n", ret);
	_INNER_FUNC_EXIT;
	return ret;
}

static DM_ERROR _receive_results_status(Session * session, Status * status)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((status) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "status is NULL!!");

	unsigned int code = get_statuscode(status);
	if (code >= 400) {
		//TODO error handling it's error~!!!!!
		//401 unauthorized
		//407 authentication required
		//404 not found
		//413 request entity too large
		//415 unspported media type or format
		//500 command failed
		ret = code;
	}

	_DEBUG_TRACE(" end, Error Type : %d\n", ret);
	_INNER_FUNC_EXIT;
	return ret;
}

static void _free_pending_statuses(GList * pendingStatus)
{
	_INNER_FUNC_ENTER;

	retm_if((pendingStatus) == NULL, "pendingStatus is NULL!!");

	GList *iter = NULL;
	for (iter = pendingStatus; iter != NULL; iter = g_list_next(iter))
		__free_pending_status(iter->data);

	g_list_free(pendingStatus);

	_INNER_FUNC_EXIT;
}

static void __free_pending_status(PendingStatus * pendingStatus)
{
	_INNER_FUNC_ENTER;

	retm_if((pendingStatus) == NULL, "pendingStatus is NULL!!");

	free(pendingStatus);
	pendingStatus = NULL;

	_INNER_FUNC_EXIT;
}

DM_ERROR create_syncml_session(ProtocolVersion protocolVersion, ProtocolType protocolType, char *server_id, char *client_id, char *pSession_id, char *pSourceUrl, char *pTargetUrl, Session ** pSession)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((pSourceUrl) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "pSourceUrl is NULL!!");
	retvm_if((pTargetUrl) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "pTargetUrl is NULL!!");

	*pSession = (Session *) calloc(1, sizeof(Session));
	if (*pSession == NULL) {
		ret = COMMON_ERR_ALLOC;
		goto error;
	}

	(*pSession)->protocolType = protocolType;
	(*pSession)->protocolVersion = protocolVersion;
/*		(*pSession)->accountId = accountId;*/
	if (server_id != NULL)
		(*pSession)->pServer_id = strdup(server_id);
	(*pSession)->service_engine_id = NO_SERVICE_ENGINE;
	(*pSession)->sessionID = strdup(pSession_id);
	(*pSession)->jsessionid = NULL;

	ret = create_location(pSourceUrl, client_id, &((*pSession)->source));
	if (ret != DM_OK)
		goto error;

	ret = create_location(pTargetUrl, NULL, &((*pSession)->target));
	if (ret != DM_OK)
		goto error;

	(*pSession)->msgID = 0;
	(*pSession)->cmdID = 1;

	(*pSession)->res_chal = NULL;
	(*pSession)->req_chal = NULL;

	(*pSession)->isHMAC = 0;	/*not using hmac */
	(*pSession)->reqhmacinfo = (Hmac *) calloc(1, sizeof(Hmac) + 1);
	if ((*pSession)->reqhmacinfo == NULL) {
		ret = COMMON_ERR_ALLOC;
		goto error;
	}
	(*pSession)->resphmacinfo = (Hmac *) calloc(1, sizeof(Hmac) + 1);
	if ((*pSession)->resphmacinfo == NULL) {
		ret = COMMON_ERR_ALLOC;
		goto error;
	}

	(*pSession)->sourceMaxMsgSize = DEFAULT_MAX_MSG_SIZE;
	(*pSession)->sourceMaxObjSize = DEFAULT_MAX_OBJ_SIZE;

	(*pSession)->targetMaxMsgSize = 0;
	(*pSession)->targetMaxObjSize = 0;

	_EXTERN_FUNC_EXIT;
	return ret;

 error:

	if (*pSession != NULL) {
		_DEBUG_TRACE("free session");
		free_session(*pSession);
	}

	_DEBUG_TRACE(" error : %d\n", ret);
	_EXTERN_FUNC_EXIT;
	return ret;

}

void free_session(Session * session)
{
	_EXTERN_FUNC_ENTER;

	retm_if((session) == NULL, "session is NULL!!");

	if (session->pServer_id != NULL) {
		free(session->pServer_id);
		session->pServer_id = NULL;
	}

	if (session->sessionID != NULL) {
		free(session->sessionID);
		session->sessionID = NULL;
	}

	free_statuses(session->status);
	session->status = NULL;

	free_statuses(session->tempStatus);
	session->status = NULL;

	if (session->target != NULL) {
		free_location(session->target);
		session->target = NULL;
	}

	if (session->source != NULL) {
		free_location(session->source);
		session->source = NULL;
	}

	if (session->orgTarget != NULL) {
		free_location(session->orgTarget);
		session->orgTarget = NULL;
	}

	if (session->cred != NULL) {
		free_cred(session->cred);
		session->cred = NULL;
	}

	if (session->res_chal != NULL) {
		free_chal(session->res_chal);
		session->res_chal = NULL;
	}

	if (session->req_chal != NULL) {
		free_chal(session->req_chal);
		session->req_chal = NULL;
	}

	if (session->reqhmacinfo != NULL) {
		free_hmac(session->reqhmacinfo);
		session->reqhmacinfo = NULL;
	}
	if (session->resphmacinfo != NULL) {
		free_hmac(session->resphmacinfo);
		session->resphmacinfo = NULL;
	}

	_free_pending_statuses(session->pendingStatus);
	session->pendingStatus = NULL;

	if (session->pLargeObjCmd != NULL) {
		free_command(session->pLargeObjCmd);
	}

	session = NULL;

	_EXTERN_FUNC_EXIT;
}

DM_ERROR receive_header(Session * session, SyncHdr * header)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	Location *pLocation = NULL;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((header) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "header is NULL!!");

	if (session->sessionID != NULL && header->sessionID != NULL) {
		if (strcmp(session->sessionID, header->sessionID) != 0) {
			ret = COMMON_ERR_GENERIC;
			goto error;
		}
	}

	session->lastRecievedMsgID = header->messageID;

	if (0 < header->maxmsgsize)
		session->targetMaxMsgSize = header->maxmsgsize;
	else
		session->targetMaxMsgSize = DEFAULT_SERVER_MAXMSGSIZE;

	if (0 < header->maxobjsize)
		session->targetMaxObjSize = header->maxobjsize;

	if (header->responseURI != NULL) {
		if (session->orgTarget == NULL) {
			session->orgTarget = session->target;
			session->target = NULL;
		}

		if (session->target != NULL)
			free_location(session->target);

		ret = create_location(header->responseURI, get_location_locname(session->orgTarget), &pLocation);
		if (ret != DM_OK)
			goto error;
		session->target = pLocation;

		if (session->target == NULL) {
			ret = COMMON_ERR_INTERNAL_NO_MEMORY;
			goto error;
		}

		_DEBUG_INFO("  target %s\n", session->target->locURI);
	}

	Status *temp = NULL;
	DM_ERROR statusData = DM_ERR_UNKNOWN;

	if (header->cred != NULL) {
		_DEBUG_INFO("header cred ");
		//ret = check_server_cred(header->cred, session);
		statusData = DM_ERR_AUTH_ACCEPTED;
		/*
		 * check server cred
		 */
	} else {
		_DEBUG_INFO("header not cred ");
		statusData = DM_OK;
	}

	ret = create_status(statusData, session->cmdID++, session->lastRecievedMsgID, 0, header->source, header->target, NULL, COMMAND_TYPE_HEADER, &temp);
	if (ret != DM_OK)
		goto error;

	session->status = g_list_append(session->status, temp);

	//free_chal(pChal);

	_EXTERN_FUNC_EXIT;
	return ret;

 error:
	//free_chal(pChal);
	_DEBUG_INFO(" error : %d\n", ret);
	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR receive_statuses(Session * session, GList * receiveStatus)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");

	GList *statusItem = receiveStatus;
	Status *status = NULL;
	Chal *pChal = NULL;
	Status *req_chal_status = NULL;
	GList *header = NULL;

	while (statusItem) {
		status = statusItem->data;

		if (status->cmdRef == 0) {
			/*status of SyncHdr */
			assert(status->type == COMMAND_TYPE_HEADER);
			assert(status->data);
			DM_ERROR errorCode = atoi(status->data);

			if (errorCode == DM_ERR_AUTH_ACCEPTED || errorCode == DM_OK || errorCode == DM_ERR_UNAUTHORIZED || errorCode == DM_ERR_AUTHENTICATION_REQUIRED) {
				//212 || 200 || 401 || 407
				//when auth type is AUTH_TYPE_BASIC does not need to send cred in syncHdr in same session
				//when auth type is AUTH_TYPE_MD5 the next nonce in Chal MUST used for the digest when the next sync session is started.

				Chal *chal = status->chal;
				if (chal != NULL) {
					_DEBUG_INFO("exist chal");

					//res_chal in status have to be stored in config_tbl because it have to be used next sync session
					_DEBUG_INFO("format type :%d\n", chal->format);

					/*next session nonce data */
					if (chal->nonce_plain != NULL) {
						set_client_nextnonce(session->pServer_id, chal->nonce_plain);
					}

					/*create next req_chal */
					if (chal->type == AUTH_TYPE_MD5) {
						_DEBUG_INFO("session have cred md5");
						set_client_authtype(session->pServer_id, chal->type);
						ret = create_chal(FORMAT_TYPE_BASE64, AUTH_TYPE_MD5, ELEMENT_DEFAULT_CLIENT_NONCE, &pChal);
						if (ret != DM_OK)
							goto error;
					} else if (chal->type == AUTH_TYPE_HMAC) {
						session->isHMAC = 1;
						_DEBUG_INFO("session have not cred , using hmac");
						set_client_authtype(session->pServer_id, chal->type);
						ret = create_chal(FORMAT_TYPE_BASE64, AUTH_TYPE_HMAC, ELEMENT_DEFAULT_CLIENT_NONCE, &pChal);
						if (ret != DM_OK)
							goto error;
					}

					if (pChal != NULL) {

						for (header = session->status; header != NULL; header = g_list_next(header)) {
							req_chal_status = (Status *) (header->data);
							if (req_chal_status != NULL) {
								_DEBUG_TRACE("req_chal_status status type : %d", req_chal_status->type);
								if (req_chal_status->type == COMMAND_TYPE_HEADER) {
									((Status *) (req_chal_status))->chal = dup_chal(pChal);
									_DEBUG_TRACE("add header chal");
									break;
								}
							}
						}
					}

				} else {
					_DEBUG_INFO("chal null");
				}

				if (errorCode == DM_ERR_AUTH_ACCEPTED || errorCode == DM_ERR_UNAUTHORIZED || errorCode == DM_ERR_AUTHENTICATION_REQUIRED) {
					if (session->cred != NULL) {
						free_cred(session->cred);
						session->cred = NULL;
					} else {
						//do nothing
					}
				}
				// 401 ||  407
				if (errorCode == DM_ERR_UNAUTHORIZED) {
					ret = DM_ERR_UNAUTHORIZED;
					goto error;
				} else if (errorCode == DM_ERR_AUTHENTICATION_REQUIRED) {
					ret = DM_ERR_AUTHENTICATION_REQUIRED;
					goto error;
				}

			} else if (errorCode == DM_ERR_SERVER_FAILURE) {
				ret = DM_ERR_SERVER_FAILURE;
				goto error;
			}
		} else {
			//status except status of SyncHdr
			GList *pendingStatusItem = session->pendingStatus;
			PendingStatus *pending = NULL;
			while (pendingStatusItem) {
				pending = pendingStatusItem->data;

				if (pending->cmdID == status->cmdRef && pending->msgID == status->msgRef) {
					//if status's cmdID and msgID exists in pendigStatus

					if (status->type == COMMAND_TYPE_ALERT) {
						ret = _receive_alert_status(session, status);
					} else if (status->type == COMMAND_TYPE_RESULTS) {
						ret = _receive_results_status(session, status);
					} else {
						/*
						   COMMAND_TYPE_GET, COMMAND_TYPE_ADD, COMMAND_TYPE_REPLACE, COMMAND_TYPE_DELETE
						   COMMAND_TYPE_COPY, COMMAND_TYPE_EXEC, COMMAND_TYPE_ATOMIC, COMMAND_TYPE_SEQUENCE
						 */
					}
					if (ret != DM_OK)
						goto error;

					session->pendingStatus = g_list_remove(session->pendingStatus, pending);
					free(pending);
				} else {
					//if status's cmdID and msgID doesn't exists in pendigStatus
					//is it error?? does this can happen??
				}
				pendingStatusItem = g_list_next(pendingStatusItem);
			}
		}
		statusItem = g_list_next(statusItem);
	}

	free_chal(pChal);
	_EXTERN_FUNC_EXIT;
	return ret;

 error:
	free_chal(pChal);
	_DEBUG_INFO(" error, Error Type %d \n", ret);
	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR receive_cmd(Session * session, GList * receiveCommand)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");

	GList *commandItem = receiveCommand;
	Command *command = NULL;

	while (commandItem) {
		command = commandItem->data;

		switch (command->type) {
		case COMMAND_TYPE_ALERT:
			ret = _receive_alert_cmd(session, command);
			break;
		case COMMAND_TYPE_GET:
			ret = _receive_get_cmd(session, command);
			break;
		case COMMAND_TYPE_RESULTS:
			ret = _receive_results_cmd(session, command);
			break;
		case COMMAND_TYPE_ADD:
		case COMMAND_TYPE_REPLACE:
		case COMMAND_TYPE_DELETE:
		case COMMAND_TYPE_COPY:
			ret = _receive_changes_cmd(session, command);
			break;
		case COMMAND_TYPE_EXEC:
			ret = _receive_exec_cmd(session, command);
			break;
		case COMMAND_TYPE_SEQUENCE:
			ret = _receice_sequence_cmd(session, command);
			break;
		case COMMAND_TYPE_ATOMIC:
			ret = _receive_atomic_cmd(session, command);
			break;
		case COMMAND_TYPE_UNKNOWN:
		case COMMAND_TYPE_HEADER:
			_DEBUG_INFO("not support command");
			break;
		default:
			break;
		}

		_DEBUG_INFO("receivce cmd ret : %d", ret);
		commandItem = g_list_next(commandItem);
		if (ret != DM_OK)
			goto error;
	}

	_EXTERN_FUNC_EXIT;
	return ret;

 error:
	_DEBUG_INFO(" error : %d\n", ret);
	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR create_prependingstatus(unsigned int msgID, unsigned int cmdID, PendingStatus ** pPendingStatus)
{
	_EXTERN_FUNC_EXIT;

	DM_ERROR ret = DM_OK;

	*pPendingStatus = (PendingStatus *) calloc(1, sizeof(PendingStatus));

	if (*pPendingStatus == NULL) {
		ret = COMMON_ERR_ALLOC;
		goto error;
	}

	(*pPendingStatus)->msgID = msgID;
	(*pPendingStatus)->cmdID = cmdID;

	ret = DM_OK;

	_EXTERN_FUNC_EXIT;
	return ret;

 error:
	_DEBUG_INFO(" error : %d\n", ret);
	_EXTERN_FUNC_EXIT;
	return ret;
}

DM_ERROR set_session_authcred(Session * session, Cred * cred)
{
	_EXTERN_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
//      retvm_if((cred) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED , "cred is NULL!!");
	if (cred != NULL) {
		session->cred = cred;
	}

	_EXTERN_FUNC_EXIT;
	return ret;
}

void reset_cmdid_session(Session * session)
{
	_EXTERN_FUNC_ENTER;

	retm_if((session) == NULL, "session is NULL!!");

	session->cmdID = 1;

	_EXTERN_FUNC_EXIT;
}

AuthType __get_Session_AuthType_enum(char *data)
{
	_EXTERN_FUNC_ENTER;

	AuthType type = AUTH_TYPE_UNKNOWN;

	retvm_if((data) == NULL, AUTH_TYPE_MD5, "data is NULL!!");

	if (strcmp(data, ELEMENT_BASIC) == 0) {
		type = AUTH_TYPE_BASIC;
	} else if (strcmp(data, ELEMENT_DIGEST) == 0) {
		type = AUTH_TYPE_MD5;
	} else if (strcmp(data, ELEMENT_HMAC) == 0) {
		type = AUTH_TYPE_HMAC;
	} else {
		type = AUTH_TYPE_UNKNOWN;
	}
	_DEBUG_VERBOSE(" end : [%d] \n", type);
	_EXTERN_FUNC_EXIT;
	return type;
}

void __set_Session_AuthType_name(AuthType authType, char **authname)
{
	_EXTERN_FUNC_ENTER;

	switch (authType) {
	case AUTH_TYPE_BASIC:
		(*authname) = strdup(ELEMENT_BASIC);
		break;
	case AUTH_TYPE_MD5:
		(*authname) = strdup(ELEMENT_DIGEST);
		break;
	case AUTH_TYPE_HMAC:
		(*authname) = strdup(ELEMENT_HMAC);
		break;
	default:
		(*authname) = strdup(ELEMENT_BASIC);
		break;
	}

	_EXTERN_FUNC_EXIT;

}

//temp code
/*
DM_ERROR get_session_clientauthtype(Session * session, AuthType * authType){
	_DEBUG_INFO(" start \n");

	DM_ERROR ret = DM_OK;
	char* data  = NULL;

	 ret = get_dmacc_authType( session->pServer_id, authType);
	if(ret != DM_OK) {
		(*authType) = AUTH_TYPE_BASIC;
		ret = DM_OK;
	}

	if(data != NULL) {
		(*authType) = __get_Session_AuthType(data);
		free(data);
		data = NULL;
	}
	_DEBUG_INFO(" end \n");
	return ret;
error :
	_DEBUG_INFO(" error end [%d]\n", ret);
	return  ret;
}

DM_ERROR get_session_serverauthtype(Session * session, AuthType* authType){
	_DEBUG_INFO(" start \n");

	DM_ERROR ret = DM_OK;
	char* data  = NULL;

	ret = get_dmacc_authType( session->pServer_id, authType);
	if(ret != DM_OK) {
		(*authType) = AUTH_TYPE_BASIC;
		ret = DM_OK;
	}

	if(data != NULL) {
		(*authType) = __get_Session_AuthType(data);

		free(data);
		data = NULL;
	}

	_DEBUG_INFO(" end \n");
	return ret;

error:
	_DEBUG_INFO(" error end [%d]\n", ret);
	return ret;
}
*/

static DM_ERROR __userInteration_operation(Command * command, GList ** response, int *result_status)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;

	retvm_if((command) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "command is NULL!!");

	CommandAlertPrivate pAlert = command->private.alert;
	AlertType type = pAlert.type;
	GList *alertItems = pAlert.items;
	GList *iter = NULL;
	Item *pItem = NULL;
	int size = g_list_length(alertItems);
	int i = 0;
	//GList * responseItems = NULL;
	char *result_data;
	/*char* findValue = NULL; */
	/*int isCancel = 0; */
	uic_data *pUic_data = (uic_data *) calloc(1, sizeof(uic_data) + 1);
	if (pUic_data == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		_INNER_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	pUic_data->pUic_option = (uic_option *) calloc(1, sizeof(uic_option) + 1);
	if ((pUic_data->pUic_option) == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		free_uic_value(&pUic_data);
		_INNER_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	pUic_data->res_data = (uic_res_data *) calloc(1, sizeof(uic_res_data) + 1);
	//uic_option *uic_value = (uic_option *)calloc(1,sizeof(uic_option));
	if ((pUic_data->res_data) == NULL) {
		_DEBUG_VERBOSE("alloc fail");
		free_uic_value(&pUic_data);
		_INNER_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	switch (type) {
	case DM_ALERT_DISPLAY:
		_DEBUG_VERBOSE("alert display ui");
		for (iter = alertItems, i = 0; iter != NULL; iter = g_list_next(iter), ++i) {
			pItem = iter->data;
			if (i == 0 && size != 1) {
				/*user interaction option */
				_DEBUG_VERBOSE("uic opation : %s", pItem->private.data);
				ret = exec_userinteration_option(pItem->private.data, &(pUic_data->pUic_option));
			} else {
				if (pItem->private.data != NULL) {
					/*user interaction execute */
					_DEBUG_VERBOSE("uic display : %s", pItem->private.data);
					pUic_data->display = strdup(pItem->private.data);
					ret = ui_display(pUic_data);
					*result_status = pUic_data->res_data->status;
				}
			}
		}
		break;
	case DM_ALERT_CONFIRM_OR_REJECT:
		_DEBUG_VERBOSE("alert confirm ui");
		_DEBUG_VERBOSE("alert confirm alert item size : %d", size);
		for (iter = alertItems, i = 0; iter != NULL; iter = g_list_next(iter), ++i) {
			pItem = iter->data;
			if (i == 0 && size != 1) {
				/*user interaction option */
				ret = exec_userinteration_option(pItem->private.data, &(pUic_data->pUic_option));
			} else {
				/*user interaction execute */
				_DEBUG_VERBOSE("uic display : %s", pItem->private.data);
				pUic_data->display = strdup(pItem->private.data);
				ret = ui_confirm_or_reject(pUic_data);
				*result_status = pUic_data->res_data->status;
			}
		}
		break;
	case DM_ALERT_TEXT_INPUT:
		_DEBUG_VERBOSE("alert input ui");
		for (iter = alertItems, i = 0; iter != NULL; iter = g_list_next(iter), ++i) {
			pItem = iter->data;
			if (i == 0 && size != 1) {
				/*user interaction option */
				/*IT ( A , N, D, T , P , I )
				 *ET (T , P )
				 */
				ret = exec_userinteration_option(pItem->private.data, &(pUic_data->pUic_option));
			} else {
				/*user interaction execute */
				_DEBUG_VERBOSE("uic display : %s", pItem->private.data);
				pUic_data->display = strdup(pItem->private.data);
				ret = ui_text_input(pUic_data, &result_data);
				*result_status = pUic_data->res_data->status;
				if (pUic_data->res_data->status == DM_OK) {
					(*response) = g_list_append((*response), result_data);
				} else {
					_DEBUG_VERBOSE("input text null");
				}
			}
		}

		break;
	case DM_ALERT_SINGLE_CHOICE:{
			_DEBUG_VERBOSE("alert single choice ui");
			for (iter = alertItems, i = 0; iter != NULL; iter = g_list_next(iter), ++i) {
				pItem = iter->data;
				if (i == 0 && size != 1) {
					/*user interaction option */
					ret = exec_userinteration_option(pItem->private.data, &(pUic_data->pUic_option));
				} else {
					/*user interaction execute */
					if (i == 1) {
						_DEBUG_VERBOSE("uic display : %s", pItem->private.data);
						pUic_data->display = strdup(pItem->private.data);
					} else {
						GList *iter_end = NULL;
						iter_end = g_list_next(iter);
						char *data = NULL;
						data = strdup(pItem->private.data);
						pUic_data->multi_data = g_list_append(pUic_data->multi_data, data);

						if (iter_end == NULL) {
							ret = ui_single_choice(pUic_data, response);
							*result_status = pUic_data->res_data->status;
							break;
						}
					}
				}
			}
		}
		break;
	case DM_ALERT_MULTIPLE_CHOICE:{
			_DEBUG_VERBOSE("alert single/multiple choice ui");
			for (iter = alertItems, i = 0; iter != NULL; iter = g_list_next(iter), ++i) {
				pItem = iter->data;
				if (i == 0 && size != 1) {
					/*user interaction option */
					if (pItem->private.data != NULL)
						ret = exec_userinteration_option(pItem->private.data, &(pUic_data->pUic_option));
				} else {
					/*user interaction execute */
					if (i == 1) {
						_DEBUG_VERBOSE("uic display : %s", pItem->private.data);
						pUic_data->display = strdup(pItem->private.data);
					} else {
						GList *iter_end = NULL;
						iter_end = g_list_next(iter);
						char *data = NULL;
						data = strdup(pItem->private.data);
						pUic_data->multi_data = g_list_append(pUic_data->multi_data, data);

						if (iter_end == NULL) {
							ret = ui_multiple_choice(pUic_data, response);
							*result_status = pUic_data->res_data->status;
							break;
						}
					}
				}
			}
		}
		break;
	case ALERT_UNKNOWN:
	case DM_ALERT_SERVER_INITIATED_MGMT:
	case DM_ALERT_CLIENT_INITIATED_MGMT:
	case DM_ALERT_NEXT_MESSAGE:
	case DM_ALERT_SESSION_ABORT:
	case DM_ALERT_CLIENT_EVENT:
	case DM_ALERT_NO_END_OF_DATA:
	case DM_ALERT_GENERIC_ALERT:
		_DEBUG_VERBOSE("not user interaction type");
		break;
	default:
		break;
	}

	/*
	   session->status
	   for(iter = responseItems; iter != NULL; iter = g_list_next(iter)){
	   pItem = iter->data;
	   temp->items = g_list_append(temp->items, pItem);
	   }
	 */
	free_uic_value(&pUic_data);

	_DEBUG_VERBOSE(" end\n");
	_INNER_FUNC_EXIT;
	return ret;

}

DM_ERROR add_response_hmacinfo(Session ** session, char *temp_hmac)
{
	_EXTERN_FUNC_ENTER;
	DM_ERROR ret = DM_OK;
	/*x-syncml-hmac: algorithm=MD5, username="156354238", mac=hvMeISwxSVUEofF2jHhQjA== */

	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((temp_hmac) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "hmac is NULL!!");

	char *start = NULL;
	char *end = NULL;

	if ((*session)->resphmacinfo == NULL) {
		_DEBUG_INFO("hmac response calloc");
		(*session)->resphmacinfo = (Hmac *) calloc(1, sizeof(Hmac) + 1);
		if ((*session)->resphmacinfo == NULL) {
			return COMMON_ERR_ALLOC;
		}
	}

	_DEBUG_INFO("temp_hmac : %s", temp_hmac);
	start = strstr(temp_hmac, "=");
	if (start == NULL) {
		_DEBUG_INFO("auth type null");
		return COMMON_ERR_INTERNAL_NOT_DEFINED;
	}
	_DEBUG_INFO("hmac : %s", start);

	end = strstr(temp_hmac, ",");
	if (end == NULL) {
		_DEBUG_INFO("auth type null");
		return COMMON_ERR_INTERNAL_NOT_DEFINED;
	}
	_DEBUG_INFO("hmac : %s", end);

	_set_hmac_value(&((*session)->resphmacinfo->authtype), start + 1, (end - start - 1));
	if ((*session)->resphmacinfo->authtype == NULL) {
		_DEBUG_INFO("alloc fail");
		_EXTERN_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	_DEBUG_INFO("auty type : %s", (*session)->resphmacinfo->authtype);

	start = strstr(end, "=");
	if (start == NULL) {
		_DEBUG_INFO("username null");
		return COMMON_ERR_INTERNAL_NOT_DEFINED;
	}
	_DEBUG_INFO("hmac : %s", start);

	end = strstr(start, ",");
	if (end == NULL) {
		_DEBUG_INFO("username null");
		return COMMON_ERR_INTERNAL_NOT_DEFINED;
	}
	_DEBUG_INFO("hmac : %s", end);

	_set_hmac_value(&((*session)->resphmacinfo->username), start + 2, (end - start - 3));
	if ((*session)->resphmacinfo->username == NULL) {
		_DEBUG_INFO("alloc fail");
		_EXTERN_FUNC_EXIT;
		return COMMON_ERR_ALLOC;
	}
	_DEBUG_INFO("username : %s", (*session)->resphmacinfo->username);

	start = strstr(end, "=");
	if (start == NULL) {
		_DEBUG_INFO("mac null");
		return COMMON_ERR_INTERNAL_NOT_DEFINED;
	}
	_DEBUG_INFO("hmac : %s", start);
	(*session)->resphmacinfo->mac = strdup(start + 1);
	_DEBUG_INFO("mac : %s", (*session)->resphmacinfo->mac);

	_EXTERN_FUNC_EXIT;
	return ret;
}

static void _set_hmac_value(char **key, char *value, int size)
{
	_INNER_FUNC_ENTER;

	retm_if((value) == NULL, "value is NULL!!");
	retm_if((size) == 0, "size is NULL!!");

	if (*key != NULL) {
		free(*key);
		*key = NULL;
	} else {
		_DEBUG_INFO("key null");
	}

	(*key) = (char *)calloc(1, size + 1);
	if ((*key) == NULL) {
		_INNER_FUNC_EXIT;
		return;
	}

	strncpy((*key), value, size);

	_INNER_FUNC_EXIT;
}

/*DM_ERROR process_commands(Command* pCommand,  int * resultStatus) {

	_DEBUG_TRACE( " start\n");

	DM_ERROR ret = DM_OK;
	switch (pCommand->type) {
	case COMMAND_TYPE_UNKNOWN:
		break;
	case COMMAND_TYPE_ALERT:
		ret = __process_alert(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_ADD:
		break;
		ret = __process_add(pCommand, resultStatus);
	case COMMAND_TYPE_REPLACE:
		ret = __process_replace(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_DELETE:
		ret = __process_delete(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_EXEC:
		ret = __process_exec(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_SEQUENCE:
		ret = __process_sequence(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_ATOMIC:
		ret = __process_atomic(pCommand, resultStatus);
		break;
	case COMMAND_TYPE_COPY:
		break;
	default:
		break;
	}

return_part:
	if (ret != DM_OK)
		goto error;

	_DEBUG_TRACE( " end\n");
	return ret;
error:
	_DEBUG_TRACE( " error end : ret %d\n", ret);
	return ret;

}*/

static DM_ERROR __rollback_atomic(Session * session, Command * pCommand)
{
	_INNER_FUNC_ENTER;

	DM_ERROR ret = DM_OK;
	retvm_if((session) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "session is NULL!!");
	retvm_if((pCommand) == NULL, COMMON_ERR_INTERNAL_NOT_DEFINED, "pCommand is NULL!!");

	Command *current_Command = NULL;
	/*Status * temp_status = NULL; */
	GList *atomic_iter = NULL;
	GList *items = NULL;
	GList *temp_items = NULL;
	Item *item = NULL;
	sync_agent_dm_mo_error_e mo_ret = sync_agent_begin_transaction_mo();
	_DEBUG_VERBOSE("transaction start : %d", mo_ret);

	for (atomic_iter = pCommand->private.sequence_atomic.commands; atomic_iter != NULL; atomic_iter = g_list_next(atomic_iter)) {
		current_Command = (Command *) atomic_iter->data;

		switch (current_Command->type) {
		case COMMAND_TYPE_ALERT:
			items = current_Command->private.alert.items;
			break;
		case COMMAND_TYPE_ADD:
		case COMMAND_TYPE_REPLACE:
		case COMMAND_TYPE_DELETE:
		case COMMAND_TYPE_COPY:
			items = current_Command->private.change.items;
			break;
		case COMMAND_TYPE_GET:
			items = current_Command->private.access.items;
			// delete result status
			break;
		case COMMAND_TYPE_UNKNOWN:
		case COMMAND_TYPE_HEADER:
		case COMMAND_TYPE_RESULTS:
		case COMMAND_TYPE_EXEC:
		case COMMAND_TYPE_SEQUENCE:
		case COMMAND_TYPE_ATOMIC:
			_DEBUG_VERBOSE("not support command in atomic");
			break;
		default:
			break;
		}

		for (temp_items = items; temp_items != NULL; temp_items = g_list_next(temp_items)) {
			item = temp_items->data;

			ret = set_status_data(session, current_Command, item, DM_ERR_ATOMIC_ROLL_BACK_OK);
		}
	}

	_DEBUG_VERBOSE(" end\n");
	_INNER_FUNC_EXIT;
	return ret;

}
